/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: icmp6.c$
* $Version : 3.8.5.0$
* $Date    : Sep-19-2012$
*
* Comments:
*
*   This file contains the implementation of a minimal
*   Internet Control Message Protocol 6. 
*
*   2011.09: Added IPv6 support by Andrey Butok
*
*END************************************************************************/

#include <rtcs.h>

#if RTCSCFG_ENABLE_IP6

#include "rtcs_prv.h"
#include "tcpip.h"
#include "icmp6_prv.h"
#include "tcp_prv.h"
#include "udp_prv.h"
#include "ip6_trace.h"

/* //TBD RFC 4443: If an ICMPv6 error message of unknown type is received at its
destination, it MUST be passed to the upper-layer process that
originated the packet that caused the error, where this can be
identified
...
In cases where the internet-layer protocol is required to pass an
ICMPv6 error message to the upper-layer process, the upper-layer
protocol type is extracted from the original packet (contained in
the body of the ICMPv6 error message) and used to select the
appropriate upper-layer process to handle the error.*/

/************************************************************************
* NAME: ICMP6_init
*
* DESCRIPTION: Initialize the ICMP layer.
*************************************************************************/
uint_32 ICMP6_init( void )
{ 
   ICMP6_CFG_STRUCT_PTR     icmp6_cfg_ptr;
   uint_32                  status;     /* return code */

   icmp6_cfg_ptr = RTCS_mem_alloc_zero(sizeof(ICMP6_CFG_STRUCT));
   if (icmp6_cfg_ptr == NULL)
   {
      return RTCSERR_OUT_OF_MEMORY;
   }

   _mem_set_type(icmp6_cfg_ptr, MEM_TYPE_ICMP6_CFG_STRUCT);
   
   RTCS_setcfg(ICMP6, icmp6_cfg_ptr);

   IP6_open(IPPROTO_ICMPV6, ICMP6_service, NULL, &status);

   return status;
}

/************************************************************************
* NAME: ICMP6_service
*
* DESCRIPTION: Process incoming ICMP6 packets.  Called from IP6_service.
*************************************************************************/
void ICMP6_service (RTCSPCB_PTR rtcs_pcb /* [IN/OUT] incoming packet */, pointer dummy /* [IN]  not used */)
{
    ICMP6_CFG_STRUCT_PTR	ICMP6_cfg_ptr;
    ICMP6_HEADER_PTR		icmp6_packet;
    uint_32                 icmp6_packet_size = RTCSPCB_SIZE(rtcs_pcb);
    uint_16					chksum;
    uchar        			type;
    IP6_HEADER_PTR          ip6_packet = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(rtcs_pcb);
    IP_IF_PTR               if_ptr;
    in6_addr                ip_src;
    in6_addr                ip_dest;
    int                     result = RTCS_OK;

    if_ptr = rtcs_pcb->IFSRC;
    
    ICMP6_cfg_ptr = RTCS_getcfg(ICMP6);   
    
    IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_TOTAL++);

    /*
    ** Make sure that
    **    sizeof(ICMP6_HEADER) <= RTCSPCB_SIZE(pcb)
    */
    if (icmp6_packet_size < sizeof(ICMP6_HEADER))
    {
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.ST_RX_SMALL_DGRAM++);
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_ICMP_BAD_HEADER);
        RTCSPCB_free(rtcs_pcb);
        return;
    } 

    icmp6_packet = (ICMP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    type = ntohc(icmp6_packet->TYPE);
    
    /* Drop Multicast loopback.*/
    {
        IP_CFG_STRUCT_PTR    IP_cfg_ptr = RTCS_getcfg(IP);
         
        if ((rtcs_pcb->IFSRC == RTCS_IF_LOCALHOST_PRV)
            && IN6_IS_ADDR_MULTICAST((in6_addr *)ip6_packet->DEST))
        {
            IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
            RTCSPCB_free(rtcs_pcb);
            return;
        }
    }

    /*
    * Verify the checksum.
    */
    chksum = IP6_Sum_pseudo(RTCSPCB_SIZE(rtcs_pcb), rtcs_pcb);
    chksum = IP_Sum_PCB(chksum, rtcs_pcb);
    if (chksum != 0xFFFF)
    {
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.ST_RX_BAD_CHECKSUM++);
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_ICMP_BAD_CHECKSUM);
        RTCSPCB_free(rtcs_pcb);
        return;
    }
    
    RTCSLOG_PCB_READ(rtcs_pcb, RTCS_LOGCTRL_PROTO(IPPROTO_ICMPV6), 0);

    /* Save source and target addresses.*/ //TBD 
    IN6_ADDR_COPY((in6_addr *)ip6_packet->SOURCE, &ip_src);
    IN6_ADDR_COPY((in6_addr *)ip6_packet->DEST, &ip_dest);
    
	/************************************************************
	* Process incoming ICMPv6 packets.
	*************************************************************/
	switch (type) 
	{
        /* ---- Neighbor Solicitation. ----------------------------- */
        case ICMP6_TYPE_NEIGHBOR_SOLICITATION:
            result = nd6_neighbor_solicitation_receive(rtcs_pcb);
        break;
        /* ---- Neighbor Advertisemnt. ----------------------------- */
        case ICMP6_TYPE_NEIGHBOR_ADVERTISEMENT:
            result = nd6_neighbor_advertisement_receive(rtcs_pcb);
            break;
        /* ---- Router Solicitation. ------------------------------- */
        /* (ICMP6_TYPE_NEIGHBOR_SOLICITATION)
         * RFC 4861: Hosts MUST silently discard any received Router Solicitation
         * Messages.
         */    		
        /* ---- Router Advertisemnt. ----------------------------- */
        case ICMP6_TYPE_ROUTER_ADVERTISEMENT:
            result = nd6_router_advertisement_receive(rtcs_pcb);
            break;         
        /* ---- Echo Request. -------------------------------------- */
        /* RFC4443 4.1: Every node MUST implement an ICMPv6 Echo responder function that
         *              receives Echo Requests and originates corresponding Echo Replies.
         */
        case ICMP6_TYPE_ECHO_REQ:
        
        #if RTCSCFG_DEBUG_MSG      
            printf("ICMP6: Echo Req Rx\n");   
        #endif            
  
            result = RTCSPCB_next(rtcs_pcb, sizeof(ICMP6_HEADER));
            if (result)
            {
                IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++);
                IF_ICMP6_STATS_ENABLED(RTCS_seterror(&ICMP6_cfg_ptr->STATS.ERR_RX, result, (uint_32)rtcs_pcb));
                RTCSLOG_PCB_FREE(rtcs_pcb, result);
                RTCSPCB_free(rtcs_pcb);
                return;
            }

            RTCSPCB_fork(rtcs_pcb);

            IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_TX_TOTAL++);
            RTCSLOG_PCB_WRITE(rtcs_pcb, RTCS_LOGCTRL_PROTO(IPPROTO_ICMPV6), 0);

            result = RTCSPCB_insert_header(rtcs_pcb, sizeof(ICMP6_HEADER));
            if (result)
            {
                IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.ST_RX_ECHO_REQ++);
                IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                IF_ICMP6_STATS_ENABLED(RTCS_seterror(&ICMP6_cfg_ptr->STATS.ERR_TX, result, (uint_32)rtcs_pcb));
                RTCSLOG_PCB_FREE(rtcs_pcb, result);
                RTCSPCB_free(rtcs_pcb);
                return;
            }

            /* Change type from Echo to Echo Reply and recalculate checksum */
            icmp6_packet = (ICMP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
            htonc(icmp6_packet->TYPE, ICMP6_TYPE_ECHO_REPLY);
            htonc(icmp6_packet->CODE, 0);
            htons(icmp6_packet->CHECKSUM, 0);
            rtcs_pcb->IP_SUM     = IP_Sum_PCB(RTCSPCB_SIZE(rtcs_pcb), rtcs_pcb);
            rtcs_pcb->IP_SUM_PTR = icmp6_packet->CHECKSUM;

            /* Send the Echo Reply whence came the Echo */
            IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.ST_TX_ECHO_REPLY++);
            IP6_send(rtcs_pcb, IPPROTO_ICMPV6, &ip_dest/*ipsrc*/, &ip_src/*ipdest*/, if_ptr, 0);
            
            rtcs_pcb = NULL; /* PCB is reused.*/
		    break;
        /* ---- Echo Reply. -------------------------------------- */		    
        case ICMP6_TYPE_ECHO_REPLY:
            {
                ICMP6_ECHO_HEADER_PTR echopacket = (ICMP6_ECHO_HEADER_PTR)icmp6_packet;
                ICMP6_PARM_PTR        parms;
                uint_16              id, seq;

                /*
                ** Make sure that
                **    sizeof(ICMP_ECHO_HEADER) <= RTCSPCB_SIZE(pcb)
                */
                if (RTCSPCB_SIZE(rtcs_pcb) < sizeof(ICMP6_ECHO_HEADER))
                {
                    IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
                    IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.ST_RX_SMALL_DGRAM++);
                    RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_ICMP_BAD_HEADER);
                    RTCSPCB_free(rtcs_pcb);
                    return;
                }

                /*
                ** get the ID and Sequence number from the packet
                */
                id  = ntohs(echopacket->ID);
                seq = ntohs(echopacket->SEQ);

                /*
                ** Find a match for the ID and sequence number
                */
                for (parms=ICMP6_cfg_ptr->ECHO_PARM_HEAD; parms; parms=parms->NEXT)
                {

                    if ((parms->id == id) && (parms->seq == seq))
                    {
                        /*  received reply for the ping request */

                        if (parms->NEXT)
                        {
                            parms->NEXT->PREV = parms->PREV;
                        } /* Endif */
                        *parms->PREV = parms->NEXT;

                        TCPIP_Event_cancel(&parms->EXPIRE);

                        /* calculate round trip time */
                        parms->timeout = (RTCS_time_get() - parms->timeout);

                        RTCSCMD_complete(parms, RTCS_OK);
                        break;
                    } 
                } 
            }
            break;		    			
	} /* switch (type) */
	
    trace_icmp6("RX", icmp6_packet); /* Trace ICMP6.*/


#if RTCSCFG_ENABLE_ICMP6_STATS
    /* Update the statistics */
    switch (type)
    {
        case ICMP6_TYPE_DEST_UNREACH:   ICMP6_cfg_ptr->STATS.ST_RX_DESTUNREACH++; break;
        case ICMP6_TYPE_TIME_EXCEED:    ICMP6_cfg_ptr->STATS.ST_RX_TIMEEXCEED++;  break;
        case ICMP6_TYPE_PARAM_PROB:     ICMP6_cfg_ptr->STATS.ST_RX_PARMPROB++;    break;
        case ICMP6_TYPE_REDIRECT:       ICMP6_cfg_ptr->STATS.ST_RX_REDIRECT++;    break;
        case ICMP6_TYPE_ECHO_REQ:       ICMP6_cfg_ptr->STATS.ST_RX_ECHO_REQ++;    break;
        case ICMP6_TYPE_ECHO_REPLY:     ICMP6_cfg_ptr->STATS.ST_RX_ECHO_REPLY++;  break;
        default:                   
            ICMP6_cfg_ptr->STATS.ST_RX_OTHER++;       
            ICMP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++;
            break;
    } 
#endif

    if (rtcs_pcb) 
    /* Free pcb.*/
    {
        RTCSLOG_PCB_FREE(rtcs_pcb, result);
        RTCSPCB_free(rtcs_pcb);
    }
}

/************************************************************************
* NAME: ICMP6_send_error
* RETURN: None.
* DESCRIPTION: Send a ICMP error with the IP header that produced it.
*   Warning         :
*     This packet assumes that the IP header is correct,
*     i.e. IP must not use this function on packets recorded
*     in ST_RX_HDR_ERRORS.
*************************************************************************/
void ICMP6_send_error ( uint_8 type     /* [IN] the ICMPv6 type to send */,    
                        uint_8 code     /* [IN] the ICMPv6 code to send */,
                        uint_32 param   /* [IN] error parameter */,
                        RTCSPCB_PTR orig_pcb   /* [IN] the packet which caused the error */)
{ 
    IP6_HEADER_PTR          ip_head = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(orig_pcb);
    ICMP6_CFG_STRUCT_PTR    ICMP6_cfg_ptr;
    RTCSPCB_PTR             pcb;
    ICMP6_ERR_HEADER_PTR    icmp_head;
    in6_addr                *ip_src;
    in6_addr                *ip_dest;
    uint_16                 ippkt_len;
    uint_32                 error;
    IP_IF_PTR               if_ptr = NULL;
    uint_32                 max_len;
   
    ICMP6_cfg_ptr = RTCS_getcfg(ICMP6); 
    
    /************************************************
     * Check input parameters.
     ************************************************/
    if(  (type > 127)
       ||(orig_pcb == NULL))
    {
        return;
    } 

    ip_head = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(orig_pcb);

    ip_src = (in6_addr *)ip_head->SOURCE;
    ip_dest = (in6_addr *)ip_head->DEST;

    ippkt_len = ntohs(ip_head->LENGTH);
    if_ptr = orig_pcb->IFSRC;


    max_len = IP6_DEFAULT_MTU - sizeof(IP6_HEADER) - sizeof(ICMP6_HEADER) - 4;
      
    /* RFC: Every ICMPv6 error message (type < 128) MUST include as much of
     * the IPv6 offending (invoking) packet (the packet that caused the
     * error) as possible without making the error message packet exceed
     * the minimum IPv6 MTU [IPv6].
     */       
    
    if (ippkt_len + IP6_HEADSIZE > max_len)
    {
        ippkt_len = max_len - IP6_HEADSIZE;
    } 
    
    if(ippkt_len > RTCSPCB_SIZE(orig_pcb) - IP6_HEADSIZE)
    {
        ippkt_len = RTCSPCB_SIZE(orig_pcb) - IP6_HEADSIZE;
    }
    
    //PFI put to one big IF.
    /*******************************************************************
     * RFC 4443:
     * (e) An ICMPv6 error message MUST NOT be originated as a result of
     * receiving the following:
     *******************************************************************/
    /* (e.1) An ICMPv6 error message. */ 
    /* (e.2) An ICMPv6 REDIRECT message [IPv6-DISC].*/
    if (ntohc(ip_head->NEXT_HEADER) == IPPROTO_ICMPV6) //TBD Extension header case.
    {
        /* Make sure the packet has at least a 'TYPE' field */
        if (ippkt_len == 0)
        {
            return;
        }
        icmp_head = (ICMP6_ERR_HEADER_PTR)((uchar_ptr)ip_head + IP6_HEADSIZE);
        if (ICMP6_TYPE_IS_ERROR(ntohc(icmp_head->HEAD.TYPE)) || (icmp_head->HEAD.TYPE[0] == ICMP6_TYPE_REDIRECT ) )
        {
            return;
        }
    }
    /*
     * (e.3) A packet destined to an IPv6 multicast address. (There are
     * two exceptions to this rule: (1) the Packet Too Big Message
     * (Section 3.2) to allow Path MTU discovery to work for IPv6
     * multicast, and (2) the Parameter Problem Message, Code 2
     * (Section 3.4) reporting an unrecognized IPv6 option (see
     * Section 4.2 of [IPv6]) that has the Option Type highestorder
     * two bits set to 10).
     * (e.4) A packet sent as a link-layer multicast (the exceptions
     * from e.3 apply to this case, too).     
     */
    if((IN6_IS_ADDR_MULTICAST(ip_dest)||(orig_pcb->TYPE & RTCSPCB_TYPE_MULTICAST)) 
        && !( (type == ICMP6_TYPE_PACKET_TOOBIG) 
            || ((type == ICMP6_TYPE_PARAM_PROB) && (code == ICMP6_CODE_PP_OPTION))) )
    {
        return;
    }
    else
    {
        if(IN6_IS_ADDR_MULTICAST(ip_dest))
            /* We may not use multicast address as source. Get real source address. */
            ip_dest = (in6_addr *)ip6_if_select_src_addr(if_ptr, ip_src /*dest*/); 
     
    }
     
    /*
     * (e.5) A packet sent as a link-layer broadcast (the exceptions
     *  from e.3 apply to this case, too).
     */
    if (orig_pcb->TYPE & RTCSPCB_TYPE_BROADCAST)
    {
        return;
    } 
    /*
     * (e.6) A packet whose source address does not uniquely identify a
     * single node -- e.g., the IPv6 Unspecified Address, an IPv6
     * multicast address, or an address known by the ICMP message
     * originator to be an IPv6 anycast address.
     */
    if(IN6_IS_ADDR_MULTICAST(ip_src) || IN6_ARE_ADDR_EQUAL(ip_src, &in6addr_any))
    {
        return;
    }

    IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_TX_TOTAL++);

    /* Allocate a PCB */
    pcb = RTCSPCB_alloc_send();
    if (pcb == NULL)
    {
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
        return;
    }


    /* Add a dependency and a pointer to the ICMP data */
    RTCSPCB_depend(pcb, orig_pcb);
    error = RTCSPCB_append_fragment(pcb, IP6_HEADSIZE + ippkt_len, (uchar_ptr)ip_head);

    if (!error)
    {
        error = RTCSPCB_insert_header(pcb, sizeof(ICMP6_HEADER) + 4);
    }

    if (error)
    {
        IF_ICMP6_STATS_ENABLED(ICMP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
        IF_ICMP6_STATS_ENABLED(RTCS_seterror(&ICMP6_cfg_ptr->STATS.ERR_TX, error, (uint_32)pcb));
        RTCSLOG_PCB_FREE(pcb, error);
        RTCSPCB_free(pcb);
        return;
    }

    RTCSLOG_PCB_WRITE(pcb, RTCS_LOGCTRL_PROTO(IPPROTO_ICMPV6), 0);

    /* Build the header */
    icmp_head = (ICMP6_ERR_HEADER_PTR)RTCSPCB_DATA(pcb);
    htonc(icmp_head->HEAD.TYPE,     type);
    htonc(icmp_head->HEAD.CODE,     code);
    htons(icmp_head->HEAD.CHECKSUM, 0);
    htonl(icmp_head->DATA,          param);

    pcb->IP_SUM     = IP_Sum_PCB(RTCSPCB_SIZE(pcb), pcb);
    pcb->IP_SUM_PTR = icmp_head->HEAD.CHECKSUM;

#if RTCSCFG_ENABLE_ICMP6_STATS
   /* Update the statistics */
   switch (type)
   {
        case ICMP6_TYPE_DEST_UNREACH:   ICMP6_cfg_ptr->STATS.ST_TX_DESTUNREACH++;break;
        case ICMP6_TYPE_TIME_EXCEED:    ICMP6_cfg_ptr->STATS.ST_TX_TIMEEXCEED++; break;
        case ICMP6_TYPE_PARAM_PROB:     ICMP6_cfg_ptr->STATS.ST_TX_PARMPROB++;  break; 
        case ICMP6_TYPE_REDIRECT:       ICMP6_cfg_ptr->STATS.ST_TX_REDIRECT++;  break;           
        case ICMP6_TYPE_PACKET_TOOBIG:  
        default:                   
            ICMP6_cfg_ptr->STATS.ST_TX_OTHER++;
            break;
   } 
#endif

   /* Send it */
   IP6_send(pcb, IPPROTO_ICMPV6, ip_dest/*ipsrc*/, ip_src/*ipdest*/, if_ptr, 0);
}

#endif /* RTCSCFG_ENABLE_IP6 */



/* EOF */
