/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: ip6.c$
* $Version : 3.8.3.0$
* $Date    : Sep-19-2012$
*
* Comments:
*
*   This file contains the implementation of the Internet
*   Protocolv6.
*
*   2011.09: Added IPv6 support by Andrey Butok
*
*END************************************************************************/

#include <rtcs.h>

#if RTCSCFG_ENABLE_IP6

#include "rtcs_prv.h"
#include "tcpip.h"
#include "ip_prv.h"
#include "ip6_trace.h"


/******************************************************************
* Definitions of some costant IP6 addresses (BSD-like).
*******************************************************************/ 
const struct in6_addr in6addr_any = IN6ADDR_ANY_INIT;
const struct in6_addr in6addr_loopback = IN6ADDR_LOOPBACK_INIT;
const struct in6_addr in6addr_nodelocal_allnodes =
	IN6ADDR_NODELOCAL_ALLNODES_INIT;
const struct in6_addr in6addr_linklocal_allnodes =
	IN6ADDR_LINKLOCAL_ALLNODES_INIT;
const struct in6_addr in6addr_linklocal_allrouters =
	IN6ADDR_LINKLOCAL_ALLROUTERS_INIT;
const struct in6_addr in6addr_linklocal_allv2routers =
	IN6ADDR_LINKLOCAL_ALLV2ROUTERS_INIT;

typedef enum{
    IP6_EXT_HEADER_HANDLER_RESULT_EXIT,      /* Stop processing.*/ 
    IP6_EXT_HEADER_HANDLER_RESULT_NEXT,      /* Go to the next Ext. Header processing.*/
    IP6_EXT_HEADER_HANDLER_RESULT_TRANSPORT  /* Go to Transport Layer processing.*/
} ip6_ext_header_handler_result_t;


/******************************************************************
* Function Prototypes
*******************************************************************/
static ip6_ext_header_handler_result_t ip6_ext_header_handler_hop_by_hop_options(RTCSPCB_PTR *rtcs_pcb_p);
static ip6_ext_header_handler_result_t ip6_ext_header_handler_destination_options(RTCSPCB_PTR *rtcs_pcb_p);
static ip6_ext_header_handler_result_t ip6_ext_header_handler_routing_header(RTCSPCB_PTR *rtcs_pcb_p);
static ip6_ext_header_handler_result_t ip6_ext_header_handler_fragment_header(RTCSPCB_PTR *rtcs_pcb_p);


/******************************************************************
* Extension Header Handler structure
*******************************************************************/
typedef struct ip6_ext_header
{
    unsigned long type;                                                     /* Identifies the type of header. */
    ip6_ext_header_handler_result_t (*handler)(RTCSPCB_PTR    *rtcs_pcb_p); /* Extension header handler.*/
    
} ip6_ext_header_t;


/******************************************************************
* Extension Header Handler List
*******************************************************************/
ip6_ext_header_t ip6_ext_header_list[] = 
{
    {IP6_TYPE_HOP_BY_HOP_OPTIONS,   ip6_ext_header_handler_hop_by_hop_options},
    {IP6_TYPE_DESTINATION_OPTIONS,  ip6_ext_header_handler_destination_options},
    {IP6_TYPE_ROUTING_HEADER,       ip6_ext_header_handler_routing_header},
    {IP6_TYPE_FRAGMENT_HEADER,      ip6_ext_header_handler_fragment_header}
    
    /* ADD HERE YOUR EXTENSION HEADER HANDLERS.*/
};

#define IP6_EXT_HEADER_LIST_SIZE    (sizeof(ip6_ext_header_list)/sizeof(ip6_ext_header_t))

/************************************************************************
* Function Prototypes
*************************************************************************/
static uint_32 ip6_process_options ( RTCSPCB_PTR  pcb );


/************************************************************************
* NAME: IP6_init
*
* DESCRIPTION: Initialize the IPv6 layer.
*
* RETURNS: RTCS_OK, or error code .
*************************************************************************/
uint_32 IP6_init(void)
{ 
    IP6_CFG_STRUCT_PTR IP6_cfg_ptr;
        
    IP6_cfg_ptr = RTCS_mem_alloc_zero(sizeof(IP6_CFG_STRUCT));

    if (IP6_cfg_ptr == NULL)
    {
        return RTCSERR_OUT_OF_MEMORY;
    }

    _mem_set_type(IP6_cfg_ptr, MEM_TYPE_IP6_CFG_STRUCT);
   
    RTCS_setcfg(IP6, IP6_cfg_ptr); 
        
    IP6_cfg_ptr->DEFAULT_HOPLIMIT = IP6_DEFAULT_HOPLIMIT;       


#if RTCSCFG_ENABLE_IP6_REASSEMBLY
   /* Initialize the reassembler */
   IP6_reasm_init();
#endif /* RTCSCFG_ENABLE_IP6_REASSEMBLY */

   return RTCS_OK;
}

/************************************************************************
* NAME: ip6_process_options
*
* DESCRIPTION: Process Hop by Hop Options and Destimation Options
*
* RETURNS: RTCS_OK, or error code => DROP packet.
*************************************************************************/
static uint_32 ip6_process_options ( RTCSPCB_PTR  pcb )
{
    IP6_HEADER_PTR           ip6_h = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(pcb);
    IP6_OPTIONS_HEADER_PTR   options_h    = (IP6_OPTIONS_HEADER_PTR)RTCSPCB_DATA(pcb);
    in6_addr *               ip_dst;
    IP6_OPTION_HEADER_PTR    option;
   
    uint_32                  offset;
    uint_32                  length = (ntohc(options_h->HDR_EXT_LENGTH)<<3) + (8-2);
   
    for(offset = 0; offset < length; )
    {
        option = (IP6_OPTION_HEADER_PTR)&options_h->OPTIONS[offset];
        
        /* The RFC2460 supports only PAD0 and PADN options.*/
        switch(ntohc(option->TYPE))
        {        
            /* Pad1 option */
            case IP6_OPTION_TYPE_PAD1:
                offset++;
                break;
            /* Pad1 option */
            case IP6_OPTION_TYPE_PADN:
                offset += sizeof(IP6_OPTION_HEADER) + ntohc(option->DATA_LENGTH);
                break;
            /* Unrecognized Options.*/
            default:     
                /* The Option Type identifiers are internally encoded such that their
                 * highest-order two bits specify the action that must be taken if the
                 * processing IPv6 node does not recognize the Option Type.*/
                switch(ntohc(option->TYPE) & IP6_OPTION_TYPE_UNRECOGNIZED_MASK)
                {
                    /* 00 - skip over this option and continue processing the header.*/
                    case IP6_OPTION_TYPE_UNRECOGNIZED_SKIP:      
                        break;
                    /* 01 - discard the packet. */    
                    case IP6_OPTION_TYPE_UNRECOGNIZED_DISCARD:   
                        return (uint_32)RTCS_ERROR;
                    /* 10 - discard the packet and, regardless of whether or not the
                     *      packets Destination Address was a multicast address, send an
                     *      ICMP Parameter Problem, Code 2, message to the packets
                     *      Source Address, pointing to the unrecognized Option Type.*/    
                    case IP6_OPTION_TYPE_UNRECOGNIZED_DISCARD_ICMP:
                        ICMP6_send_error ( ICMP6_TYPE_PARAM_PROB,    
                                           ICMP6_CODE_PP_OPTION,
                                           (uint_32)option->TYPE - (uint_32)ip6_h ,
                                           pcb );
                        return (uint_32)RTCS_ERROR; 
                    /* 11 - discard the packet and, only if the packets Destination
                     *      Address was not a multicast address, send an ICMP Parameter
                     *      Problem, Code 2, message to the packets Source Address,
                     *      pointing to the unrecognized Option Type.*/                                                 
                    case IP6_OPTION_TYPE_UNRECOGNIZED_DISCARD_UICMP:
                        ip_dst  = (in6_addr *)(ip6_h->DEST);
                        if(!IN6_IS_ADDR_MULTICAST(ip_dst))
                            ICMP6_send_error ( ICMP6_TYPE_PARAM_PROB,    
                                               ICMP6_CODE_PP_OPTION,
                                               (uint_32)option->TYPE - (uint_32)ip6_h ,
                                               pcb );
                        return (uint_32)RTCS_ERROR;
                }
                
                offset += sizeof(IP6_OPTION_HEADER) + ntohc(option->DATA_LENGTH);
                break;
        }
    }
    
    /* Set up the transport layer protocol and delta. */
    RTCSPCB_SET_TRANS_PROTL(pcb, ntohc(options_h->NEXT_HEADER));
    RTCSPCB_SET_TRANS_DELTA(pcb, length+2);
    
    return RTCS_OK;
}

/************************************************************************
* NAME: ip6_ext_header_handler_hop_by_hop_options
* 
* DESCRIPTION: 
*
* RETURNS:
*************************************************************************/
static ip6_ext_header_handler_result_t ip6_ext_header_handler_hop_by_hop_options(RTCSPCB_PTR    *rtcs_pcb_p)
{
    IP6_CFG_STRUCT_PTR		IP6_cfg_ptr = RTCS_getcfg(IP6);
    
    /* RFC 2460 4: The Hop-by-Hop Options header,
     * when present, must immediately follow the IPv6 header. Its presence
     * is indicated by the value zero in the Next Header field of the IPv6
     * header.*/
    uint_32     error;
    uchar       hdrlen;
    RTCSPCB_PTR rtcs_pcb =   *rtcs_pcb_p;
   
    hdrlen = RTCSPCB_TRANSPORT_DELTA(rtcs_pcb);
    error = RTCSPCB_next(rtcs_pcb, hdrlen);
    if (error)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++); //PFI
        IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
        RTCSLOG_PCB_FREE(rtcs_pcb, error);
        RTCSPCB_free(rtcs_pcb);
                    
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
        
    if(ip6_process_options ( rtcs_pcb ) == (uint_32)RTCS_ERROR)
    {
        /* DROP it */ //PFI
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCS_OK);
        RTCSPCB_free(rtcs_pcb);
          
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, RTCS_OK);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }

    return IP6_EXT_HEADER_HANDLER_RESULT_NEXT;
}

/************************************************************************
* NAME: ip6_ext_header_destination_options
* 
* DESCRIPTION: 
*
* RETURNS:
*************************************************************************/
static ip6_ext_header_handler_result_t ip6_ext_header_handler_destination_options(RTCSPCB_PTR    *rtcs_pcb_p)
{
    uint_32     error;
    uchar       hdrlen;
    RTCSPCB_PTR rtcs_pcb =   *rtcs_pcb_p;
    IP6_CFG_STRUCT_PTR		IP6_cfg_ptr = RTCS_getcfg(IP6);
    
    hdrlen = RTCSPCB_TRANSPORT_DELTA(rtcs_pcb); //PFI
    error = RTCSPCB_next(rtcs_pcb, hdrlen);
                
    if (error)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++); //PFI
        IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
        RTCSLOG_PCB_FREE(rtcs_pcb, error);
        RTCSPCB_free(rtcs_pcb);
                            
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
                
    if(ip6_process_options ( rtcs_pcb ) == (uint_32)RTCS_ERROR)
    {
        /* DROP it */ //PFI
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCS_OK);
        RTCSPCB_free(rtcs_pcb);
                  
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, RTCS_OK);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
                
   return IP6_EXT_HEADER_HANDLER_RESULT_NEXT;             
}

/************************************************************************
* NAME: ip6_ext_header_routing_header
* 
* DESCRIPTION: 
*
* RETURNS:
*************************************************************************/
static ip6_ext_header_handler_result_t ip6_ext_header_handler_routing_header(RTCSPCB_PTR    *rtcs_pcb_p)
{
    uint_32     error;
    uchar       hdrlen;
    RTCSPCB_PTR rtcs_pcb =   *rtcs_pcb_p;
    IP6_ROUTING_HEADER_PTR   routing_h;
    IP6_CFG_STRUCT_PTR		IP6_cfg_ptr = RTCS_getcfg(IP6);
                    
    hdrlen = RTCSPCB_TRANSPORT_DELTA(rtcs_pcb);
    error = RTCSPCB_next(rtcs_pcb, hdrlen);
    if (error)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++); //PFI
        IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
        RTCSLOG_PCB_FREE(rtcs_pcb, error);
        RTCSPCB_free(rtcs_pcb);
                                
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
                    
    routing_h    = (IP6_ROUTING_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
                    
    /* RFC 2460 4.4: If, while processing a received packet, a node encounters a Routing
    * header with an unrecognized Routing Type value, the required behavior
    * of the node depends on the value of the Segments Left field, as
    * follows:
    * -If Segments Left is non-zero, the node must discard the packet and
    *  send an ICMP Parameter Problem, Code 0, message to the packets
    *  Source Address, pointing to the unrecognized Routing Type.
    */
    if( ntohc(routing_h->SEGMENTS_LEFT) > 0)
    {
        ICMP6_send_error ( ICMP6_TYPE_PARAM_PROB,    
                            ICMP6_CODE_PP_HEADER,
                            (uint_32)routing_h->ROUTING_TYPE - (uint_32)RTCSPCB_DATA_NETWORK(rtcs_pcb) ,
                            rtcs_pcb);
                        
        /* Discard the packet.*/
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCS_OK); //PFI
        RTCSPCB_free(rtcs_pcb);
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, RTCS_OK);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
    /* -If Segments Left is zero, the node must ignore the Routing header
     *  and proceed to process the next header in the packet, whose type
     *  is identified by the Next Header field in the Routing header.*/ 
                    
    /* Set up the transport layer protocol and delta. */
    RTCSPCB_SET_TRANS_PROTL(rtcs_pcb, ntohc(routing_h->NEXT_HEADER));
    RTCSPCB_SET_TRANS_DELTA(rtcs_pcb, (ntohc(routing_h->HDR_EXT_LENGTH)<<3) + (8));    

    return IP6_EXT_HEADER_HANDLER_RESULT_NEXT;             
}

/************************************************************************
* NAME: ip6_ext_header_handler_fragment_header
* 
* DESCRIPTION: Process IPv6 Fragment header.
* 
* RETURNS:
*************************************************************************/
static ip6_ext_header_handler_result_t ip6_ext_header_handler_fragment_header(RTCSPCB_PTR    *rtcs_pcb_p)
{
    uint_32     error;
    uchar       hdrlen;
    RTCSPCB_PTR rtcs_pcb =   *rtcs_pcb_p;
    IP6_CFG_STRUCT_PTR		IP6_cfg_ptr = RTCS_getcfg(IP6);
    
#if RTCSCFG_ENABLE_IP6_REASSEMBLY
                
    hdrlen = RTCSPCB_TRANSPORT_DELTA(rtcs_pcb);
    error = RTCSPCB_next(rtcs_pcb, hdrlen);
    if (error)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++); //PFI
        IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
        RTCSLOG_PCB_FREE(rtcs_pcb, error);
        RTCSPCB_free(rtcs_pcb);
                    
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
                
    error = IP6_reasm(rtcs_pcb, rtcs_pcb_p);
    IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_FRAG_RECVD++);
    if (*rtcs_pcb_p)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_FRAG_REASMD++);
        return IP6_EXT_HEADER_HANDLER_RESULT_TRANSPORT;
    } 
    else
    {
        if (error)
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_FRAG_DISCARDED++);
        }
                    
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
        return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
    }
#else   /* IPv6 REASSEMBLY is disabled */
    IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_FRAG_DISCARDED++);
    RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_FEATURE_NOT_ENABLED);
    RTCSPCB_free(rtcs_pcb);
    return IP6_EXT_HEADER_HANDLER_RESULT_EXIT;
#endif /* RTCSCFG_ENABLE_IP6_REASSEMBLY */             

}

/************************************************************************
* NAME: ip6_ext_header_process
* 
* DESCRIPTION: Process IPv6 Extension Headers.
* 
* RETURNS: RTCS_OK - continue processing by transport layer.
*          RTCS_ERROR - stop processing.
*************************************************************************/
static int ip6_ext_header_process(RTCSPCB_PTR *rtcs_pcb_p)
{
    int     try_upper_layer;
    uint_8  next_header;
    
    /* RFC 2460 4:
     * Therefore, extension headers must
     * be processed strictly in the order they appear in the packet; a
     * receiver must not, for example, scan through a packet looking for a
     * particular kind of extension header and process that header prior to
     * processing all preceding ones.
     */
    
    /* Process headers.*/
    try_upper_layer = FALSE;
    while(try_upper_layer == FALSE)
    {
        ip6_ext_header_t *ext_header;
        int j;
        next_header = RTCSPCB_TRANSPORT_PROTL(*rtcs_pcb_p);
        
        ext_header = NULL;
        
        for(j = 0; j<IP6_EXT_HEADER_LIST_SIZE; j++)
        {
            if(next_header == ip6_ext_header_list[j].type)
            {
                ext_header = &ip6_ext_header_list[j];
                break;
            }
        }
        
        if(ext_header)
        {
            ip6_ext_header_handler_result_t handler_result = ext_header->handler(rtcs_pcb_p);
                    
            switch(handler_result)
            {
                case IP6_EXT_HEADER_HANDLER_RESULT_EXIT:
                    return RTCS_ERROR;
                case IP6_EXT_HEADER_HANDLER_RESULT_NEXT:
                    break;
                case IP6_EXT_HEADER_HANDLER_RESULT_TRANSPORT:
                    try_upper_layer = TRUE;
                    break;
            };  
        }
        else
        {
            try_upper_layer = TRUE;
        }
    }  
    
    return RTCS_OK;          
}

/************************************************************************
* NAME: IP6_service
* 
* DESCRIPTION: Parses a received IP6 packet.
*
* RETURNS: RTCS_OK or error code
*************************************************************************/
void IP6_service( RTCSPCB_PTR    rtcs_pcb   /* [IN] the packet to send */)
{ 
	IP6_CFG_STRUCT_PTR		IP6_cfg_ptr;
	IP6_HEADER_PTR			packet = (IP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    uint_16                	pktlen;
    uint_32                 error;
    uchar                   hdrlen; //PFI
    
    RTCSLOG_FNE2(RTCSLOG_FN_IP6_service, rtcs_pcb);
    
    IP6_cfg_ptr = RTCS_getcfg(IP6);

	IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_TOTAL++);
        
    RTCSPCB_DATA_NETWORK(rtcs_pcb) = RTCSPCB_DATA(rtcs_pcb);
    
    /*******************************************************************
     * Start IPv6 header  processing.
     *******************************************************************/
    
    /* Calculate & Check Size of the packet.*/
    pktlen = IP6_HEADER_LEN + ntohs(packet->LENGTH);
    
    if (RTCSPCB_SIZE(rtcs_pcb) < pktlen)
    {
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_HDR_ERRORS++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_SMALL_PKT++);
		RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_BAD_HEADER);
		RTCSPCB_free(rtcs_pcb);
		RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, 3);
      	return;
	}
	
   	/*
   	 * IP version must be 6.
   	 */
   	if ((ntohc(packet->VERSION_TCLASS) & IP6_HEADER_VERSION_MASK) != IP6_HEADER_VERSION)
   	{
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_HDR_ERRORS++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_BAD_VERSION++);
		RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_BAD_HEADER);
		RTCSPCB_free(rtcs_pcb);
		RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, 4);
		return;
	}
	 
	/*
	 * Validate source address.
	 */
	if (IN6_IS_ADDR_MULTICAST((in6_addr*)packet->SOURCE))
    {
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_HDR_ERRORS++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_BAD_SOURCE++);
		RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_BAD_ADDRESS);
		RTCSPCB_free(rtcs_pcb);
		RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, 5);
		return;
	}     
    
	/*
	 * Validate destination address.
	 */
	if( !ip6_if_is_my_addr((_rtcs_if_handle)rtcs_pcb->IFSRC, (in6_addr*)packet->DEST) &&
   		!ip6_if_is_my_solicited_multicast_addr((_rtcs_if_handle)rtcs_pcb->IFSRC, (in6_addr*)packet->DEST) &&
   		!IN6_ARE_ADDR_EQUAL(&in6addr_linklocal_allnodes, (in6_addr*)packet->DEST)
   		//TBD All routers
	)
	{
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_ADDR_ERRORS++);
		RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_BAD_ADDRESS);
		RTCSPCB_free(rtcs_pcb);
		RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, 6);
		return;
	}
	//TBD ROUTER, FARWARD PACKETS
	
	
	/*
	** In this function, the PCB always has only one fragment, so it's
	** always safe to call RTCSPCB_shrink() here.
	*/
	error = RTCSPCB_shrink(rtcs_pcb, RTCSPCB_SIZE(rtcs_pcb) - pktlen);
	if (error)
	{
		IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++);
		IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
		RTCSLOG_PCB_FREE(rtcs_pcb, error);
		RTCSPCB_free(rtcs_pcb);
		RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, 8);
		return;
    } 

    trace_ip6("RX", packet); /* Trace IP6 RX Header.*/
	
	
    /* Set up the transport layer protocol and delta. */
    RTCSPCB_SET_TRANS_PROTL(rtcs_pcb, ntohc(packet->NEXT_HEADER));
    RTCSPCB_SET_TRANS_DELTA(rtcs_pcb, sizeof(IP6_HEADER));
	
	/********************************************
	* Extension headers processing
    *********************************************/
    if(ip6_ext_header_process(&rtcs_pcb) == RTCS_ERROR)
    {
        return;
    }
    
    
    /* Note: (http://www.cisco.com/web/about/ac123/ac147/archived_issues/ipj_9-3/ipv6_internals.html)
     * Note that there is no standard extension header format, meaning that when a host 
     * encounters a header that it does not recognize in the protocol chain, the only thing 
     * it can do is discard the packet. Worse, firewalls and routers configured to filter IPv6 
     * have the same problem: as soon as they encounter an unknown extension header, 
     * they must decide to allow or disallow the packet, even though another header deeper 
     * inside the packet would possibly trigger the opposite behavior. In other words, an IPv6 
     * packet with a TCP payload that would normally be allowed through could be blocked if 
     * there is an unknown extension header between the IPv6 and TCP headers.
     */
    
    /********************************************
	* Upper Layer or 
	* Transport layer processing (ICMP/TCP/UDP)
    *********************************************/
    { 
        ICB_STRUCT_PTR       search_ptr;
        uchar                protocol;
        uint_32              error;
       
        /* Find an ICB for the received packet's protocol */
        protocol = RTCSPCB_TRANSPORT_PROTL(rtcs_pcb);
        RTCS_LIST_SEARCH(IP6_cfg_ptr->ICB6_HEAD, search_ptr)
        {
            if (search_ptr->PROTOCOL == protocol)
            {
                break;
            }
        } 

        if (!search_ptr)
        { /* No protocol found.*/
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_DISCARDED++);
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_NO_PROTO++);
            
            /* RFC 2460 4:If, as a result of processing a header, a node is required to proceed
             * to the next header but the Next Header value in the current header is
             * unrecognized by the node, it should discard the packet and send an
             * ICMP Parameter Problem message to the source of the packet, with an
             * ICMP Code value of 1 ("unrecognized Next Header type encountered")
             * and the ICMP Pointer field containing the offset of the unrecognized
             * value within the original packet. The same action should be taken if
             * a node encounters a Next Header value of zero in any header other
             * than an IPv6 header.*/
            ICMP6_send_error ( ICMP6_TYPE_PARAM_PROB,    
                                           ICMP6_CODE_PP_NEXT_HEADER,
                                           (uint_32)RTCSPCB_DATA(rtcs_pcb) - (uint_32)RTCSPCB_DATA_NETWORK(rtcs_pcb) ,
                                           rtcs_pcb ); //TBD - not tested. Ponter to next header...
            
            RTCSLOG_PCB_FREE(rtcs_pcb, RTCS_OK);
            RTCSPCB_free(rtcs_pcb);
          
            RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, RTCS_OK);
            return;
        } 

        hdrlen = RTCSPCB_TRANSPORT_DELTA(rtcs_pcb);
        error = RTCSPCB_next(rtcs_pcb, hdrlen);
        if (error)
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_RX_ERRORS++);
            IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_RX, error, (uint_32)rtcs_pcb));
            RTCSLOG_PCB_FREE(rtcs_pcb, error);
            RTCSPCB_free(rtcs_pcb);
            
            RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, error);
            return;
        }

        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_DELIVERED++);
        search_ptr->SERVICE(rtcs_pcb, search_ptr->PRIVATE);
    } 
   
    RTCSLOG_FNX2(RTCSLOG_FN_IP6_service, RTCS_OK);

} 


/************************************************************************
* NAME: IP6_complete_send
* PARAMS:
*      pointer           ifdest      // [IN] the destination interface.
*      RTCSPCB_PTR _PTR_ rtcs_pcb_ptr_ptr // [IN] the packet to send.
*      _ip_address       hopsrc      // [IN] the destination interface.
*       _ip_address      ipdest      // [IN] the ultimate dest.
*      uint_32           protocol    // [IN] the transport layer protocol
* DESCRIPTION: Generates an IP header for outgoing packets.  This function
*           must not consume pcb.
*************************************************************************/
static uint_32 IP6_complete_send ( IP_IF_PTR ifdest, RTCSPCB_PTR _PTR_ rtcs_pcb_ptr_ptr, in6_addr *hopsrc, in6_addr *ipdest, uint_32 protocol )
{
    RTCSPCB_PTR            rtcs_pcb = *rtcs_pcb_ptr_ptr;
    IP6_CFG_STRUCT_PTR     IP6_cfg_ptr;
    IP6_HEADER_PTR         packet;
    uchar                  proto;
    uchar                  ttl;
    uint_32                error;
    uint_16                chksum;

    IP6_cfg_ptr = RTCS_getcfg(IP6);

    error = RTCSPCB_insert_header(rtcs_pcb, sizeof(IP6_HEADER));
    if (error)
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
        IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_TX, error, (uint_32)rtcs_pcb));
        /* The PCB will be freed for us. */
        return error;
    }

    RTCSLOG_PCB_WRITE(rtcs_pcb, RTCS_LOGCTRL_PROTO(IPPROTO_IP), IP_VERSION_6);
    packet = (IP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    RTCSPCB_DATA_NETWORK(rtcs_pcb) = RTCSPCB_DATA(rtcs_pcb);
   
    proto  = IPPROTO_GET(protocol);

    ttl = rtcs_pcb->LINK_OPTIONS.TX.TTL;
    if (!ttl) 
        ttl = IPTTL_GET(protocol);
    if (!ttl)
        ttl = ifdest->IP6_IF.ND6->cur_hop_limit; 
    
    /* Fill the header.*/
    htonc(packet->VERSION_TCLASS,  (IP_VERSION_6 << 4) );
    htonc(packet->TCLASS_FLOWL,      0);
    htons(packet->FLOWL,   0x0000);
   
    htons(packet->LENGTH,   RTCSPCB_SIZE(rtcs_pcb) - sizeof(IP6_HEADER));
    htonc(packet->NEXT_HEADER, proto);
    htonc(packet->HOP_LIMIT,   ttl);
   
    IN6_ADDR_COPY(hopsrc, (in6_addr*)packet->SOURCE);
    IN6_ADDR_COPY(ipdest, (in6_addr*)packet->DEST);

    /* Set up the transport layer protocol and delta. */
    RTCSPCB_SET_TRANS_PROTL(rtcs_pcb, proto);
    RTCSPCB_SET_TRANS_DELTA(rtcs_pcb, sizeof(IP_HEADER));

    /*
    ** If the transport layer needs a pseudo header checksum,
    ** calculate it now.
    */
    if (rtcs_pcb->IP_SUM_PTR)
    {
        chksum = IP6_Sum_pseudo(rtcs_pcb->IP_SUM, rtcs_pcb);
        chksum = IP_Sum_invert(chksum);
        htons(rtcs_pcb->IP_SUM_PTR, chksum);
    }
    
    return RTCS_OK;
}

/************************************************************************
* NAME: IP6_send_dgram
* 
* DESCRIPTION: Sends an IP datagram out an interface, fragmenting it if
*        necessary.
*************************************************************************/
uint_32 IP6_send_dgram ( IP_IF_PTR ifdest,          /* [IN] the outgoing interface */
                         RTCSPCB_PTR inrtcs_pcb,    /* [IN] the packet to send */
                         in6_addr *hopsrc,          /* [IN] the hop src */
                         in6_addr *hopdest,         /* [IN] the hop dest */
                         in6_addr *ipdest,          /* [IN] the ultimate dest */
                         uint_32 protocol           /* [IN] the transport layer protocol */)
{ 
    IP6_CFG_STRUCT_PTR IP6_cfg_ptr = RTCS_getcfg(IP6);
    uint_32 error;
   
    RTCSLOG_FNE2(RTCSLOG_FN_IP_send_dgram, ifdest);
    
    error = IP6_complete_send(ifdest, &inrtcs_pcb, hopsrc, ipdest, protocol); //TBD
    if (error)
    { 
        RTCSLOG_PCB_FREE(inrtcs_pcb, error);
        RTCSPCB_free(inrtcs_pcb);
      
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, error);
        return error;
    } 

    /* If the packet doesn't exceed the MTU, send it directly */
    if (RTCSPCB_SIZE(inrtcs_pcb) <= (ifdest->IP6_IF.ND6 ? ifdest->IP6_IF.ND6->mtu : ifdest->MTU))
    {
        /*
         * Send via interface.
         */
        error = ifdest->DEVICE.SEND6(ifdest, inrtcs_pcb, hopsrc, hopdest);
      
        RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, error);
        return error;
    } 
    else
    /* Packet exceeds the MTU. */
    {
        uint_32         dmtumax;
        uint_32         dmtu;
        IP6_HEADER_PTR  srciph;
        uint_32         srcdatalen;
        uint_32         srclen;
        uint_32         offset;
        uchar_ptr       srcdata;
        uint_32         iphlen;
        PCB_FRAGMENT_PTR    frag_ptr;
        uchar           next_header;
        uint_32         frag_id;
   
   
        /*
         * Fragmentation is used in IPv6 when the IPv6 packet will not fit
         * inside the path MTU to its destination.  When fragmentation is
         * performed, an IPv6 node uses a fragment header, as specified in
         * Section 4.5 of the IPv6 base specification [RFC2460], to break down
         * the datagram into smaller fragments that will fit in the path MTU.
         */
        int dfrag  = IPDFRAG_GET(protocol); /* Check the "Do not fragment flag". Just in case.*/
        
        iphlen = IP6_HEADER_LEN + sizeof(IP6_FRAGMENT_HEADER);
        srciph = (IP6_HEADER_PTR)RTCSPCB_DATA(inrtcs_pcb);
        srcdatalen = RTCSPCB_SIZE(inrtcs_pcb) - IP6_HEADER_LEN;
          
        /* If the datagram exceeds the MTU, but DF is set => drop it.*/
        if (dfrag)
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_TX_FRAG_DISCARDED++);
            RTCSLOG_PCB_FREE(inrtcs_pcb, RTCSERR_IP_CANTFRAG);
            RTCSPCB_free(inrtcs_pcb);
          
            RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, RTCSERR_IP_CANTFRAG);
            return RTCSERR_IP_CANTFRAG;
        }
        
        /* Sanity check.  If small MTU. */
        if (iphlen > ifdest->IP6_IF.ND6->mtu)
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_TX_FRAG_DISCARDED++);
            RTCSLOG_PCB_FREE(inrtcs_pcb, RTCSERR_IP_SMALLMTU);
            RTCSPCB_free(inrtcs_pcb);
          
            RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, RTCSERR_IP_SMALLMTU);
            return RTCSERR_IP_SMALLMTU;
        }
        //TBD other extensions.

        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_TX_FRAG_FRAGD++);
        dmtumax = ifdest->IP6_IF.ND6->mtu - iphlen;
        dmtu = dmtumax & (((uint_32)(-1))<<3);
        offset =  0;
        srcdata = RTCSPCB_DATA(inrtcs_pcb) + IP6_HEADER_LEN;
        srclen = inrtcs_pcb->HEADER_FRAG_USED - IP6_HEADER_LEN;
        frag_ptr = inrtcs_pcb->PCBPTR->FRAG;
        next_header = srciph->NEXT_HEADER[0]; 
        
        /* ID */
        frag_id = IP6_cfg_ptr->NEXT_ID6;
        IP6_cfg_ptr->NEXT_ID6++;
        
        do
        {
            IP6_HEADER_PTR          dstiph;
            IP6_FRAGMENT_HEADER_PTR frag_header;
            uchar_ptr       dstdata;
            uint_32         dstlen = (srcdatalen > dmtumax) ? dmtu : srcdatalen;
            RTCSPCB_PTR     rtcs_pcb;
            uint_16         fragment;

            /*
             * Fragment packets:
             * +------------------+--------+--------------+
             * |  Unfragmentable  |Fragment|    first     |
             * |       Part       | Header |   fragment   |
             * +------------------+--------+--------------+
             * 
             * +------------------+--------+--------------+
             * |       Part       | Header |   fragment   |
             * +------------------+--------+--------------+
             * o
             * o
             * o
             * +------------------+--------+----------+
             * |  Unfragmentable  |Fragment|   last   |
             * |       Part       | Header | fragment |
             * +------------------+--------+----------+
             * Each fragment packet is composed of:
             *     
             * (1) The Unfragmentable Part of the original packet, with the
             *     Payload Length of the original IPv6 header changed to contain
             *     the length of this fragment packet only (excluding the length
             *     of the IPv6 header itself), and the Next Header field of the
             *     last header of the Unfragmentable Part changed to 44.
             * (2) A Fragment header containing:
             *     
             *     The Next Header value that identifies the first header of
             *     the Fragmentable Part of the original packet.
             *     
             *     A Fragment Offset containing the offset of the fragment,
             *     in 8-octet units, relative to the start of the
             *     Fragmentable Part of the original packet.  The Fragment
             *     Offset of the first ("leftmost") fragment is 0.
             *     
             *     An M flag value of 0 if the fragment is the last
             *     ("rightmost") one, else an M flag value of 1.
             *     
             *     The Identification value generated for the original
             *     packet.
             *     
             * (3) The fragment itself.
             */

            
            /* Allocate the PCB */
            rtcs_pcb = RTCSPCB_alloc_send();
            if (rtcs_pcb == NULL)
            {
                IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                RTCSLOG_PCB_FREE(inrtcs_pcb, RTCSERR_PCB_ALLOC);
                RTCSPCB_free(inrtcs_pcb);
         
                RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, RTCSERR_PCB_ALLOC);
                return RTCSERR_PCB_ALLOC;
            }
            
            /* Allocate the memory for the IP data */
            dstdata = RTCS_mem_alloc_system(dstlen);
            if (dstdata == NULL)
            {
                IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_OUT_OF_MEMORY);
                RTCSPCB_free(rtcs_pcb);
                RTCSLOG_PCB_FREE(inrtcs_pcb, RTCSERR_OUT_OF_MEMORY);
                RTCSPCB_free(inrtcs_pcb);
         
                RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, RTCSERR_OUT_OF_MEMORY);
                return RTCSERR_OUT_OF_MEMORY;
            }
            _mem_set_type(dstdata, MEM_TYPE_IP_DATA);
            
            /* Append the data to the PCB */
            error = RTCSPCB_append_fragment_autofree(rtcs_pcb, dstlen, dstdata);
            if (error)
            {
                IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_TX, error, (uint_32)rtcs_pcb));
                _mem_free(dstdata);
                RTCSLOG_PCB_FREE(rtcs_pcb, error);
                RTCSPCB_free(rtcs_pcb);
                RTCSLOG_PCB_FREE(inrtcs_pcb, error);
                RTCSPCB_free(inrtcs_pcb);
         
                RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, error);
                return error;
            }
            
            /* Allocate the IPV6 header and Fragment header */
            error = RTCSPCB_insert_header(rtcs_pcb, iphlen);
            if (error)
            {
                IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_TX, error, (uint_32)rtcs_pcb));
                /* don't _mem_free(dstdata) here because RTCSPCB_free(pcb) will do it */
                RTCSLOG_PCB_FREE(rtcs_pcb, error);
                RTCSPCB_free(rtcs_pcb);
                RTCSLOG_PCB_FREE(inrtcs_pcb, error);
                RTCSPCB_free(inrtcs_pcb);
         
                RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, error);
                return error;
            }
            
            /*************************************
             * Build IP Header 
             *************************************/
             /* Copy the IP header */
            dstiph = (IP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
            RTCSPCB_DATA_NETWORK(rtcs_pcb) = RTCSPCB_DATA(rtcs_pcb);
            _mem_copy(srciph, dstiph, iphlen);
            
            /* Update the LENGTH field */
            htons(dstiph->LENGTH, sizeof(IP6_FRAGMENT_HEADER) + dstlen);     
            
            /* Update next header = Fragment Header.*/
            htonc(dstiph->NEXT_HEADER, IP6_TYPE_FRAGMENT_HEADER);
            
            /*************************************
             * Build the FRAGMENT Header 
             *************************************/
            frag_header = (IP6_FRAGMENT_HEADER_PTR)(RTCSPCB_DATA(rtcs_pcb)+IP6_HEADER_LEN);
            
            fragment = offset;
            offset     += dstlen;
            srcdatalen -= dstlen; 
        
            /* An MF bit must be added on all fragments except the last */
            if (srcdatalen > 0)
            {
                fragment |= IP6_FRAGMENT_MF;
            }
            htons(frag_header->OFFSET_MORE, fragment);
            
            /* Update next header.*/
            htonc(frag_header->NEXT_HEADER, next_header);
            
            /* ID */
            htonl(frag_header->ID, frag_id);

            /*************************************
             * Copy Data 
             *************************************/
            for (;;)
            {
                uint_32 copylen = (srclen > dstlen) ? dstlen : srclen;
                
                _mem_copy(srcdata, dstdata, copylen);
                srcdata += copylen;
                dstdata += copylen;
                srclen  -= copylen;
                dstlen  -= copylen;
                if (dstlen == 0)
                    break;
                frag_ptr++;
                srcdata = frag_ptr->FRAGMENT;
                srclen = frag_ptr->LENGTH;
            }
            
            
            /*
            * Send via interface.
            */
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_TX_FRAG_SENT++);
            error = ifdest->DEVICE.SEND6(ifdest, rtcs_pcb, hopsrc, hopdest);
            if (error)
            {
                RTCSLOG_PCB_FREE(inrtcs_pcb, error);
                RTCSPCB_free(inrtcs_pcb);
         
                RTCSLOG_FNX2(RTCSLOG_FN_IP6_send_dgram, error);
                return error;
            }
        } 
        while (srcdatalen > 0);
        
        RTCSLOG_PCB_FREE(inrtcs_pcb, RTCS_OK);
        RTCSPCB_free(inrtcs_pcb);
   
        RTCSLOG_FNX2(RTCSLOG_FN_IP_send_dgram, RTCS_OK);
        return RTCS_OK;  
    }
} 

/************************************************************************
* NAME: IP6_send
* PARAMS:
*      RTCSPCB_PTR    rtcs_pcb   //[IN] the packet to send.
*      uint_32        protocol   //[IN] the transport layer protocol.
*      in6_addr       *ipsrc     //[IN] the destination interface (0 = any).
*      in6_addr       *ipdest    //[IN] the ultimate destination.
*      uint_32        flags      //[IN] optional flags.
* RETURS: RTCS_OK or error code
* DESCRIPTION: Sends an IP6 packet generated on the local host.
*************************************************************************/
uint_32 IP6_send ( RTCSPCB_PTR  rtcs_pcb        /* [IN] the packet to send */,
                   uint_32 protocol             /* [IN] the transport layer protocol */,
                   in6_addr *ip_src             /* [IN] the output interface (0 = any) */,
                   in6_addr *ip_dest            /* [IN] the ultimate destination */,
                   _rtcs_if_handle ihandle_dest /* [IN] the outgoing interface (0 = default) */,
                   uint_32 flags                /* [IN] optional flags */ )
{ 
    IP6_CFG_STRUCT_PTR    IP6_cfg_ptr;
    IP_IF_PTR      if_dest = (IP_IF_PTR)ihandle_dest;
    
    IP6_cfg_ptr = RTCS_getcfg(IP6);

    IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_TOTAL++);


    /* Validate destination address.*/
    /* RFC3513: The unspecified address must not be used as the destination address
     * of IPv6 packets or in IPv6 Routing Headers.*/ 
    if((ip_dest == NULL) || IN6_IS_ADDR_UNSPECIFIED(ip_dest))
    {
        IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_DISCARDED++);
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_BAD_ADDRESS);
        RTCSPCB_free(rtcs_pcb);
        return RTCSERR_IP_BAD_ADDRESS;
    }
   
    
    /* 
     * The specified source address may
     * influence the candidate set, by affecting the choice of outgoing
     * interface.  If the application or upper-layer specifies a source
     * address that is in the candidate set for the destination, then the
     * network layer MUST respect that choice.  If the application or
     * upper-layer does not specify a source address, then the network layer
     * uses the source address selection algorithm
     */

    if( ip_src == NULL)
    /* Determine a source address.*/
    {
        ip_src = (in6_addr *)ip6_if_select_src_addr(if_dest, ip_dest);
            
        if(ip_src == NULL)
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_DISCARDED++);
            RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_UNREACH);
            RTCSPCB_free(rtcs_pcb);
            return RTCSERR_IP_UNREACH;
        } 
    }
    

    if(if_dest == NULL)
    /* Determine an output interface.*/
    {
        if_dest = ip6_if_get_by_addr(ip_src);

        if(if_dest == NULL) /* Ther is no any initializaed IF.*/
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_DISCARDED++);
            RTCSLOG_PCB_FREE(rtcs_pcb, RTCSERR_IP_UNREACH);
            RTCSPCB_free(rtcs_pcb);
            return RTCSERR_IP_UNREACH;
        }
    }  
    
    
    
    rtcs_pcb->IFSRC = if_dest; /* Set source Interface.*/
    
    /* If this datagram is for us.*/
    if(ip6_if_is_my_addr(if_dest, ip_dest) == TRUE)
    {  
        if_dest = RTCS_IF_LOCALHOST_PRV; /* Route it to loopback.*/
    }
    
#if RTCSCFG_IP6_LOOPBACK_MULTICAST    
    /* Send Multicast packets also to the loopback.*/
    if( (if_dest != RTCS_IF_LOCALHOST_PRV)
       &&(IN6_IS_ADDR_MULTICAST(ip_dest)) )
    {
        RTCSPCB_PTR dup_pcb;
        
        /* Duplicate the PCB and send it to the loopback. */
        dup_pcb = RTCSPCB_alloc_dup(rtcs_pcb);
        if (dup_pcb)
        {
            dup_pcb->IFSRC = RTCS_IF_LOCALHOST_PRV;
            IP6_send_dgram(RTCS_IF_LOCALHOST_PRV, dup_pcb, ip_src, ip_dest, ip_dest, protocol);
        } 
        else
        {
            IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
        }
    }
#endif /* RTCSCFG_IP6_LOOPBACK_MULTICAST */
    
    return IP6_send_dgram(if_dest, rtcs_pcb, ip_src, ip_dest, ip_dest, protocol);
}



/************************************************************************
* NAME: ip6_get_common_prefix_length
*
* DESCRIPTION: RFC3484 2.2. Common Prefix Length
*  We define the common prefix length CommonPrefixLen(A, B) of two
*  addresses A and B as the length of the longest prefix (looking at the
*  most significant, or leftmost, bits) that the two addresses have in
*  common.  It ranges from 0 to 128.
*************************************************************************/
int ip6_common_prefix_length(in6_addr *ip_addr_1, in6_addr *ip_addr_2)
{
    int length = 0;
    int i;
    int bit_i;
    uint_8 byte_xor;
    
    for(i=0; i < sizeof(in6_addr); i++)
    {
        /* XOR */
        byte_xor = ip_addr_1->s6_addr[i] ^ ip_addr_2->s6_addr[i];
                        
        for(bit_i = 0; bit_i<8; bit_i++)
        {
            if((byte_xor & 0x80) == 0x00)
            {
                length++;
                byte_xor<<=1; /* Shift to the next bit*/
            }
            else
            {
                break;
            }
        }
    } 
    
    return length;
}

/************************************************************************
* NAME: ip6_addr_scope
*
* DESCRIPTION: Returns scope of the IPv6 address (Node-local, 
* link-local, site-local or global.).
*************************************************************************/
int ip6_addr_scope(in6_addr *ip_addr)
{
    /* Local Host. */
    if(IN6_IS_ADDR_LINKLOCAL(ip_addr))
        return IP6_ADDR_SCOPE_LINKLOCAL;
    if(IN6_IS_ADDR_SITELOCAL(ip_addr))    
    	return IP6_ADDR_SCOPE_SITELOCAL;
    
    /* Multicast. */
    if(IN6_IS_ADDR_MULTICAST(ip_addr))
    {
        int scope = IP6_ADDR_MULTICAST_SCOPE(ip_addr);
        
        switch(scope)
        {
            case IP6_ADDR_SCOPE_INTFACELOCAL:
                return IP6_ADDR_SCOPE_INTFACELOCAL;
            case IP6_ADDR_SCOPE_LINKLOCAL:    
                return IP6_ADDR_SCOPE_LINKLOCAL;
            case IP6_ADDR_SCOPE_SITELOCAL:    
                return IP6_ADDR_SCOPE_SITELOCAL;                
            default:    
                return IP6_ADDR_SCOPE_GLOBAL;         
        }
    }
    
    /* Loopback interface - special case.*/
    if(IN6_ARE_ADDR_EQUAL(ip_addr, &in6addr_loopback))
        return IP6_ADDR_SCOPE_INTFACELOCAL;    

    return IP6_ADDR_SCOPE_GLOBAL;
}

/************************************************************************
* NAME: ip6_addr_pefix_cmp
*
* DESCRIPTION: Compares first "prefix_length" bits of the addresses.
*************************************************************************/
boolean ip6_addr_pefix_cmp(in6_addr *addr_1, in6_addr *addr_2, uint_32 prefix_length)
{
    boolean result;
    uint_32 prefix_length_bytes = prefix_length>>3;
    uint_32 prefix_length_bits_mask = ((1<<(prefix_length%8))-1) << (8-(prefix_length%8));
    
    if((prefix_length <= 128) 
        && (memcmp(addr_1, addr_2, prefix_length_bytes) == 0) 
        && ((addr_1->s6_addr[prefix_length_bytes] & prefix_length_bits_mask) == (addr_2->s6_addr[prefix_length_bytes] & prefix_length_bits_mask)) )
    {
        result = TRUE;
    }
    else
    {
        result = FALSE;
    }

    return result;
}

/************************************************************************
* NAME: ip6_get_solicited_multicast_addr
*
* DESCRIPTION: Get IPv6 solicited-node multicast address.
* It has the prefix FF02:0:0:0:0:1:FF00:0000/104 concatenated 
* with the 24 low-order bits of a corresponding IPv6 unicast 
* or anycast address.
*************************************************************************/
void ip6_get_solicited_multicast_addr(in6_addr *ip_addr, in6_addr *solicited_multicast_addr)
{
    solicited_multicast_addr->s6_addr[0]=0xFF;
    solicited_multicast_addr->s6_addr[1]=0x02;
    solicited_multicast_addr->s6_addr[2]=0x00;
    solicited_multicast_addr->s6_addr[3]=0x00;
    solicited_multicast_addr->s6_addr[4]=0x00;
    solicited_multicast_addr->s6_addr[5]=0x00;
    solicited_multicast_addr->s6_addr[6]=0x00;
    solicited_multicast_addr->s6_addr[7]=0x00;
    solicited_multicast_addr->s6_addr[8]=0x00;
    solicited_multicast_addr->s6_addr[9]=0x00;
    solicited_multicast_addr->s6_addr[10]=0x00;
    solicited_multicast_addr->s6_addr[11]=0x01;
    solicited_multicast_addr->s6_addr[12]=0xFF;
    solicited_multicast_addr->s6_addr[13]=ip_addr->s6_addr[13];
    solicited_multicast_addr->s6_addr[14]=ip_addr->s6_addr[14];
    solicited_multicast_addr->s6_addr[15]=ip_addr->s6_addr[15];
}

#endif /* RTCSCFG_ENABLE_IP6 */

/************************************************************************
* NAME: IN6_ADDR_COPY
*
* DESCRIPTION: Copy IPv6 address.
* NOTE: Actually it is workaround for Keil/IAR ARM compilers.
*************************************************************************/
void IN6_ADDR_COPY(in6_addr *from_addr, in6_addr *to_addr)
{
    int i;
    for(i = 0; i< sizeof(struct in6_addr); i++)
        (to_addr)->s6_addr[i] = (from_addr)->s6_addr[i];
}

/* EOF */
