/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: ip6_if.c$
* $Version : 3.8.3.0$
* $Date    : Aug-28-2012$
*
* Comments:
*
*   This file contains the implementation of the
*   packet driver interface.
*
*   2011.09: Added IPv6 support by Andrey Butok
*
*END************************************************************************/

#include <rtcs.h>

#if RTCSCFG_ENABLE_IP6

#include "rtcs_prv.h"
#include "tcpip.h"
#include "ip_prv.h"
#include "route.h"
#include <arp.h>

/************************************************************************
*     Function Prototypes
*************************************************************************/
static int ip6_if_addr_autoconf_set(IP_IF_PTR if_ptr, struct in6_addr *ipaddr);
static uint_32 ip6_if_policy_label( in6_addr *addr);
static void ip6_if_assign_scope_id( IP_IF_PTR if_ptr );

/************************************************************************
*     Policy Table (RFC3484)
*************************************************************************/
typedef struct ip6_if_policy_entry
{
    in6_addr    prefix;             /* Prefix of an IP address. */
    uint_32     prefix_length;      /* Prefix length (in bits). The number of leading bits
                                     * in the Prefix that are valid. */
    uint_32     precedence;         /* Precedence value used for sorting destination addresses.*/
    uint_32     label;              /* The label value Label(A) allows for policies that prefer a particular
                                     * source address prefix for use with a destination address prefix.*/         
} IP6_IF_POLICY_ENTRY, _PTR_ IP6_IF_POLICY_ENTRY_PTR;

/* RFC3484 Default policy table:*/
const IP6_IF_POLICY_ENTRY ip6_if_policy_table[] =
{
    {IN6ADDR(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1),          128,    50, 0},
    {IN6ADDR(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0),          0,      40, 1},
    {IN6ADDR(0x20,0x02,0,0,0,0,0,0,0,0,0,0,0,0,0,0),    16,     30, 2},
    {IN6ADDR(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0),          96,     20, 3},
    {IN6ADDR(0,0,0,0,0,0,0,0,0,0,0xFF,0xFF,0,0,0,0),    96,     10, 4}    
};
#define IP6_IF_POLICY_TABLE_SIZE    (sizeof(ip6_if_policy_table)/sizeof(IP6_IF_POLICY_ENTRY))

static unsigned long last_scope_id = 0;

/************************************************************************
* NAME: ip6_if_get_addr
*
* DESCRIPTION: This function is used to retrieve all IP addresses registerred 
*              with the given interface.
*              Returns TRUE if successful and data structure filled. FALSE in case of error.
*              It returns FALSE if n-th address is not available.
*************************************************************************/
boolean ip6_if_get_addr (_rtcs_if_handle ihandle, uint_32 n, IP6_IF_ADDR_INFO_PTR addr_info) //OK
{
    int i;
    boolean result = FALSE;
    IP_IF_PTR if_ptr = (IP_IF_PTR) ihandle;
    
    if(if_ptr && addr_info)
    {
        for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
        {
            /* Skip NOT_USED addresses. */
            if(if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED)
            {    
                if(n == 0)
                {
                    _mem_copy(&if_ptr->IP6_IF.address[i], addr_info, sizeof(if_ptr->IP6_IF.address[i]));
                    result = TRUE;
                    break;     
                }
                n--;
            }    
        } 
    }
    
    return result;
}

/************************************************************************
* NAME: ip6_if_is_my_addr
*
* DESCRIPTION: Checks if an unicast address is attached/bound to the interface.
*              Returns TRUE if the address is attached/bound, otherwise FALSE.
*************************************************************************/
boolean ip6_if_is_my_addr(_rtcs_if_handle ihandle, in6_addr *ip_addr)
{
    boolean result;
    IP_IF_PTR if_ptr = (IP_IF_PTR) ihandle;
    
  	if(ip6_if_get_addr_info(if_ptr, ip_addr) != NULL)
  	    result = TRUE;
  	else
  	    result = FALSE;
    
    return result;
}

/************************************************************************
* NAME: ip6_if_get_addr_info
*
* DESCRIPTION: Gets address information structure that corresponds to the ip_addr.
*************************************************************************/
IP6_IF_ADDR_INFO_PTR ip6_if_get_addr_info(IP_IF_PTR if_ptr, in6_addr *ip_addr)
{
    int i;
    IP6_IF_ADDR_INFO_PTR result = NULL;
    
    if(if_ptr && ip_addr)
    {
		for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
        {
            /* Skip NOT_USED addresses. */
            if((if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED) &&
               IN6_ARE_ADDR_EQUAL(ip_addr, &if_ptr->IP6_IF.address[i].ip_addr))
            {    
                    result = &if_ptr->IP6_IF.address[i];
                    break;     
            }    
        }     
  	}
    
    return result;
}

/************************************************************************
* NAME: ip6_if_is_my_solicited_multicast_addr
*
* DESCRIPTION: Checks if a solicited multicast address is attached/bound to the interface.
*              Returns TRUE if the solicited multicast address is attached/bound, otherwise FALSE.
*************************************************************************/
boolean ip6_if_is_my_solicited_multicast_addr(_rtcs_if_handle ihandle, in6_addr *ip_addr)
{
    int i;
    boolean result = FALSE;
    IP_IF_PTR if_ptr = (IP_IF_PTR) ihandle;
    
    if(if_ptr && ip_addr)
    {
		for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
        {
            /* Skip NOT_USED addresses. */
            if((if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED) &&
               IN6_ARE_ADDR_EQUAL(ip_addr, &if_ptr->IP6_IF.address[i].solicited_multicast_addr))
            {    
                    result = TRUE;
                    break;     
            }    
        }     
  	}
    
    return result;
}

//TBD look at get_ifid() in BSD
/************************************************************************
* NAME: ip6_if_addr_autoconf_set
*
* DESCRIPTION: Overwrite the last 64 bits with the Interface Identifier.
*              "ipaddr" contains prefix to form an address.
*************************************************************************/
static int ip6_if_addr_autoconf_set(IP_IF_PTR if_ptr, struct in6_addr *ipaddr)
{
    int result = RTCS_OK;
    
    /* Build Interface identifier.
     * Set the 8 last bytes of the IP address based on the Layer 2 identifier.
     */
    switch(if_ptr->DEV_ADDRLEN)
    {
        /* 802.15.4 16-bit short address.*/
        case 2:        
            /* [RFC4944]
             * In these cases, a "pseudo 48-bit address" is formed as follows.  First, the
             * left-most 32 bits are formed by concatenating 16 zero bits to the 16-
             * bit PAN ID (alternatively, if no PAN ID is known, 16 zero bits may be
             * used).  This produces a 32-bit field as follows: 16_bit_PAN:16_zero_bits
             * Then, these 32 bits are concatenated with the 16-bit short address.
             * This produces a 48-bit address.
             */
            /* [16_bit_PAN:16_zero_bits:16_bit_short_address] => [32_zero_bits:16_bit_short_address]
             */
            /* The interface identifier is formed from this 48-bit address as per
             * the "IPv6 over Ethernet" specification [RFC2464].
             */
            _mem_zero (&(ipaddr->s6_addr8[8]), 6);
            ipaddr->s6_addr8[11] = 0xff;
            ipaddr->s6_addr8[12] = 0xfe;
            _mem_copy(&(if_ptr->DEV_ADDR[0]), &(ipaddr->s6_addr8[14]), 2);
             break;
        /* IEEE 802.11 48-bit addresses. MAC address for Ethernet */      
        case 6: 
            /* [RFC2464] The OUI of the Ethernet address (the first three octets) becomes the
             * company_id of the EUI-64 (the first three octets).  The fourth and
             * fifth octets of the EUI are set to the fixed value FFFE hexadecimal.
             * The last three octets of the Ethernet address become the last three
             * octets of the EUI-64.
             */
            _mem_copy(if_ptr->DEV_ADDR, &(ipaddr->s6_addr8[8]),  3);
            ipaddr->s6_addr8[11] = 0xff;
            ipaddr->s6_addr8[12] = 0xfe;
            _mem_copy(&(if_ptr->DEV_ADDR[3]), &(ipaddr->s6_addr8[13]), 3);
            
            /* The Interface Identifier is then formed from the EUI-64 by
             * complementing the "Universal/Local" (U/L) bit, which is the next-to-
             * lowest order bit of the first octet of the EUI-64.
             */            
            ipaddr->s6_addr8[8] ^= 0x02; 
            break;
        /* IEEE EUI-64 identifier. IEEE 802.15.4 */    
        case 8: 
            /* In this case, the Interface Identifier is formed
             * from the EUI-64 according to the "IPv6 over Ethernet" specification
             * [RFC2464].
             */
            _mem_copy(if_ptr->DEV_ADDR, &(ipaddr->s6_addr8[8]), 8);
            
            /* The Interface Identifier is then formed from the EUI-64 by
             * complementing the "Universal/Local" (U/L) bit, which is the next-to-
             * lowest order bit of the first octet of the EUI-64.
             */ 
            ipaddr->s6_addr8[8] ^= 0x02;
            break;
            /* TBD for others.*/
        default:
        #if RTCSCFG_DEBUG_MSG
            printf("Error: Cannot build Interface Identifier.\n");
        #endif            
            result = RTCS_ERROR;
            break;
    }
    
    return result;
}

/************************************************************************
* NAME: IP6_IF_bind_addr
*
* DESCRIPTION: Binds an IP address to a hardware interface. Public API.
*************************************************************************/
void IP6_IF_bind_addr ( IP6_IF_PARM_PTR  parms)
{ 
    uint_32 result = ip6_if_bind_addr ( (IP_IF_PTR) parms->ihandle, &parms->ip_addr, parms->ip_addr_type, IP6_IF_ADDRESS_LIFETIME_INFINITE, ND6_PREFIX_LENGTH_DEFAULT );

    RTCSCMD_complete(parms, result); 
}    

/************************************************************************
* NAME: IP6_IF_unbind_addr
*
* DESCRIPTION: Unbinds an IP address from a hardware interface. Public API.
*************************************************************************/
void IP6_IF_unbind_addr ( IP6_IF_PARM_PTR  parms )
{
    in6_addr    address;
    IP_IF_PTR   if_ptr;
    int         i;
    IP6_IF_ADDR_INFO_PTR if_addr_ptr = NULL;

    /* Get input parameters.*/
    address = parms->ip_addr;
    if_ptr = (IP_IF_PTR) parms->ihandle;
    
    /* Find the address.*/
    for(i = 0; i < IP6_IF_ADDRESSES_MAX; i++)
    {
        if((if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED) && 
            (IN6_ARE_ADDR_EQUAL(&if_ptr->IP6_IF.address[i].ip_addr, &address)))
        {
            /* The bound address is found.*/
            if_addr_ptr = &if_ptr->IP6_IF.address[i];  
            break;
        }
    }
    
    if(if_addr_ptr == NULL)
    {
        RTCSCMD_complete(parms, RTCSERR_IP_BIND_ADDR);
    }
    else
    {
        /* Leave Multicast group.*/
        if(ip6_if_leave (if_ptr, &if_addr_ptr->solicited_multicast_addr) == RTCS_OK)
        {
            printf("Leaved solicited_multicast_addr.\n");
        }
        else
        {
            printf("Failed leaving of  solicited_multicast_addr.\n");
        }
        
        if_addr_ptr->ip_addr_state = IP6_ADDR_STATE_NOT_USED; /* Mark as Not Used.*/
        RTCSCMD_complete(parms, RTCS_OK);
    }
} 

/************************************************************************
* NAME: ip6_if_unbind_addr
*
* DESCRIPTION: Unbinds an IP address from a hardware interface. Internal.
*************************************************************************/
void ip6_if_unbind_addr ( IP_IF_PTR if_ptr, IP6_IF_ADDR_INFO_PTR if_addr )
{
    if(if_ptr && if_addr && (if_addr->ip_addr_state != IP6_ADDR_STATE_NOT_USED))
    {
        /* Leave Multicast group.*/
        ip6_if_leave (if_ptr, &if_addr->solicited_multicast_addr);
        /* Mark as Not Used.*/
        if_addr->ip_addr_state = IP6_ADDR_STATE_NOT_USED; 
    }
} 

/************************************************************************
* NAME: ip6_if_bind_addr
*
* DESCRIPTION: Binds an IP address to a hardware interface. Internal.
*************************************************************************/
uint_32 ip6_if_bind_addr ( IP_IF_PTR if_ptr, in6_addr *ip_addr, ip6_if_addr_type addr_type, uint_32 lifetime /*in seconds*/, uint_32 prefix_length )
{ 
    int i;
    IP6_IF_ADDR_INFO_PTR    if_addr_ptr = NULL;
    uint_32                 result = RTCS_OK ;
    
    /* Check input parameters. */
    if ((if_ptr == NULL) || (ip_addr == NULL) || IN6_IS_ADDR_MULTICAST(ip_addr))
    {
        result = RTCSERR_INVALID_PARAMETER;
        goto BIND_COMPLETE;
    } 
   
    /* Find free address entry 
     */
    for(i = 0; i < IP6_IF_ADDRESSES_MAX; i++)
    {
        if(if_ptr->IP6_IF.address[i].ip_addr_state == IP6_ADDR_STATE_NOT_USED)
        {
            if_addr_ptr = &if_ptr->IP6_IF.address[i];
            break; /* Found free entry.*/
        }
    }
    
    /* If we did not find free address space.
     */
    if(if_addr_ptr == NULL)
    {
        result = RTCSERR_IP6_IF_NO_ADDRESS_SPACE;
        goto BIND_COMPLETE;
    }

    /*
     * Copying address.
     */
    IN6_ADDR_COPY(ip_addr, &if_addr_ptr->ip_addr); 
   
    /* If the address is zero.
     */
    if(IN6_ARE_ADDR_EQUAL(&if_addr_ptr->ip_addr, &in6addr_any))
    {
        /* Set link-local address.
         */
        if_addr_ptr->ip_addr.s6_addr[0] = 0xFE;
        if_addr_ptr->ip_addr.s6_addr[1] = 0x80;
    }
            
    if_addr_ptr->ip_addr_type = addr_type; /* Set type.*/
    
    /* If we are doing Autoconfiguration, the ip_addr points to prefix.*/ 
    if(addr_type == IP6_ADDR_TYPE_AUTOCONFIGURABLE)
    {
        /* Construct address from prefix and interface id. */
        if((prefix_length != ND6_PREFIX_LENGTH_DEFAULT) || ip6_if_addr_autoconf_set(if_ptr, &if_addr_ptr->ip_addr) == RTCS_ERROR)
        {
            result = RTCSERR_INVALID_PARAMETER;
            goto BIND_COMPLETE;
        }
    }
    
    /* Check if addr already exists. Do it here to cover Autoconfiguration case.*/
    if(ip6_if_get_addr_info(if_ptr, &if_addr_ptr->ip_addr) != NULL)
    {
        /* The address is already bound.*/
        goto BIND_COMPLETE;
    }    
    
    /* An address on which the Duplicate Address Detection procedure is
     * applied is said to be tentative until the procedure has completed
     * successfully.
     */
    if_addr_ptr->ip_addr_state = IP6_ADDR_STATE_TENTATIVE;
    
    /* Save creation time, in seconds.*/
    if_addr_ptr->creation_time = RTCS_time_get()/1000;
    
    /* Set lifetime, in seconds.*/
    if_addr_ptr->lifetime = lifetime;
    
    if(if_ptr->IP6_IF.ND6)
    {
        /* Get&Set the solicited-node multicast group-address for assigned ip_addr. */
        ip6_get_solicited_multicast_addr(&if_addr_ptr->ip_addr, &if_addr_ptr->solicited_multicast_addr);    
        
        /*************************************************************************
        * Join Multicast ADDRESSES.
        * When a multicast-capable interface becomes enabled, the node MUST
        * join the all-nodes multicast address on that interface, as well as
        * the solicited-node multicast address corresponding to each of the IP
        * addresses assigned to the interface.
        **************************************************************************/
        /* Join solicited multicast address group.*/
        ip6_if_join (if_ptr, &if_addr_ptr->solicited_multicast_addr);

        /* Start Duplicate Address Detection (DAD).
         * RFC4862:  The Duplicate Address Detection algorithm is performed on all addresses,
         * independently of whether they are obtained via stateless
         * autoconfiguration or DHCPv6.
         */
        nd6_dad_start(if_ptr, if_addr_ptr);
    }


BIND_COMPLETE:

    return result;
} 

/************************************************************************
* NAME: ip6_if_addr_timer
*
* DESCRIPTION: Timer that checks expiring of bound addresses.
*************************************************************************/
void ip6_if_addr_timer ( IP_IF_PTR if_ptr )
{ 
    int i;
    
    for(i = 0; i < IP6_IF_ADDRESSES_MAX; i++)
    {
        /* Check lifetime for address.*/
        if((if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED)
           && (if_ptr->IP6_IF.address[i].lifetime != IP6_IF_ADDRESS_LIFETIME_INFINITE)
           && ((RTCS_time_get()/1000) > (if_ptr->IP6_IF.address[i].creation_time + if_ptr->IP6_IF.address[i].lifetime)) //PFI
            )
        {
            /* RFC4862 5.5.4: An address (and its association with an interface) becomes invalid
             * when its valid lifetime expires. An invalid address MUST NOT be used
             * as a source address in outgoing communications and MUST NOT be
             * recognized as a destination on a receiving interface.*/
            
            ip6_if_unbind_addr (if_ptr, &if_ptr->IP6_IF.address[i]);
        }
    }
}

/************************************************************************
* NAME: ip6_if_join
*
* DESCRIPTION:  Adds a multicast group into the physical interface
*************************************************************************/
uint_32 ip6_if_join ( IP_IF_PTR if_ptr, const in6_addr *group_ip)
{
   if (!if_ptr->DEVICE.JOIN6)
   {
      return RTCS_OK;
   } 

   /* Join the group on the physical interface */
   return if_ptr->DEVICE.JOIN6(if_ptr, group_ip);
}

/************************************************************************
* NAME: ip6_if_join
*
* DESCRIPTION:  Removes a multicast group from the physical interface
*************************************************************************/
uint_32 ip6_if_leave( IP_IF_PTR if_ptr, const in6_addr *group_ip)
{ 
   if (!if_ptr->DEVICE.LEAVE6)
   {
      return RTCS_OK;
   } 

   /* Leave the group on the physical interface */
   return if_ptr->DEVICE.LEAVE6(if_ptr, group_ip);
} 

/************************************************************************
* NAME: ip6_if_policy_label
*
* DESCRIPTION:  Returns label value from policy table.
*************************************************************************/
static uint_32 ip6_if_policy_label( in6_addr *addr)
{ 
    int i;
    uint_32 label = 0;
    int biggest_prefix_length = -1;
    
    /* Find the entry in the list. */
    for(i=0; i<IP6_IF_POLICY_TABLE_SIZE; i++)
    {
        int prefix_length = ip6_if_policy_table[i].prefix_length;
           
        if(ip6_addr_pefix_cmp((in6_addr*)&ip6_if_policy_table[i].prefix, addr, prefix_length) == TRUE)
        {
            if(prefix_length > biggest_prefix_length)
            {
                biggest_prefix_length = prefix_length;
                label = ip6_if_policy_table[i].label;
            }
        
        }
    }
    return label;
} 

/************************************************************************
* NAME: ip6_if_select_src_addr
*
* DESCRIPTION:  Selects the best source address to use with a 
*               destination address, Based on RFC3484.
*************************************************************************/ 
const in6_addr * ip6_if_select_src_addr(IP_IF_PTR if_ptr /* Optional.*/, in6_addr *dest_addr)
{
    int         i;
    int         j;
    in6_addr    *best_addr = NULL;
    int         best_scope;
    int         dest_scope;  
    int         new_scope; 
    int         best_prefix_length; 
    int         new_prefix_length; 
    uint_32     dest_label;
    uint_32     best_label;
    uint_32     new_label;
    IP_IF_PTR   if_dest_cur;
    IP_IF_PTR   if_dest_best;
    
   
    if(dest_addr) 
    {
    
        for(j=0, if_dest_cur = ip6_if_get_by_n(0); if_dest_cur != NULL ;j++, if_dest_cur = ip6_if_get_by_n(j))
        {
            dest_scope = ip6_addr_scope(dest_addr);
            dest_label = ip6_if_policy_label(dest_addr);
                
            /* Just continue the first loop.*/
            for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
            { 
                /* Skip not used enries. */
                if(if_dest_cur->IP6_IF.address[i].ip_addr_state == IP6_ADDR_STATE_NOT_USED)
                {
                        continue;
                }
                else if(best_addr == NULL)
                {
                    /* Just take the first used address, by default.=> Link-local address.*/
                    best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                    if_dest_best = if_dest_cur;
                }
                    
                /* RFC3484 Source Address Selection.*/
                /* Rule 1:  Prefer same address.*/
                if(IN6_ARE_ADDR_EQUAL(dest_addr, &if_dest_cur->IP6_IF.address[i].ip_addr))
                {
                    best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                    if_dest_best = if_dest_cur;
                    break;
                }
                    
                /* Rule 2:  Prefer appropriate scope.*/
                /* If Scope(SA) < Scope(SB): If Scope(SA) < Scope(D), then prefer SB
                 * and otherwise prefer SA.  Similarly, if Scope(SB) < Scope(SA): If
                 * Scope(SB) < Scope(D), then prefer SA and otherwise prefer SB.
                 */
                best_scope = ip6_addr_scope(best_addr);
                new_scope = ip6_addr_scope(&if_dest_cur->IP6_IF.address[i].ip_addr);
                     
                if(best_scope < new_scope)
                {
                    if(best_scope < dest_scope)
                    {
                        best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr; //PFI take pointer at the begining.
                        if_dest_best = if_dest_cur;
                    }
                    continue;
                }
                else if( new_scope < best_scope)
                {
                    if( new_scope >= dest_scope)
                    {
                        best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                        if_dest_best = if_dest_cur;
                    }
                    continue;
                }
                     
                /* Rule 3:  Avoid deprecated addresses.
                 * XXX: Not implemented - we do not store depricated addresses."*/
                     
                /* Rule 4:  Prefer home addresses.
                 * XXX: Not implemented - we do nit have Mobile IPv6.*/
                      
                /* Rule 5:  Prefer outgoing interface.
                 */
                if((if_dest_best == if_ptr ) && (if_dest_cur != if_ptr))
                {
                    continue;
                }
                if((if_dest_best != if_ptr ) && (if_dest_cur == if_ptr))
                {
                    best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                    if_dest_best = if_dest_cur;
                    continue;
                }
                     
                   
                /* Rule 6:  Prefer matching label.
                 * If Label(SA) = Label(D) and Label(SB) <> Label(D), then prefer SA.
                 * Similarly, if Label(SB) = Label(D) and Label(SA) <> Label(D), then
                 * prefer SB.     
                 */
                best_label = ip6_if_policy_label(best_addr);
                new_label = ip6_if_policy_label(&if_dest_cur->IP6_IF.address[i].ip_addr);
                     
                if(best_label == dest_label)
                {
                    if(new_label != dest_label)
                        continue; /* Prefer SA.*/
                }
                if(new_label == dest_label)
                {
                    if(new_label != dest_label)
                    {
                        best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                        if_dest_best = if_dest_cur;
                        continue;
                    }
                }
                     
                /* Rule 7:  Prefer public addresses.
                 * If SA is a public address and SB is a temporary address, then prefer
                 * SA.  Similarly, if SB is a public address and SA is a temporary
                 * address, then prefer SB.
                 * XXX: We do not support "temporary"/"random" addresses.*/
                     
                /* Rule 8:  Use longest matching prefix.
                 * If CommonPrefixLen(SA, D) > CommonPrefixLen(SB, D), then prefer SA.
                 * Similarly, if CommonPrefixLen(SB, D) > CommonPrefixLen(SA, D), then
                 * prefer SB.
                 */
                best_prefix_length = ip6_common_prefix_length(best_addr, dest_addr);
                new_prefix_length = ip6_common_prefix_length(&if_dest_cur->IP6_IF.address[i].ip_addr, dest_addr);
                       
                         
                if(new_prefix_length > best_prefix_length)
                /* Found better one.*/
                {
                    best_addr = &if_dest_cur->IP6_IF.address[i].ip_addr;
                    if_dest_best = if_dest_cur;
                }
            } /* for(IP6_IF_ADDRESSES_MAX) */
        }/* for(if_dest_cur) */
    } /* if(dest_addr) */

    return best_addr;
}

/************************************************************************
* INTERFACE LIST.
*************************************************************************/
static IP_IF_PTR ip6_if_list = NULL;

/************************************************************************
* NAME: ip6_if_add
*
* DESCRIPTION: This function adds interface into the list.
*************************************************************************/
void ip6_if_add( IP_IF_PTR if_ptr )
{
    if_ptr->IP6_IF.next_if = ip6_if_list;

    if(if_ptr->IP6_IF.next_if != 0)
        if_ptr->IP6_IF.next_if->IP6_IF.prev_if = if_ptr;

    if_ptr->IP6_IF.prev_if = 0;
    
    ip6_if_list = if_ptr;
    
    /* Assign Scope ID. */
    ip6_if_assign_scope_id(if_ptr);
}

/************************************************************************
* NAME: ip6_if_del
*
* DESCRIPTION: This function removes interafce from the list. 
*************************************************************************/
void ip6_if_del( IP_IF_PTR if_ptr )
{
    if(if_ptr->IP6_IF.prev_if == 0)
        ip6_if_list = if_ptr->IP6_IF.next_if;
    else
        if_ptr->IP6_IF.prev_if->IP6_IF.next_if = if_ptr->IP6_IF.next_if;

    if(if_ptr->IP6_IF.next_if != 0)
        if_ptr->IP6_IF.next_if->IP6_IF.prev_if = if_ptr->IP6_IF.prev_if;
}

/************************************************************************
* NAME: ip6_if_assign_scope_id
*
* DESCRIPTION: This function assignes unique Scope ID to the interface.
*************************************************************************/
static void ip6_if_assign_scope_id( IP_IF_PTR if_ptr )
{
    uint_32 scope_id;
    IP_IF_PTR current;
    int try_again =1;

    scope_id = last_scope_id + 1;
    
    while(try_again == 1) 
    {
        try_again = 0;
        
        for (current = ip6_if_list; current; current = current->IP6_IF.next_if)
        {
            if(scope_id == current->IP6_IF.scope_id)
            {
                try_again = 1;
                scope_id++;
                break;     
            }
        }
    }
    last_scope_id = scope_id;
    if_ptr->IP6_IF.scope_id = scope_id;
}

/************************************************************************
* NAME: ip6_if_get_by_scope_id
*
* DESCRIPTION: This function returns pointer the interafce according 
*              its Scope ID. 
*              It returns NULL if the interface with passed Scope ID 
*              is not available.
*************************************************************************/
IP_IF_PTR ip6_if_get_by_scope_id( uint_32 scope_id )
{
    IP_IF_PTR result = NULL;
    IP_IF_PTR current;
   
    for (current = ip6_if_list; current; current = current->IP6_IF.next_if)
    {
        if(current->IP6_IF.scope_id == scope_id)
        {
            result = current;
            break;     
        }
    }
    
    return result;
}

/************************************************************************
* NAME: ip6_if_get_scope_id
*
* DESCRIPTION: Gets Scope ID assigned to the interface.
*************************************************************************/
uint_32 ip6_if_get_scope_id(IP_IF_PTR if_ptr)
{
    uint_32 result = 0;
    
    if(if_ptr)
    {
        result = if_ptr->IP6_IF.scope_id;
  	}
    
    return result;
}

/************************************************************************
* NAME: ip6_if_del
*
* DESCRIPTION: This function returns pointer to n-th interafce according 
*              its index (from zero). 
*              It returns NULL if id-th interface is not available.
*************************************************************************/
IP_IF_PTR ip6_if_get_by_n( uint_32 n )
{
    IP_IF_PTR result = NULL;
    IP_IF_PTR current;
   
    for (current = ip6_if_list; current; current = current->IP6_IF.next_if, n--)
    {
        if(n == 0)
        {
            result = current;
            break;     
        }
    }
    
    return result;
}

/************************************************************************
* NAME: ip6_if_get_by_addr
*
* DESCRIPTION: This function returns pointer to interafce which has 
*              the provided address. 
*              It returns NULL if there is no interface with the address..
*************************************************************************/
IP_IF_PTR ip6_if_get_by_addr(in6_addr *addr)
{
    IP_IF_PTR   result = NULL;
    IP_IF_PTR   if_dest_cur;
    int         i;
    
    /* If the source address is explicitly specified by the user the specified address is used.*/
    if(addr)
    {
        /* Find the IF which has the assigned address.*/
        for(i=0, if_dest_cur = ip6_if_get_by_n(0); if_dest_cur != NULL ;i++, if_dest_cur = ip6_if_get_by_n(i))
        {
            if(ip6_if_is_my_addr(if_dest_cur, addr) == TRUE)
            {
                result = if_dest_cur;
                break; /* Found.*/
            }
        }
    }
    
    return result;
}


//////////////// For DEBUG needs only. //////////////////////////////////

/************************************************************************
* NAME: nd6_debug_print_prefix_list
* RETURS: None.
* DESCRIPTION: Prints prefix list. For DEBUG needs only.
*************************************************************************/
void ip6_if_debug_print_addr_info(IP_IF_PTR if_ptr )
{
    int i;
    IP6_IF_ADDR_INFO_PTR entry ;
    
    if(if_ptr == NULL)
        if_ptr = ip6_if_get_by_n(0); /* Get the first one.*/

    printf("Bind Address List:\n");
    for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
    {
        if(if_ptr->IP6_IF.address[i].ip_addr_state != IP6_ADDR_STATE_NOT_USED)
        {
            char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
            entry = &if_ptr->IP6_IF.address[i];

            inet_ntop(AF_INET6, (char*)entry->ip_addr.s6_addr, numaddr, sizeof(numaddr));     
            /* Print entry.*/
            printf("[%d] %s/%d State(%d) Auto(%d) LifeTime(%d)[%d]\n", i, numaddr, entry->prefix_length, entry->ip_addr_state,  entry->ip_addr_type, entry->lifetime, entry->creation_time); 
        }
    }
}

#endif /* RTCSCFG_ENABLE_IP6 */
