/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: ip6_reasm.c$
* $Version : 3.8.1.0$
* $Date    : Jun-14-2012$
*
* Comments:
*
*   This file contains the code to handle IPv6 reassembly
*
*END************************************************************************/

//TBD PFI

#include <stddef.h>
#include <rtcs.h>

#if RTCSCFG_ENABLE_IP6 && RTCSCFG_ENABLE_IP6_REASSEMBLY

#include "rtcs_prv.h"
#include "tcpip.h"
#include "ip_prv.h"
#include "icmp_prv.h"


/*
** Datagrams awaiting reassembly are stored in sets of 256-byte
** blocks.  We allocate a partition of memory blocks; each block
** can be used in one of three ways:
**
**    1) It can store up to 256 data bytes from the datagram.
**       This is an IP6REASM_BLK, called a 'direct block'.
**       This block also has 32-bit bitfield indicating which
**       part(s) of the direct block contain(s) valid data.
**       Each bit represents 8 bytes, so (32 bits) * (8 bytes/bit)
**       = 256 bytes.
**
**    2) It can be used as an array of pointers to IP6REASM_BLKs
**       This is an IP6REASM_INDIDX, called an 'indirect block'.
**       The indirect block could hold up to 65 direct block
**       pointers (257 in 32-bit wide memory), but we only
**       store 64 (or 256) to simplify the math.  i.e.
**       IP6REASM_IND_SIZE could be
**          sizeof(IP6REASM_BLK)/sizeof(IP6REASM_BLK_PTR),
**       but we only set it to
**          IP6REASM_BLK_SIZE/sizeof(IP6REASM_BLK_PTR),
**
**    3) It can hold state information about an IP datagram, an
**       IP6_DGRAM structure.  In particular, it has the datagram
**       'key' (the IP source, dest, protocol and ID), and as
**       many pointers to indirect blocks as it needs to address
**       a maximum sized datagram (256 direct blocks), i.e. 4
**       (or 1).  However, the IP6_DGRAM structure is quite a
**       bit smaller than an IP6REASM_BLK, so difference is
**       filled with pointers to direct blocks -- as many as
**       will fit without exceeding the size of an IP6REASM_BLK.
**       When allocating blocks for incoming IP datagram fragments,
**       the direct block pointers in the IP6_DGRAM are used first,
**       followed by the indirect block pointers.
*/


/* The size of the IP reassembly partition */
#define IP6REASM_PART_MAXSIZE     (10*1024)
#define IP6REASM_PART_GROW        8 /* ~2 KB */
#define IP6REASM_PART_MAX         (IP6REASM_PART_MAXSIZE / sizeof(IP6REASM_BLK))


/* Size of a block in octets */
#define IP6REASM_BLK_SIZE         256
#define IP6REASM_BLK_NUM(offset)  ((offset) / IP6REASM_BLK_SIZE)
#define IP6REASM_BLK_OFS(offset)  ((offset) % IP6REASM_BLK_SIZE)
#define IP6REASM_NB_BLK           (0x10000 / IP6REASM_BLK_SIZE)

/* Number of direct block pointers that will fit in an IP6_DGRAM */
#define IP6REASM_NB_DIRECT        ((sizeof(IP6REASM_BLK)-offsetof(IP6_DGRAM,DIRECT)) \
                                  / sizeof(IP6REASM_BLK_PTR))
#define IP6REASM_IS_DIRECT(block) ((block) < IP6REASM_NB_DIRECT)

/* Number of direct block pointers in each indirect block */
#define IP6REASM_IND_SIZE         (IP6REASM_BLK_SIZE / sizeof(IP6REASM_BLK_PTR))
#define IP6REASM_IND_NUM(block)   (((block)-IP6REASM_NB_DIRECT) / IP6REASM_IND_SIZE)
#define IP6REASM_IND_OFS(block)   (((block)-IP6REASM_NB_DIRECT) % IP6REASM_IND_SIZE)
#define IP6REASM_NB_IND           (IP6REASM_NB_BLK / IP6REASM_IND_SIZE)


typedef struct ip6reasm_blk
{
   uint_32  BUSYINT; /* the bitfield, 1 bit set = 8 bytes copied */
                     /* the LSB is DATA[0-7] */
                     /* the MSB is DATA[249-256] */
   uchar    DATA[IP6REASM_BLK_SIZE];
} IP6REASM_BLK, _PTR_ IP6REASM_BLK_PTR;

/* the structure of an indirect block */
typedef struct ip6reasm_indidx
{
   IP6REASM_BLK_PTR   DIRECT[IP6REASM_IND_SIZE];
} IP6REASM_INDIDX, _PTR_ IP6REASM_INDIDX_PTR;

typedef struct ip6_dgram
{
   /* the "key" of the dgram.  */
   in6_addr IPSRC;
   in6_addr IPDST;
   uint_8      PROTO;
 //  uint_8      RESERVED; //PFI ??
   uint_32     ID;

   /* move-to-front list of IP6_DGRAMs */
   struct ip6_dgram _PTR_       NEXT;
   struct ip6_dgram _PTR_ _PTR_ PREV;

   /* the IP header of the complete datagram */
   IP6_HEADER   IPH;

   /* Information from the PCB with the first fragment */
   uint_32              TYPE;
   pointer              IFSRC;
   RTCS_LINKOPT_STRUCT  LINKOPT;

   uint_32    TOTLEN; /* the data of the complete ip datagram payload */
   uint_32    CURLEN; /* the current data size */
   uint_32    MAXLEN; /* the last fragment seen */

   TCPIP_EVENT TIMER; /* the timer to handle datagram expiry */

   IP6REASM_INDIDX_PTR   INDIR[IP6REASM_NB_IND];
   IP6REASM_BLK_PTR      DIRECT[1];
} IP6_DGRAM, _PTR_ IP6_DGRAM_PTR;


static IP6_DGRAM_PTR  IP6REASM_head;  /* list of all IP6_DGRAMs */
static _rtcs_part    IP6REASM_part;  /* the partition id for the buffer */


static IP6REASM_BLK_PTR IP6REASM_blk_get      (IP6_DGRAM_PTR, uint_32);
static uint_32         IP6REASM_blk_write    (IP6_DGRAM_PTR, IP6REASM_BLK_PTR, uint_32, uchar_ptr, uint_32);
static void            IP6REASM_blk_read_all (IP6_DGRAM_PTR, uchar_ptr, uint_32);
static void            IP6REASM_blk_del_all  (IP6_DGRAM_PTR);

static IP6_DGRAM_PTR IP6REASM_get_dgram ( in6_addr *  ipsrc /* [IN] the IP source */,   
                                          in6_addr *  ipdst /* [IN] the IP destination */,   
                                          uint_8      proto /* [IN] the IP protocol */,   
                                          uint_32     id    /* [IN] the IP ID */);
static void         IP6REASM_del_dgram   (IP6_DGRAM_PTR);
static void         IP6REASM_add_frag    (IP6_DGRAM_PTR, RTCSPCB_PTR);
static RTCSPCB_PTR  IP6REASM_reasm_dgram (IP6_DGRAM_PTR);

static boolean IP6REASM_expire (TCPIP_EVENT_PTR);

//TBD PFI Combine & share code with IPv4 reasambly code.


/************************************************************************
* NAME: IP6_reasm_init
* RETURNS: RTCS_OK or error code
* DESCRIPTION: called by IP_init().
*************************************************************************/
uint_32 IP6_reasm_init (void)
{
   RTCS_DLIST_INIT(IP6REASM_head);

   /*
   ** Allocate the memory partition.  Use an initial count of zero, since we
   ** usually never need to reassemble datagrams.
   */
   IP6REASM_part =  RTCS_part_create(sizeof(IP6REASM_BLK) /* size of one block */, 
                                    0 /* initial number of blocks */,
                                    IP6REASM_PART_GROW /* number of blocks to grow by */, 
                                    IP6REASM_PART_MAX /* maximum number of blocks */, NULL, NULL);
   if (IP6REASM_part == 0)
   {
      return RTCSERR_OUT_OF_MEMORY;
   }

   return RTCS_OK;
}

/************************************************************************
* NAME: IP6REASM_blk_get
* RETURNS: pointer to block containing offset
* DESCRIPTION: Retrieve the block containing a particular datagram offset,
*              allocating one if necessary.
*************************************************************************/
static IP6REASM_BLK_PTR IP6REASM_blk_get
   (
      IP6_DGRAM_PTR   dgram,  /* [IN] the IP datagram */
      uint_32        offset  /* [IN] the offset into the datagram */
   )
{
   IP6REASM_BLK_PTR   blk;
   uint_32           nb;
   uint_32           ni;

   nb = IP6REASM_BLK_NUM(offset);
   if (IP6REASM_IS_DIRECT(nb)) {
      blk = dgram->DIRECT[nb];
      if (!blk) {
         blk = RTCS_part_alloc(IP6REASM_part);
         if (blk) {
            blk->BUSYINT = 0;
            dgram->DIRECT[nb] = blk;
         }
      }
      return blk;
   }

   /* If we get here, it's an indirect block */

   ni = IP6REASM_IND_NUM(nb);
   if (ni >= IP6REASM_NB_IND) {
      return NULL;
   }

   if (!dgram->INDIR[ni]) {
      dgram->INDIR[ni] = RTCS_part_alloc_zero(IP6REASM_part);
      if (!dgram->INDIR[ni]) {
         return NULL;
      }
   }

   nb = IP6REASM_IND_OFS(nb);
   blk = dgram->INDIR[ni]->DIRECT[nb];
   if (!blk) {
      blk = RTCS_part_alloc(IP6REASM_part);
      if (blk) {
         blk->BUSYINT = 0;
         dgram->INDIR[ni]->DIRECT[nb] = blk;
      }
   }
   return blk;

}

/************************************************************************
* NAME: IP6REASM_blk_write
* RETURNS: number of bytes written
* DESCRIPTION: Writes data to blk at offset.
*************************************************************************/
static uint_32 IP6REASM_blk_write
   (
      IP6_DGRAM_PTR      dgram,
      IP6REASM_BLK_PTR   blk,
      uint_32           offset,
      uchar_ptr         ptr,
      uint_32           len
   )
{
   uint_32 blkoffset, bitoffset, numbits;
   uint_32 mask, oldbits;

   blkoffset = IP6REASM_BLK_OFS(offset);

   if (blkoffset + len > IP6REASM_BLK_SIZE) {
      len = IP6REASM_BLK_SIZE - blkoffset;
   }
   _mem_copy(ptr, blk->DATA + blkoffset, len);

   bitoffset = blkoffset >> 3;  /* first bit to set in BUSYINT */
   numbits   = (len+7)   >> 3;  /* number of bits to set */
   /* mask = (1 << (bitoffset+numbits)) - (1 << bitoffset) */
   mask = (2 << (bitoffset+numbits-1)) - (1 << bitoffset);

   oldbits = mask & blk->BUSYINT;
   while (oldbits) {
      oldbits &= oldbits - 1;
      numbits--;
   } 
   blk->BUSYINT |= mask;

   dgram->CURLEN += numbits << 3;
   if (dgram->MAXLEN < offset + len) {
      dgram->MAXLEN = offset + len;
   }

   return len;

} 

/************************************************************************
* NAME: IP6REASM_blk_read_all
* RETURNS: void
* DESCRIPTION: Reads data into ptr.
*************************************************************************/
static void IP6REASM_blk_read_all
   (
      IP6_DGRAM_PTR   dgram,
      uchar_ptr      ptr,
      uint_32        len
   )
{
   uint_32              i, j;
   IP6REASM_INDIDX_PTR   indir;

   /* Read the direct blocks */
   for (i = 0; i < IP6REASM_NB_DIRECT; i++) {
      if (len <= IP6REASM_BLK_SIZE) {
         _mem_copy(dgram->DIRECT[i]->DATA, ptr, len);
         return;
      }
      _mem_copy(dgram->DIRECT[i]->DATA, ptr, IP6REASM_BLK_SIZE);
      ptr += IP6REASM_BLK_SIZE;
      len -= IP6REASM_BLK_SIZE;
   }

   /* Read the indirect blocks */
   for (i = 0; i < IP6REASM_NB_IND; i++) {
      indir = dgram->INDIR[i];
      for (j = 0; j < IP6REASM_IND_SIZE; j++) {
         if (len <= IP6REASM_BLK_SIZE) {
            _mem_copy(indir->DIRECT[j]->DATA, ptr, len);
            return;
         }
         _mem_copy(indir->DIRECT[j]->DATA, ptr, IP6REASM_BLK_SIZE);
         ptr += IP6REASM_BLK_SIZE;
         len -= IP6REASM_BLK_SIZE;
      }
   }

}

/************************************************************************
* NAME: IP6REASM_blk_del_all
* RETURNS: void
* DESCRIPTION: Free all blocks allocated to a datagram.
*************************************************************************/
static void IP6REASM_blk_del_all (IP6_DGRAM_PTR    dgram   /* [IN] the IP datagram */)
{
    uint_32              i, j;
    IP6REASM_INDIDX_PTR   indir;

    /* free the direct blocks */
    for (i = 0; i < IP6REASM_NB_DIRECT; i++)
    {
        if (!dgram->DIRECT[i])
            continue;
        RTCS_part_free(dgram->DIRECT[i]);
    }

    /* free the indirect blocks */
    for (i = 0; i < IP6REASM_NB_IND; i++)
    {
        indir = dgram->INDIR[i];
        if (!indir)
            continue;
        for (j = 0; j < IP6REASM_IND_SIZE; j++)
        {
            if (!indir->DIRECT[j])
                continue;
            RTCS_part_free(indir->DIRECT[j]);
        }
        RTCS_part_free(indir);
    }
}

/************************************************************************
* NAME: IP6REASM_get_dgram
* RETURNS: pointer the matching IP datagram.
* DESCRIPTION: Find the IP datagram matching the given (ipsrc, ipdst, proto, id)
*     key.
*     If the structure isn't in the list, allocate one.
*************************************************************************/
static IP6_DGRAM_PTR IP6REASM_get_dgram ( in6_addr *  ipsrc /* [IN] the IP source */,   
                                          in6_addr *  ipdst /* [IN] the IP destination */,   
                                          uint_8      proto /* [IN] the IP protocol */,   
                                          uint_32     id    /* [IN] the IP ID */)
{
    register IP6_DGRAM_PTR dgram;

    RTCS_DLIST_PEEK(IP6REASM_head, dgram);
    if (dgram)
    {
        /* There's a good probability our IP6_DGRAM is at the head of the list */
        if (IN6_ARE_ADDR_EQUAL( ipsrc,  &dgram->IPSRC )
            && IN6_ARE_ADDR_EQUAL( ipdst, &dgram->IPDST )
            && (proto == dgram->PROTO)
            && (id    == dgram->ID))
        {
            return dgram;
        }

        /* It's not at the head.  Scan the list for it */
        RTCS_DLIST_SEARCH_REST(IP6REASM_head, dgram)
        {
            if (IN6_ARE_ADDR_EQUAL( ipsrc,  &dgram->IPSRC )
                && IN6_ARE_ADDR_EQUAL( ipdst, &dgram->IPDST )
                && (proto == dgram->PROTO)
                && (id    == dgram->ID))
            {
                /* Found it.  Move it to the head of the list */
                RTCS_DLIST_DEL(IP6REASM_head, dgram);
                RTCS_DLIST_INS(IP6REASM_head, dgram);
                return dgram;
            }
        } /* End SEARCH */
    }

    /* Not found.  Allocate one */
    dgram = RTCS_part_alloc_zero(IP6REASM_part);
    if (!dgram)
        return NULL;

    /* Initialize it */
    IN6_ADDR_COPY(ipsrc, &dgram->IPSRC); 
    IN6_ADDR_COPY(ipdst, &dgram->IPDST); 
    
    dgram->PROTO = proto;
    dgram->ID    = id;

    /* If insufficient fragments are received to complete reassembly of a
     * packet within 60 seconds of the reception of the first-arriving
     * fragment of that packet, reassembly of that packet must be
     * abandoned and all the fragments that have been received for that
     * packet must be discarded.
     */ 
    
    dgram->TIMER.TIME  = IP6REASM_TTL;
    dgram->TIMER.EVENT = IP6REASM_expire;
    dgram->TIMER.PRIVATE = dgram;
    TCPIP_Event_add(&dgram->TIMER);

    /* Add it to the list*/
    RTCS_DLIST_INS(IP6REASM_head, dgram);
    return dgram;

} 

/************************************************************************
* NAME: IP6REASM_del_dgram
* RETURNS: void
* DESCRIPTION: Free an IP datagram.
*************************************************************************/
static void IP6REASM_del_dgram ( IP6_DGRAM_PTR   dgram    /* [IN] the IP datagram */)
{
   /* Cancel the expire timer and free all memory */
   TCPIP_Event_cancel(&dgram->TIMER);
   RTCS_DLIST_DEL(IP6REASM_head, dgram);
   IP6REASM_blk_del_all(dgram);
   RTCS_part_free(dgram);
}

/************************************************************************
* NAME: IP6REASM_del_dgram
* RETURNS: void
* DESCRIPTION: Adds a received IP fragment to an IP datagram.
*************************************************************************/
static void IP6REASM_add_frag
   (
      IP6_DGRAM_PTR     dgram,   /* [IN] the dgram */
      RTCSPCB_PTR       inpcb    /* [IN] the received PCB */
   )
{ 
    IP6_HEADER_PTR ip6_h = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(inpcb);
   
    IP6_FRAGMENT_HEADER_PTR ip6_frag_h    = (IP6_FRAGMENT_HEADER_PTR)RTCSPCB_DATA(inpcb);
   
    uint_32           iphlen  = (uint_32)ip6_frag_h - (uint_32)ip6_h + sizeof(IP6_FRAGMENT_HEADER);
    uint_32           totlen  = IP6_HEADER_LEN + ntohs(ip6_h->LENGTH);
    int_32            offset  = ntohs(ip6_frag_h->OFFSET_MORE);
    int_32            first   = offset & IP6_FRAGMENT_OFFSET_MASK;
   
   
    uint_32           datalen = totlen - iphlen;
    uchar_ptr         data    = (uchar_ptr)ip6_h + iphlen;
    IP6REASM_BLK_PTR  blk;
    uint_32           len;

    if (offset & IP6_FRAGMENT_MF)
    {
        /* Fragments (except the last) must be multiples of 8 bytes */
        if ((datalen & 0x07) != 0)
        {
            /* If the length of a fragment, as derived from the fragment packets
             * Payload Length field, is not a multiple of 8 octets and the M flag
             * of that fragment is 1, then that fragment must be discarded and an
             * ICMP Parameter Problem, Code 0, message should be sent to the
             * source of the fragment, pointing to the Payload Length field of
             * the fragment packet. */
            ICMP6_send_error ( ICMP6_TYPE_PARAM_PROB     /* [IN] the ICMPv6 type to send */,    
                                ICMP6_CODE_PP_HEADER     /* [IN] the ICMPv6 code to send */,
                                (uint_32)ip6_h->LENGTH - (uint_32)ip6_h  /* [IN] error parameter */,
                                inpcb   /* [IN] the packet which caused the error */ );
            return;
        }
    }
    else
    {
        /* If this fragment is the last one, record the total length */
        dgram->TOTLEN = first + datalen;
    }

    /* Once we get the first fragment, record the IP header and 8 data bytes */
    if (first == 0)
    {
        _mem_copy(ip6_h, &dgram->IPH, IP6_HEADER_LEN);
        dgram->TYPE    = inpcb->TYPE;
        dgram->IFSRC   = inpcb->IFSRC;
        dgram->LINKOPT = inpcb->LINK_OPTIONS.RX;
    }

    while (datalen)
    {
        blk = IP6REASM_blk_get(dgram, first);
        if (!blk)
        {
            return;
        } 

        len = IP6REASM_blk_write(dgram, blk, first, data, datalen);
        first   += len;
        data    += len;
        datalen -= len;
    }
}

/************************************************************************
* NAME: IP6REASM_reasm_dgram
* RETURNS: RTCSPCB containing the completed datagram.
* DESCRIPTION: Reassemble a completed datagram.
*************************************************************************/
static RTCSPCB_PTR IP6REASM_reasm_dgram ( IP6_DGRAM_PTR   dgram    /* [IN] the dgram descriptor */)
{
   RTCSPCB_PTR outpcb;
   PCB_PTR     bpcb;
   PCB_FRAGMENT _PTR_ pcb_frag_ptr;
   uchar_ptr   data;
   uint_32     iphlen  = IP6_HEADER_LEN;
   uint_32     ip_totlen = iphlen + dgram->TOTLEN;

   bpcb = RTCS_mem_alloc_system(sizeof(PCB) + sizeof(PCB_FRAGMENT) + ip_totlen);
   if (!bpcb)
   {
      return NULL;
   }

   data = (uchar_ptr)bpcb + sizeof(PCB) + sizeof(PCB_FRAGMENT);
   bpcb->FREE    = (void(_CODE_PTR_)(PCB_PTR))_mem_free;
   bpcb->PRIVATE = NULL;
 
   pcb_frag_ptr = bpcb->FRAG;
   pcb_frag_ptr->LENGTH   = ip_totlen;
   pcb_frag_ptr->FRAGMENT = data;
   pcb_frag_ptr++;
   pcb_frag_ptr->LENGTH   = 0;
   pcb_frag_ptr->FRAGMENT = NULL;

   /* Copy the IP header */
   _mem_copy(&dgram->IPH, data, iphlen);
   data += iphlen;

   /* Copy the stored data in the new packet */
   IP6REASM_blk_read_all(dgram, data, dgram->TOTLEN);

   /* Put it in an RTCSPCB */
   outpcb = RTCSPCB_alloc_recv(bpcb);
   if (outpcb == NULL)
   {
      PCB_free(bpcb);
      return NULL;
   } 

   outpcb->IFSRC           = dgram->IFSRC;
   outpcb->TYPE            = dgram->TYPE;
   outpcb->LINK_OPTIONS.RX = dgram->LINKOPT;
   RTCSPCB_DATA_NETWORK(outpcb) = RTCSPCB_DATA(outpcb);
   RTCSPCB_SET_TRANS_PROTL(outpcb, dgram->PROTO);
   RTCSPCB_SET_TRANS_DELTA(outpcb, iphlen);

   /* Delete the local structure */
   IP6REASM_del_dgram(dgram);

   return outpcb;

}

/************************************************************************
* NAME: IP6REASM_expire
* RETURNS: boolean
* DESCRIPTION: Called by the Timer.  Expire an IP datagram.
*************************************************************************/
static boolean IP6REASM_expire ( TCPIP_EVENT_PTR   event    /* [IN/OUT] the expire event */)
{ 
    IP6_CFG_STRUCT_PTR IP6_cfg_ptr = RTCS_getcfg(IP6);
    IP6_DGRAM_PTR      dgram = event->PRIVATE;
    uint_32  error;

    IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.ST_RX_FRAG_DISCARDED++);
 
    /* If the first fragment (i.e., the one
     * with a Fragment Offset of zero) has been received, an ICMP Time
     * Exceeded -- Fragment Reassembly Time Exceeded message should be
     * sent to the source of that fragment.
     */
    if (dgram->DIRECT[0]
        && dgram->DIRECT[0]->BUSYINT & 1
        && dgram->TYPE & RTCSPCB_TYPE_UNICAST)
    {
        uchar_ptr       buffer;
        RTCSPCB_PTR     orig_pcb;
        
        /*************************************
         * Reconstact PCB for ICMP error.
         *************************************/
        
        /* Allocate a PCB */
        orig_pcb = RTCSPCB_alloc_send();
        if (orig_pcb != NULL)
        {
            error = RTCSPCB_insert_header(orig_pcb, sizeof(IP6_HEADER) + sizeof(IP6_FRAGMENT_HEADER));
            if (error)
            {
                IF_IP6_STATS_ENABLED(IP6_cfg_ptr->STATS.COMMON.ST_TX_MISSED++);
                IF_IP6_STATS_ENABLED(RTCS_seterror(&IP6_cfg_ptr->STATS.ERR_TX, error, (uint_32)orig_pcb));
            }
            else
            {
                IP6_FRAGMENT_HEADER_PTR frag_header;
                
                buffer = RTCSPCB_DATA(orig_pcb);


                _mem_copy(&dgram->IPH, buffer, sizeof(IP6_HEADER));
                
                
                orig_pcb->IFSRC           = dgram->IFSRC;
                orig_pcb->TYPE            = dgram->TYPE;
                orig_pcb->LINK_OPTIONS.RX = dgram->LINKOPT;
                RTCSPCB_DATA_NETWORK(orig_pcb) = RTCSPCB_DATA(orig_pcb);
                RTCSPCB_SET_TRANS_PROTL(orig_pcb, dgram->PROTO);
                RTCSPCB_SET_TRANS_DELTA(orig_pcb, sizeof(IP6_HEADER));
                
                
                /*************************************
                * Build the FRAGMENT Header. 
                *************************************/
                frag_header = (IP6_FRAGMENT_HEADER_PTR)(RTCSPCB_DATA(orig_pcb) + IP6_HEADER_LEN);
                htons(frag_header->OFFSET_MORE, IP6_FRAGMENT_MF);
                htonc(frag_header->NEXT_HEADER, dgram->PROTO);
                htonl(frag_header->ID, dgram->ID);
                
                /* Send a ICMP Time Exceeded */
                ICMP6_send_error (ICMP6_TYPE_TIME_EXCEED, ICMP6_CODE_TE_FRG_REASSEMBLY,
                                0   /* [IN] error parameter */,
                                orig_pcb   /* [IN] the packet which caused the error */);
            }
            /* Free orig_pcb.*/
            RTCSPCB_free(orig_pcb);
        }
    }

    IP6REASM_del_dgram(dgram);
    return FALSE;
} 

/************************************************************************
* NAME: IP6_reasm
* RETURNS: RTCS_OK or error code
* DESCRIPTION: Process an IP fragment.  If an error occurs, outpcb is
*               guaranteed to be NULL.
*  Warning:
*     It is assumed the PCB is not fragmented.  This is
*     true for all PCBs that come from the link layer.
*     Loopback PCBs generally are fragmented, but IP
*     doesn't FRAGMENT them, so they don't go through
*     this function.
*
*  Warning:
*     I just realized that the word 'fragment' is overloaded
*     in the warning above.  There are PCB fragments and IP
*     fragments.  I've rewritten the previous warning so that
*     PCB fragments are 'fragments' (lowercase) and IP fragments
*     are 'FRAGMENTS' (uppercase).
*************************************************************************/
uint_32 IP6_reasm ( RTCSPCB_PTR  pcb            /* [IN] the packet to deliver */,     
                    RTCSPCB_PTR _PTR_ outpcb    /* [OUT] the reassembled packet or NULL */)
{
   IP6_HEADER_PTR ip6_h = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(pcb);
   IP6_FRAGMENT_HEADER_PTR ip6_frag_h    = (IP6_FRAGMENT_HEADER_PTR)RTCSPCB_DATA(pcb);
   
   /*
    * An original packet is reassembled only from fragment packets that
    * have the same Source Address, Destination Address, and Fragment
    * Identification.
    */
   
   in6_addr *    ip_src  = (in6_addr *)(ip6_h->SOURCE);
   in6_addr *    ip_dst  = (in6_addr *)(ip6_h->DEST);
   uint_8        next_header  = ntohc(ip6_frag_h->NEXT_HEADER);
   uint_32       id     = ntohl(ip6_frag_h->ID);
   
   IP6_DGRAM_PTR   dgram;

   *outpcb = NULL;

   dgram = IP6REASM_get_dgram(ip_src, ip_dst, next_header, id);
   if (!dgram)
   {
      RTCSLOG_PCB_FREE(pcb, RTCSERR_OUT_OF_MEMORY);
      RTCSPCB_free(pcb);
      return RTCSERR_OUT_OF_MEMORY;
   }    

   IP6REASM_add_frag(dgram, pcb);
   

    /*
    ** We're done when:
    **    0 < TOTLEN = MAXLEN <= CURLEN
    **
    ** Note: we can get TOTLEN < CURLEN because CURLEN is
    ** always a multiple of eight, but TOTLEN may not be.
    */
    if (dgram->TOTLEN > 0
        && dgram->MAXLEN == dgram->TOTLEN
        && dgram->CURLEN >= dgram->MAXLEN)
    {
        *outpcb = IP6REASM_reasm_dgram(dgram);
    }

    /* Free the old pcb */
    RTCSLOG_PCB_FREE(pcb, RTCS_OK);
    RTCSPCB_free(pcb);

    return RTCS_OK;
} 


#endif /* RTCSCFG_ENABLE_IP6 && RTCSCFG_ENABLE_IP6_REASSEMBLY */
/* EOF */
