/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: iplocal6.c$
* $Version : 3.8.1.0$
* $Date    : Jun-14-2012$
*
* Comments:
*
*   This file contains the interface between IP6 and
*   the local host.
*
*   2011.09: Added IPv6 support by Andrey Butok.
*
*END************************************************************************/

#include <rtcs.h>
#include "rtcs_prv.h"
#include "tcpip.h"
#include "ip_prv.h"
#include "icmp_prv.h"

#if RTCSCFG_ENABLE_IP6

/************************************************************************
* NAME: IP6_open
*
* DESCRIPTION: Registers a transport protocol with IP6.
*              TBD. Combine it wuth IPv4.
*************************************************************************/
ICB_STRUCT_PTR IP6_open ( uchar protocol,   /* [IN]  Protocol to use */
                          IP_SERVICE_FN service, /* [IN]  Packet receive function */
                          pointer private,  /* [IN]  Owner's config block */
                          uint_32_ptr status /* [OUT] return code */)
{
   IP6_CFG_STRUCT_PTR    IP6_cfg_ptr;
   ICB_STRUCT_PTR icb, search_ptr;

   IP6_cfg_ptr = RTCS_getcfg(IP6);

   /* Make sure the protocol isn't already is use */
   RTCS_LIST_SEARCH(IP6_cfg_ptr->ICB6_HEAD, search_ptr)
   {
      if (search_ptr->PROTOCOL == protocol)
      {
         *status = RTCSERR_IP_PROT_OPEN;
         return NULL;
      }
   }

   icb = RTCS_mem_alloc(sizeof(ICB_STRUCT));
   if (!icb)
   {
      *status = RTCSERR_IP_ICB_ALLOC;
      return NULL;
   }

   _mem_set_type(icb, MEM_TYPE_ICB);

   icb->PROTOCOL = protocol & 0xFF;
   icb->SERVICE  = service;
   icb->PRIVATE  = private;

   /*
   ** Add the record into the chain of open ICBs
   */
   RTCS_LIST_INS(IP6_cfg_ptr->ICB6_HEAD, icb);

   /*
   ** Open new ICB succeeded
   */
   *status = RTCS_OK;
   return icb;
}

/************************************************************************
* NAME: IP6LOCAL_open
*
* DESCRIPTION: Do IPv6-specific initialization for the 
*              localhost interface.
*************************************************************************/
void IP6LOCAL_open (IP_IF_PTR if_ptr )
{
    
    /* Add to IF list.*/ //TBD make it common for all interfaces.
    ip6_if_add(if_ptr);
    
    /* RFC3513: 2.5.3 The Loopback Address
    * The unicast address 0:0:0:0:0:0:0:1 is called the loopback address.
    * It may be used by a node to send an IPv6 packet to itself.  It may
    * never be assigned to any physical interface.   It is treated as
    * having link-local scope, and may be thought of as the link-local
    * unicast address of a virtual interface (typically called "the
    * loopback interface") to an imaginary link that goes nowhere.
    * 
    * The loopback address must not be used as the source address in IPv6
    * packets that are sent outside of a single node.  An IPv6 packet with
    * a destination address of loopback must never be sent outside of a
    * single node and must never be forwarded by an IPv6 router.  A packet
    * received on an interface with destination address of loopback must be
    * dropped.
    */
    
    ip6_if_bind_addr (if_ptr, (in6_addr *)&in6addr_loopback, IP6_ADDR_TYPE_MANUAL, IP6_IF_ADDRESS_LIFETIME_INFINITE, 128); 
}

/************************************************************************
* NAME: IP6LOCAL_send
* RETURNS : RTCS_OK or error code.
* DESCRIPTION: Sends an IPv6 packet to loopback.
*************************************************************************/
uint_32 IP6LOCAL_send(  IP_IF_PTR if_ptr     /* [IN] the IP interface structure */,
                        RTCSPCB_PTR rtcs_pcb /* [IN] the packet to send */,
                        in6_addr *src        /* [IN] the next-hop source address */,
                        in6_addr *dest       /* [IN] the next-hop destination address */)
{ 
    printf("IP6LOCAL\n");
    if (!RTCSCMD_service(rtcs_pcb, IP6_service))
    {
        IF_IPIF_STATS_ENABLED(if_ptr->STATS.COMMON.ST_RX_MISSED++);
        RTCSLOG_PCB_FREE(rtcs_pcb, RTCS_OK);
        RTCSPCB_free(rtcs_pcb);
    }

    return RTCS_OK;
}
#endif /* RTCSCFG_ENABLE_IP6 */
