/**HEADER********************************************************************
* 
* Copyright (c) 2011 by Andrey Butok, Freescale Semiconductor;
* All Rights Reserved
*
*
*************************************************************************** 
*
* THIS SOFTWARE IS PROVIDED BY FREESCALE "AS IS" AND ANY EXPRESSED OR 
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  
* IN NO EVENT SHALL FREESCALE OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
* THE POSSIBILITY OF SUCH DAMAGE.
*
**************************************************************************
*
* $FileName: nd6.c$
* $Version : 3.8.2.0$
* $Date    : Oct-2-2012$
*
* Comments:
*
*   This file contains the implementation of the Neighbor Discovery for IP6.
*
*   2011.10: Added IPv6 support by Andrey Butok
*
*END************************************************************************/

//TBD PFI Parsering options.
#include <string.h>
#include <rtcs.h>

#if RTCSCFG_ENABLE_IP6

#include "rtcs_prv.h"
#include "tcpip.h"
#include "ip_prv.h"

#include "ip6_trace.h"
#include "nd6.h"

/******************************************************************
* Definitions
*******************************************************************/


/******************************************************************
* Function Prototypes
*******************************************************************/
static boolean nd6_timer (TCPIP_EVENT_PTR   event);
static void nd6_dad_timer(IP_IF_PTR if_ptr );
static void nd6_dad_failed(IP_IF_PTR if_ptr, IP6_IF_ADDR_INFO_PTR addr_info);
static void nd6_rd_timer(IP_IF_PTR if_ptr );

/************************************************************************
* NAME: ND6_open
* RETURNS: RTCS_OK or error code
* DESCRIPTION: Initializes the Neighbor Disscovery on an interface.
*************************************************************************/
uint_32 nd6_open (IP_IF_PTR   if_ptr)
{
    ND6_CFG_PTR         nd6_cfg;
    IP6_CFG_STRUCT_PTR  IP6_cfg_ptr = RTCS_getcfg(IP6);
    int                 i;

    /* Allocate an ND configuration data.*/
    nd6_cfg = RTCS_mem_alloc_zero(sizeof(ND6_CFG)); //PFI
    if (!nd6_cfg)
    {
        return RTCSERR_ND6_CFG;
    } 
    _mem_set_type(nd6_cfg, MEM_TYPE_ND6_CFG);

//PFI Delete it

    /* --- Initialize Neighbor Cache & Default Router List. ----*/
    for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
    {
        nd6_cfg->neighbor_cache[i].state = ND6_NEIGHBOR_STATE_NOTUSED;
    }

//PFI add it via function, as in FNET.    
    /* --- Initialize Prefix List. ----*/
    
    /* The link-local prefix is considered to be on the
     * prefix list with an infinite invalidation timer
     * regardless of whether routers are advertising a
     * prefix for it.
     */
    in6addr(&nd6_cfg->prefix_list[0].prefix,0xFE,0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
    nd6_cfg->prefix_list[0].prefix_length = ND6_PREFIX_LENGTH_DEFAULT;
    nd6_cfg->prefix_list[0].lifetime = ND6_PREFIX_LIFETIME_INFINITE;
    nd6_cfg->prefix_list[0].state = ND6_PREFIX_STATE_USED;
    nd6_cfg->prefix_list[0].creation_time = RTCS_time_get()/1000;
    for(i=1; i < ND6_PREFIX_LIST_SIZE; i++)
    {
        nd6_cfg->prefix_list[i].state = ND6_PREFIX_STATE_NOTUSED;
    }

    /* ---- Initialize interface variables. ----*/
    /* The recommended MTU for the link. */
    nd6_cfg->mtu = if_ptr->MTU;  
    if(nd6_cfg->mtu < IP6_DEFAULT_MTU)
        nd6_cfg->mtu = IP6_DEFAULT_MTU;
    
    /* Cur Hop Limit value */
    nd6_cfg->cur_hop_limit = IP6_cfg_ptr->DEFAULT_HOPLIMIT;
    
    /* Reachable time. used for NU.*/
    nd6_cfg->reachable_time = ND6_REACHABLE_TIME;
    
    /* Time between retransmissions of Neighbor Solicitation messages.*/
    nd6_cfg->retrans_timer = ND6_RETRANS_TIMER;
    
        
    //TBD other variables.
    
    /* --- Register timer to check ND lists and N cache. ---*/
    nd6_cfg->timer.TIME = ND6_TIMER_PERIOD; 
    nd6_cfg->timer.PRIVATE = if_ptr; 
    nd6_cfg->timer.EVENT = nd6_timer;
    
    TCPIP_Event_add(&nd6_cfg->timer);
    
    if_ptr->IP6_IF.ND6 = nd6_cfg;
    return RTCS_OK;
}

/************************************************************************
* NAME: ND6_close
* DESCRIPTION: Shuts down the Neighbor Disscovery on an interface.
*************************************************************************/
void nd6_close (IP_IF_PTR   if_ptr)
{
   ND6_CFG_PTR    nd6_cfg;

   /* Deallocate ND data.*/
   nd6_cfg = if_ptr->IP6_IF.ND6;
   if (nd6_cfg)
   {
        /* Cancel timer.*/
        TCPIP_Event_cancel(&nd6_cfg->timer);
        
        /* Free allocated memory.*/
        _mem_free(nd6_cfg);
        if_ptr->IP6_IF.ND6 = NULL;
    }
}

/************************************************************************
* NAME: nd6_timer
* DESCRIPTION: Called by the Timer.
*************************************************************************/
static boolean nd6_timer ( TCPIP_EVENT_PTR   event )
{
    IP_IF_PTR   if_ptr = event->PRIVATE;
    ND6_CFG_PTR    nd6_cfg = if_ptr->IP6_IF.ND6;
    ND6_NEIGHBOR_ENTRY_PTR neighbor_entry;
    int i;

    /* DAD timer.*/
    nd6_dad_timer(if_ptr); 
    

    /* Neighbor Cache processing.*/
    for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
    {
        neighbor_entry = &nd6_cfg->neighbor_cache[i];
        
        /* Check router expiration */
        if((neighbor_entry->state != ND6_NEIGHBOR_STATE_NOTUSED)
           && (neighbor_entry->is_router == 1) 
           && ((RTCS_time_get()/1000) > (neighbor_entry->creation_time + neighbor_entry->router_lifetime)))
        {
            nd6_router_list_del(neighbor_entry);
        }
        
        switch(neighbor_entry->state)
        {
            /* =========== INCOMPLETE ============= */
            case ND6_NEIGHBOR_STATE_INCOMPLETE:
                if((uint_32)(RTCS_time_get() - neighbor_entry->state_time ) > nd6_cfg->retrans_timer )
                {
                    neighbor_entry->solicitation_send_counter++;
                    if(neighbor_entry->solicitation_send_counter >= ND6_MAX_MULTICAST_SOLICIT)
                    /* AR is failed.*/
                    {
                        /* RFC4861 7.3.3: If address resolution fails, the entry SHOULD be deleted, so that subsequent
                         * traffic to that neighbor invokes the next-hop determination procedure
                         * again. Invoking next-hop determination at this point ensures that
                         * alternate default routers are tried.
                         */                        
                        nd6_neighbor_cache_del(neighbor_entry);
                        //TBD ICMP error to upper layer.
                    #if RTCSCFG_DEBUG_MSG                        
                        printf("ND6: AR is FAILED \n");
                    #endif                      
                    }
                    else
                    /* AR: transmit NS.*/
                    {
                        in6_addr *dest_addr = (in6_addr *)&neighbor_entry->ip_addr;
                        
                        /* Select source address.*/ //PFI
                        in6_addr *src_addr = &neighbor_entry->solicitation_src_ip_addr;
                        neighbor_entry->state_time = RTCS_time_get();
                        /* AR: Transmitting a Neighbor Solicitation message targeted at the neighbor.*/
                        nd6_neighbor_solicitation_send(if_ptr, src_addr, NULL /* NULL for AR */, dest_addr);
                    }
                }
                break;
            /* =========== REACHABLE ============= */    
            case ND6_NEIGHBOR_STATE_REACHABLE:
                /*
                 * RFC 4861: When ReachableTime milliseconds have passed since receipt of the last
                 * reachability confirmation for a neighbor, the Neighbor Cache entrys
                 * state changes from REACHABLE to STALE.
                 */
                if((uint_32)(RTCS_time_get() - neighbor_entry->state_time ) > nd6_cfg->reachable_time)
                /* REACHABLE to STALE */
                { 
                    neighbor_entry->state = ND6_NEIGHBOR_STATE_STALE;
                }
                break;
           /* =========== DELAY ================ */    
           case ND6_NEIGHBOR_STATE_DELAY:     
                 /*
                 * If the entry is still in the DELAY state when the timer expires, 
                 * the entrys state changes to PROBE.
                 */
                if((uint_32)(RTCS_time_get() - neighbor_entry->state_time ) > ND6_DELAY_FIRST_PROBE_TIME)
                /* DELAY to PROBE */
                { 
                    in6_addr *src_addr;
                    in6_addr *dest_addr = (in6_addr *)&neighbor_entry->ip_addr;

                    
                    neighbor_entry->state = ND6_NEIGHBOR_STATE_PROBE;
                    neighbor_entry->solicitation_send_counter = 0;
                    /* 
                     * Upon entering the PROBE state, a node sends a unicast Neighbor
                     * Solicitation message to the neighbor using the cached link-layer
                     * address.
                     */

                    neighbor_entry->state_time = RTCS_time_get();
                     
                    /* Select source address.*/ //PFI
                    src_addr = (in6_addr *)ip6_if_select_src_addr(if_ptr, dest_addr);
                     
                    if(src_addr)
                    {
                        nd6_neighbor_solicitation_send(if_ptr, src_addr, dest_addr/*Unicast*/, dest_addr);
                    }
                }
                break;
            /* =========== PROBE ================ */    
            case ND6_NEIGHBOR_STATE_PROBE:  //PFI the same as INCOMPLETE
                /* RFC4861 7.3.3:
                 * While in the PROBE state, a node retransmits Neighbor
                 * Solicitation messages every RetransTimer milliseconds until
                 * reachability confirmation is obtained. Probes are retransmitted even
                 * if no additional packets are sent to the neighbor. If no response is
                 * received after waiting RetransTimer milliseconds after sending the
                 * MAX_UNICAST_SOLICIT solicitations, retransmissions cease and the
                 * entry SHOULD be deleted. Subsequent traffic to that neighbor will
                 *  recreate the entry and perform address resolution again.
                 */
                if((uint_32)(RTCS_time_get() - neighbor_entry->state_time ) > nd6_cfg->retrans_timer )
                {
                    neighbor_entry->solicitation_send_counter++;
                    if(neighbor_entry->solicitation_send_counter >= ND6_MAX_UNICAST_SOLICIT)
                    /* AR is failed.*/
                    {
                        /* RFC4861 7.3.3: If no response is
                         * received after waiting RetransTimer milliseconds after sending the
                         * MAX_UNICAST_SOLICIT solicitations
                         */                        
                        nd6_neighbor_cache_del(neighbor_entry);
                    }
                    else
                    /* NUD: transmit NS.*/
                    {
                        in6_addr *dest_addr = (in6_addr *)&neighbor_entry->ip_addr;
                        
                        /* Select source address.*/ //PFI
                        in6_addr *src_addr = (in6_addr *)ip6_if_select_src_addr(if_ptr, dest_addr);
                        if(src_addr)
                        {
                            neighbor_entry->state_time = RTCS_time_get();
                            dest_addr = (in6_addr *)&neighbor_entry->ip_addr;
                            nd6_neighbor_solicitation_send(if_ptr, src_addr, dest_addr/*Unicast*/, dest_addr);
                        }
                    }
                }              
                break;                    
            case ND6_NEIGHBOR_STATE_NOTUSED:
            default:
                break;
        }
    }
    
    
    /* Check lifetime for prefixes.*/
    for(i=1; i<ND6_PREFIX_LIST_SIZE; i++)
    {
        if((nd6_cfg->prefix_list[i].state != ND6_PREFIX_STATE_NOTUSED)
            && (nd6_cfg->prefix_list[i].lifetime != ND6_PREFIX_LIFETIME_INFINITE)
            && ((RTCS_time_get()/1000) > (nd6_cfg->prefix_list[i].creation_time + nd6_cfg->prefix_list[i].lifetime)) ) //PFI
             
        {
            nd6_prefix_list_del(&nd6_cfg->prefix_list[i]);
        }
    }
    
    /* Check lifetime for addresses.*/
    ip6_if_addr_timer(if_ptr);
        
    /* RD timer. */
    nd6_rd_timer(if_ptr);  
    
    /* Init next timer event.*/
    nd6_cfg->timer.TIME = ND6_TIMER_PERIOD;

    return TRUE;
}

/************************************************************************
* NAME: nd6_neighbor_cache_get
*
* DESCRIPTION: Get entry from Neighbor cache that corresponds ip_addr.
*               It returns NULL if no entry is found.
*************************************************************************/
ND6_NEIGHBOR_ENTRY_PTR nd6_neighbor_cache_get(IP_IF_PTR if_ptr, in6_addr *ip_addr)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_NEIGHBOR_ENTRY_PTR result = NULL;
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    if (nd6_cfg)
    {
        /* Find the entry in the cache. */
        for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
        {
            if( (nd6_cfg->neighbor_cache[i].state != ND6_NEIGHBOR_STATE_NOTUSED) &&
                IN6_ARE_ADDR_EQUAL(&nd6_cfg->neighbor_cache[i].ip_addr, ip_addr))
            {
                result = &nd6_cfg->neighbor_cache[i];
                break;
            }
        }
    }
    return result;
}

/************************************************************************
* NAME: nd6_neighbor_cache_add
*
* DESCRIPTION: Adds (TBD update) entry into the Neighbor cache.
*************************************************************************/
//PFI ll_addr_t =>*
ND6_NEIGHBOR_ENTRY_PTR nd6_neighbor_cache_add(IP_IF_PTR if_ptr, in6_addr *ip_addr, ll_addr_t ll_addr, nd6_neighbor_state_t state)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_NEIGHBOR_ENTRY_PTR entry = NULL;

   
    nd6_cfg = if_ptr->IP6_IF.ND6;
    if (nd6_cfg)
    {
        /* Find not used entry.*/
        for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
        {
            if( (nd6_cfg->neighbor_cache[i].state == ND6_NEIGHBOR_STATE_NOTUSED))
            {
                entry = &nd6_cfg->neighbor_cache[i];
                break;
            }
        }
        
        /* If no free entry is found.*/
        if(entry == NULL)
        { 
            entry = &nd6_cfg->neighbor_cache[0];
            /* Try to find the oldest entry.*/
            for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
            {
                if(nd6_cfg->neighbor_cache[i].creation_time < entry->creation_time)
                {
                    entry = &nd6_cfg->neighbor_cache[i];
                }
            }
        }
        
        /* Fill the informationn.*/
        
        _mem_zero(entry, sizeof(ND6_NEIGHBOR_ENTRY));    /* Clear enty structure.*/
        
        
        IN6_ADDR_COPY(ip_addr, &entry->ip_addr);
        if(ll_addr != NULL)
            LL_ADDR_COPY(ll_addr, entry->ll_addr, if_ptr->DEV_ADDRLEN);
        entry->creation_time = RTCS_time_get()/1000;
        entry->is_router = 0;
        entry->router_lifetime = 0;
        entry->state = state;
        
        //TBD Init timers reachable; last send
    }
    return entry;
}

/************************************************************************
* NAME: nd6_neighbor_cache_del
*
* DESCRIPTION: Deletes an entry from the Neighbor Cache.
*************************************************************************/
void nd6_neighbor_cache_del(ND6_NEIGHBOR_ENTRY_PTR neighbor_entry)
{
    if (neighbor_entry)
    {
        neighbor_entry->state = ND6_NEIGHBOR_STATE_NOTUSED;
        
        /* Free waiting queue.*/
        if(neighbor_entry->waiting_pcb) //TBD for bigger queue.
        {
            RTCSPCB_free(neighbor_entry->waiting_pcb); 
        }
    }
}

/************************************************************************
* NAME: nd6_neighbor_enqueue_pcb
*
* DESCRIPTION: Put PCB to the queue, waiting for address resolution to complete.
*************************************************************************/
void nd6_neighbor_enqueue_waiting_pcb(ND6_NEIGHBOR_ENTRY_PTR neighbor_entry, RTCSPCB_PTR  waiting_pcb)
{
    if (neighbor_entry && waiting_pcb)
    {
        
        /* When a queue  overflows, the new arrival SHOULD replace the oldest entry.*/
        //TBD for bigger queue.
        if(neighbor_entry->waiting_pcb != NULL)
        {
            RTCSPCB_free(neighbor_entry->waiting_pcb); /* Free the oldest one.*/
        }
        
        neighbor_entry->waiting_pcb = waiting_pcb;
    }
}

/************************************************************************
* NAME: nd6_neighbor_enqueue_pcb
*
* DESCRIPTION:  Sends waiting PCBs, if any.
*************************************************************************/
void nd6_neighbor_send_waiting_pcb(IP_IF_PTR if_ptr, ND6_NEIGHBOR_ENTRY_PTR neighbor_entry)
{
    if (neighbor_entry->waiting_pcb != NULL)
    {
        /* Send.*/
        if_ptr->DEVICE.SEND6(if_ptr, neighbor_entry->waiting_pcb, NULL /* not needed.*/, &neighbor_entry->ip_addr);
        
        neighbor_entry->waiting_pcb = NULL;
    }
}

/************************************************************************
* NAME: nd6_prefix_list_get
*
* DESCRIPTION: Get entry from Prefix List that corresponds to "prefix".
* It returns NULL if no entry is found.
*************************************************************************/
ND6_PREFIX_ENTRY_PTR nd6_prefix_list_get(IP_IF_PTR if_ptr, in6_addr *prefix)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_PREFIX_ENTRY_PTR result = NULL;
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    if (nd6_cfg)
    {
        /* Find the entry in the list.
        */
        for(i=0; i<ND6_PREFIX_LIST_SIZE; i++)
        {
            if( (nd6_cfg->prefix_list[i].state != ND6_PREFIX_STATE_NOTUSED) &&
                IN6_ARE_ADDR_EQUAL(&nd6_cfg->prefix_list[i].prefix, prefix))
            {
                result = &nd6_cfg->prefix_list[i];
                break;
            }
        }
    }
    return result;
}

/************************************************************************
* NAME: nd6_prefix_list_add
*
* DESCRIPTION: Adds (TBD update) entry into the Prefix List.
*************************************************************************/
ND6_PREFIX_ENTRY_PTR nd6_prefix_list_add(IP_IF_PTR if_ptr, in6_addr *prefix, uint_32 prefix_length, uint_32 lifetime)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_PREFIX_ENTRY_PTR entry = NULL;
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    if (nd6_cfg)
    {
        /* Find an unused entry in the cache. Skip 1st Link_locak prefix.
         */
        for(i=1; i<ND6_PREFIX_LIST_SIZE; i++)
        {
            if(nd6_cfg->prefix_list[i].state == ND6_PREFIX_STATE_NOTUSED)
            {
                entry = &nd6_cfg->prefix_list[i];
                break;
            }
        }
        
        
        /* If no free entry is found.
         */
        if(entry == NULL)
        { 
            entry = &nd6_cfg->prefix_list[1];
            /* Try to find the oldest entry.
             */
            for(i=1; i<ND6_PREFIX_LIST_SIZE; i++)
            {
                if(nd6_cfg->prefix_list[i].creation_time < entry->creation_time)
                {
                    entry = &nd6_cfg->prefix_list[i];
                }
            }
        }
        
        /* Fill the informationn.
         */
        IN6_ADDR_COPY(prefix, &entry->prefix);
        entry->prefix_length = prefix_length;
        entry->lifetime = lifetime;
        entry->creation_time = RTCS_time_get()/1000;
        entry->state = ND6_PREFIX_STATE_USED;
    }
    return entry;
}

/************************************************************************
* NAME: nd6_prefix_list_del
*
* DESCRIPTION: Deletes an entry from the Prefix List.
*************************************************************************/
void nd6_prefix_list_del(ND6_PREFIX_ENTRY_PTR prefix_entry)
{
    if (prefix_entry)
    {
        prefix_entry->state = ND6_PREFIX_STATE_NOTUSED;
    }
}

/************************************************************************
* NAME: nd6_addr_is_onlink
*
* DESCRIPTION: Checks if the address is on-link. 
*               Returns TRUE if it is on-line, FALSE otherwise.
*************************************************************************/
boolean nd6_addr_is_onlink(IP_IF_PTR if_ptr, in6_addr *addr)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    boolean result = FALSE;
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    if (nd6_cfg)
    {
        /* Find the entry in the list. */
        for(i=0; i<ND6_PREFIX_LIST_SIZE; i++)
        {
            if( (nd6_cfg->prefix_list[i].state != ND6_PREFIX_STATE_NOTUSED) &&
                (ip6_addr_pefix_cmp(&nd6_cfg->prefix_list[i].prefix, addr, nd6_cfg->prefix_list[i].prefix_length) == TRUE) )
            {
                result = TRUE;
                break;
            }
        }
    }
    return result;
}

/************************************************************************
* NAME: nd6_router_list_add
*
* DESCRIPTION: Adds entry into the Router List.
*************************************************************************/
void nd6_router_list_add( ND6_NEIGHBOR_ENTRY_PTR neighbor_entry, uint_32 lifetime)
{
    if (neighbor_entry)
    {
        if(lifetime) 
        {
            neighbor_entry->is_router = 1;
            neighbor_entry->router_lifetime = lifetime;
            neighbor_entry->creation_time = RTCS_time_get()/1000;
        }
        else
        /* If the address is already present in the hosts Default Router
         * List and the received Router Lifetime value is zero, immediately
         * time-out the entry.
         */
        {
            nd6_router_list_del(neighbor_entry);
        }
    }

}

/************************************************************************
* NAME: nd6_router_list_del
*
* DESCRIPTION: Deletes an entry from the Router List.
*************************************************************************/
void nd6_router_list_del( ND6_NEIGHBOR_ENTRY_PTR neighbor_entry)
{
    if (neighbor_entry)
    {
        neighbor_entry->is_router = 0;
        neighbor_entry->router_lifetime = 0;
    }
}

/************************************************************************
* NAME: nd6_default_router_get
*
* DESCRIPTION: Chooses default router. Returns NULL if no router exists.
*************************************************************************/
ND6_NEIGHBOR_ENTRY_PTR nd6_default_router_get(IP_IF_PTR if_ptr)
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_NEIGHBOR_ENTRY_PTR result = NULL;
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    
    if (nd6_cfg)
    {
        /* Find the best router. */
        for(i=0; i < ND6_NEIGHBOR_CACHE_SIZE; i++)
        {
            if( (nd6_cfg->neighbor_cache[i].state != ND6_NEIGHBOR_STATE_NOTUSED) && (nd6_cfg->neighbor_cache[i].is_router == 1) )
            {
                result = &nd6_cfg->neighbor_cache[i];
                /* Routers that are reachable or probably reachable (i.e., in any
                 * state other than INCOMPLETE) SHOULD be preferred over routers
                 * whose reachability is unknown or suspect.
                 */
                if(nd6_cfg->neighbor_cache[i].state != ND6_NEIGHBOR_STATE_INCOMPLETE)
                {
                    break;
                }
            }
        }

    }
    
    /* //TBD 
     * RFFC4861: When no routers on the list are known to be reachable or
     * probably reachable, routers SHOULD be selected in a round-robin
     * fashion, so that subsequent requests for a default router do not
     * return the same router until all other routers have been
     * selected.
     * Cycling through the router list in this case ensures that all
     * available routers are actively probed by the Neighbor
     * Unreachability Detection algorithm. A request for a default
     * router is made in conjunction with the sending of a packet to a
     * router, and the selected router will be probed for reachability
     * as a side effect.
     */
    
    return result;
}


/************************************************************************
* NAME: nd6_neighbor_solicitation_receive
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Handles received Neighbor Solicitation message.
*************************************************************************/
uint_32 nd6_neighbor_solicitation_receive(RTCSPCB_PTR rtcs_pcb)
{
    ICMP6_HEADER_PTR        icmp6_packet = (ICMP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    uint_32                 icmp6_packet_size = RTCSPCB_SIZE(rtcs_pcb);
    ND6_NS_HEADER_PTR     ns_packet = (ND6_NS_HEADER_PTR)icmp6_packet;
    ND6_OPTION_LLA_HEADER_PTR nd_option_slla = NULL;
    uint_32                 nd_option_offset;
    IP6_IF_ADDR_INFO_PTR    target_if_addr_info;
    IP6_HEADER_PTR          ip6_packet = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(rtcs_pcb);

    IP_IF_PTR               if_ptr = rtcs_pcb->IFSRC;
    int                     result = RTCS_OK;

#if RTCSCFG_DEBUG_MSG
    printf("ND6:NS Received \n");
#endif    
	
    /************************************************************
    * Validation.
    *************************************************************/	
    if(
        (icmp6_packet_size < sizeof(ND6_NS_HEADER))                 /* Make sure that sizeof(ICMP6_NS_HEADER) <= RTCSPCB_SIZE(pcb).*/
        /* Validation RFC4861 (7.1.1).*/
        ||(ip6_packet->HOP_LIMIT[0] != ND6_HOP_LIMIT)               /* The IP Hop Limit field has a value of 255.*/
        ||(icmp6_packet->CODE[0] != 0)                              /* ICMP Code is 0.*/ 
        ||IN6_IS_ADDR_MULTICAST((in6_addr*)ns_packet->TARGET_ADDR)  /* Target Address is not a multicast address. */
        ||ip6_if_is_my_addr((_rtcs_if_handle) if_ptr, (in6_addr*)ip6_packet->SOURCE) /* Duplicate IP address check. */
        )
    { 		
        result = RTCSERR_ICMP_BAD_HEADER;
    }
    else
    {   /* Validation passed.
         */
     
        /* Handle posible options.
         */
        nd_option_offset = sizeof(ND6_NS_HEADER);
        while(icmp6_packet_size > nd_option_offset + sizeof(ND6_OPTION_HEADER)) 
        {
            ND6_OPTION_HEADER_PTR nd_option;
    				    
            nd_option =  (ND6_OPTION_HEADER_PTR) ((uint_8 *)icmp6_packet + nd_option_offset) ;
            /* Validation RFC4861 (7.1.1). All included options have a length that is greater than zero.
             */
            if(nd_option->LENGTH[0] == 0)
            { 		
                result = RTCSERR_ICMP_BAD_HEADER;
                goto DROP;
            }
            
            /* Handle Source link-layer address option only.
             */
            if((nd_option->TYPE[0] == ND6_OPTION_SOURCE_LLA)
    			   && ( ((nd_option->LENGTH[0] << 3) - sizeof(ND6_OPTION_HEADER)) >= if_ptr->DEV_ADDRLEN) )
            {
                nd_option_slla = (ND6_OPTION_LLA_HEADER_PTR)nd_option; /* Source Link-layer Address option is found.*/
            }
            /* else, silently ignore any options they do not recognize
             * and continue processing the message.
             */
    				    
            nd_option_offset += (nd_option->LENGTH[0] << 3);
        }

        if(nd_option_slla  != NULL)
        {
            /* Validation RFC4861 (7.1.1): If the IP source address is the 
             * unspecified address, there is no source link-layer address option in the message.
             * NOTE: RFC4861 (7):Duplicate Address Detection sends Neighbor Solicitation
             * messages with an unspecified source address targeting its own
             * "tentative" address.
             */
            if(IN6_IS_ADDR_UNSPECIFIED((in6_addr*)ip6_packet->SOURCE))
            {
                /* If the Source Address is the unspecified address, the node MUST NOT
                 * create or update the Neighbor Cache entry.
                 */
                result = RTCSERR_ICMP_BAD_HEADER;
                goto DROP;
            }
            else
            {
                ND6_NEIGHBOR_ENTRY_PTR neighbor_cache_entry;
                            
                /* RFC 48617.2.3: the recipient
                 * SHOULD create or update the Neighbor Cache entry for the IP Source
                 * Address of the solicitation. 
                 */
                neighbor_cache_entry = nd6_neighbor_cache_get(if_ptr, (in6_addr*)ip6_packet->SOURCE);
                if(neighbor_cache_entry == NULL)
                {
                    /* If an entry does not already exist, the
                     * node SHOULD create a new one and set its reachability state to STALE
                     * as specified in Section 7.3.3.
                     * If a Neighbor Cache entry is created, the IsRouter flag SHOULD be set
                     * to FALSE. This will be the case even if the Neighbor Solicitation is
                     * sent by a router since the Neighbor Solicitation messages do not
                     * contain an indication of whether or not the sender is a router. In
                     * the event that the sender is a router, subsequent Neighbor
                     * Advertisement or Router Advertisement messages will set the correct
                     * IsRouter value. 
                     */
                     nd6_neighbor_cache_add(if_ptr, (in6_addr*)ip6_packet->SOURCE, nd_option_slla->addr, ND6_NEIGHBOR_STATE_STALE);
                }
                else
                {
                    /* If an entry already exists, and the
                     * cached link-layer address differs from the one in the received Source
                     * Link-Layer option, the cached address should be replaced by the
                     * received address, and the entrys reachability state MUST be set to
                     * STALE.
                     * If a Neighbor Cache entry already exists, its
                     * IsRouter flag MUST NOT be modified.
                     */
                    if(!LL_ADDR_ARE_EQUAL(nd_option_slla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN))
                    {
                        LL_ADDR_COPY(nd_option_slla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN);
                        neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;
                    }
                    else
                    {
                        /* RFC4861: Appendix C */
                        if(neighbor_cache_entry->state == ND6_NEIGHBOR_STATE_INCOMPLETE)
                            neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;
                    }
                    
                    
                    /* - It sends any packets queued for the neighbor awaiting address
                     *   resolution.*/
                    nd6_neighbor_send_waiting_pcb(if_ptr, neighbor_cache_entry);
                    
                }
            }
        }
                    
        /* Get Target Address Info, according tp Target Address of NS message.
         */
        target_if_addr_info = ip6_if_get_addr_info(if_ptr, (in6_addr*)ns_packet->TARGET_ADDR);
                    
        /* Check if we are the target.
         */
        if(target_if_addr_info != NULL)
        {
            /* Duplicate Address Detection (DAD) sends Neighbor Solicitation
             * messages with an unspecified source address targeting its own
             * "tentative" address.
             */
            if(IN6_IS_ADDR_UNSPECIFIED((in6_addr*)ip6_packet->SOURCE))
            { /* === Duplicate Address Detection ===*/

                /* IP Destination address must be our solitation address.
                 */
                if(ip6_if_is_my_solicited_multicast_addr(if_ptr, (in6_addr*)ip6_packet->DEST))
                {
                    if(target_if_addr_info->ip_addr_state != IP6_ADDR_STATE_TENTATIVE)
                    {
                        /* If the source of the solicitation is the unspecified address, the
                         * node MUST set the Solicited flag to zero and multicast the
                         * advertisement to the all-nodes address.
                         */
                        nd6_neighbor_advertisement_send(if_ptr, &target_if_addr_info->ip_addr /*ipsrc*/, (in6_addr *)&in6addr_linklocal_allnodes/*ipdest*/, ND6_NA_FLAG_OVERRIDE);
                    }
                    else
                    {
                        nd6_dad_failed(if_ptr, target_if_addr_info); /* => DAD is failed. */
                        /* UST be silently discarded, if the Target Address is a "tentative" 
                         * address on which Duplicate Address Detection is being performed [ADDRCONF].*/
                    }

                }
                /* else DROP IT.
                 */
            }
            else if(    /* === Address Resolution === */
                    ip6_if_is_my_solicited_multicast_addr(if_ptr, (in6_addr*)ip6_packet->DEST) 
                    ||  /* === Neighbor Unreachability Detection === */
                    ((IN6_ARE_ADDR_EQUAL((in6_addr*)ns_packet->TARGET_ADDR, (in6_addr*)ip6_packet->DEST))) )
            { 
                /*
                 * Sends a Neighbor Advertisement response.
                 */ 
                nd6_neighbor_advertisement_send(if_ptr, &target_if_addr_info->ip_addr /*ipsrc*/, (in6_addr *)ip6_packet->SOURCE/*ipdest*/, ND6_NA_FLAG_SOLICITED | ND6_NA_FLAG_OVERRIDE);
            }
            else
            { /* Bad packet.*/
                result = RTCSERR_ICMP_BAD_HEADER;
            }
        }
        /* else  Not for us => Discarded.
         */
    }
DROP:    
    return result;    
}

/************************************************************************
* NAME: nd6_neighbor_advertisement_receive
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Handles received Neighbor Advertisement message.
*************************************************************************/
uint_32 nd6_neighbor_advertisement_receive(RTCSPCB_PTR rtcs_pcb)
{
    ICMP6_HEADER_PTR        icmp6_packet = (ICMP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    uint_32                 icmp6_packet_size = RTCSPCB_SIZE(rtcs_pcb);
    ND6_NA_HEADER_PTR     na_packet = (ND6_NA_HEADER_PTR)icmp6_packet;
    ND6_OPTION_LLA_HEADER_PTR nd_option_tlla = NULL;
    uint_32                 nd_option_offset;
    IP6_IF_ADDR_INFO_PTR    target_if_addr_info;
    IP6_HEADER_PTR          ip6_packet = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(rtcs_pcb);

    IP_IF_PTR               if_ptr = rtcs_pcb->IFSRC;
    int                     result = RTCS_OK;
    ND6_NEIGHBOR_ENTRY_PTR  neighbor_cache_entry;
    int is_solicited;
    int is_router;
    int is_override;
    int is_ll_addr_changed;;

	
    /************************************************************
    * Validation.
    *************************************************************/	
    if(
        (icmp6_packet_size < sizeof(ND6_NA_HEADER))               /* Make sure that sizeof(ICMP6_NA_HEADER) <= RTCSPCB_SIZE(pcb).*/
        /* Validation RFC4861 (7.1.2).*/
        ||(ip6_packet->HOP_LIMIT[0] != ND6_HOP_LIMIT)               /* The IP Hop Limit field has a value of 255.*/
        ||(icmp6_packet->CODE[0] != 0)                              /* ICMP Code is 0.*/ 
        ||IN6_IS_ADDR_MULTICAST((in6_addr*)na_packet->TARGET_ADDR)  /* Target Address is not a multicast address. */
        ||( ((is_solicited=(na_packet->FLAG[0] & ND6_NA_FLAG_SOLICITED)) != 0)
            && IN6_IS_ADDR_MULTICAST((in6_addr*)ip6_packet->DEST) ) /* If the IP Destination Address is a multicast 
                                                                     * address the Solicited flag is zero.*/
        )
    { 		
        result = RTCSERR_ICMP_BAD_HEADER;
    }
    else
    {   
        
        /************************************************************
         * Handle posible options.
         ************************************************************/
        nd_option_offset = sizeof(ND6_NA_HEADER);
        while(icmp6_packet_size > nd_option_offset + sizeof(ND6_OPTION_HEADER)) 
        {
            ND6_OPTION_HEADER_PTR nd_option;
    				    
            nd_option =  (ND6_OPTION_HEADER_PTR) ((uint_8 *)icmp6_packet + nd_option_offset) ;
            /* Validation RFC4861 (7.1.2). All included options have a length that is greater than zero.
             */
            if(nd_option->LENGTH[0] == 0)
            { 		
                result = RTCSERR_ICMP_BAD_HEADER;
                goto DROP;
            }
            
            /* Handle Target Link-Layer Address option only.
             */
            if((nd_option->TYPE[0] == ND6_OPTION_TARGET_LLA)
    			   && ( ((nd_option->LENGTH[0] << 3) - sizeof(ND6_OPTION_HEADER)) >= if_ptr->DEV_ADDRLEN) )
            {
                nd_option_tlla = (ND6_OPTION_LLA_HEADER_PTR)nd_option; /* Target Link-layer Address option is found.*/
            }
            /* else, silently ignore any options they do not recognize
             * and continue processing the message.
             */
    				    
            nd_option_offset += (nd_option->LENGTH[0] << 3);
        }
        
        /* Get Target Address Info, according to Target Address of NA message.
        */
        target_if_addr_info = ip6_if_get_addr_info(if_ptr, (in6_addr*)na_packet->TARGET_ADDR);
        if(target_if_addr_info != NULL)
        {
            /* Duplicated address!!!!! */
            if(target_if_addr_info->ip_addr_state == IP6_ADDR_STATE_TENTATIVE) 
                nd6_dad_failed(if_ptr, target_if_addr_info); /* => DAD is failed. */
            
            goto DROP;
        }

        /* A Neighbor Advertisements that passes the validity checks is called a
         * "valid advertisement".
         */
        
        /* Extract flag values.*/
        is_router = (na_packet->FLAG[0] & ND6_NA_FLAG_ROUTER)? 1:0;
        is_override = na_packet->FLAG[0] & ND6_NA_FLAG_OVERRIDE;
        
        
        /************************************************************
         * Handle NA message.
         ************************************************************/
        
        /* RFC4861 7.2.5: Neighbor Cache is searched for the targets entry.
         */
        neighbor_cache_entry = nd6_neighbor_cache_get(if_ptr, (in6_addr*)ip6_packet->SOURCE);
        if(neighbor_cache_entry == NULL)
        {
            /* If no entry exists, the advertisement SHOULD be silently discarded.
             */
            goto DROP;
        }
        
        /* If the targets Neighbor Cache entry is in the INCOMPLETE state.*/
        if(neighbor_cache_entry->state == ND6_NEIGHBOR_STATE_INCOMPLETE)
        {
            /* If the link layer has addresses and no Target Link-Layer Address option is
             * included, the receiving node SHOULD silently discard the received
             * advertisement.*/
            if(nd_option_tlla == NULL)
            {
                goto DROP;
            }
            /* Otherwise, the receiving node performs the following
             * steps:
             * - It records the link-layer address in the Neighbor Cache entry.
             */
            LL_ADDR_COPY(nd_option_tlla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN);
              
            /* - If the advertisements Solicited flag is set, the state of the
             *   entry is set to REACHABLE; otherwise, it is set to STALE.
             */
            if (is_solicited)
            {
                neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_REACHABLE;
                /* Reset Reachable Timer. */
                neighbor_cache_entry->state_time = RTCS_time_get();
            }
            else
            {
                neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;            
            }
              
            /* - It sets the IsRouter flag in the cache entry based on the Router
             *   flag in the received advertisement.
             */
            neighbor_cache_entry->is_router = is_router;
             
            /* - It sends any packets queued for the neighbor awaiting address
             *   resolution.
             */ 
             nd6_neighbor_send_waiting_pcb(if_ptr, neighbor_cache_entry);
        }
        else
        {
            /* If the targets Neighbor Cache entry is in any state other than
             * INCOMPLETE.
             */
             if( nd_option_tlla )
             {
                /* If supplied link-layer address differs. */
                is_ll_addr_changed = !LL_ADDR_ARE_EQUAL(nd_option_tlla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN);
             }
             else
             {
                is_ll_addr_changed = 0;  
             }
             
             /* I. If the Override flag is clear and the supplied link-layer address
              *     differs from that in the cache.
              */
             if((is_override == 0) && is_ll_addr_changed)
             {
                /* a. If the state of the entry is REACHABLE, set it to STALE, but
                 *    do not update the entry in any other way.
                 */
                 if(neighbor_cache_entry->state == ND6_NEIGHBOR_STATE_REACHABLE) 
                 {
                    neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;
                 }
                /* b. Otherwise, the received advertisement should be ignored and
                 *    MUST NOT update the cache.
                 */
                 goto DROP;
             }
             /* II. If the Override flag is set, or the supplied link-layer address
              *     is the same as that in the cache, or no Target Link-Layer Address
              *     option was supplied, the received advertisement MUST update the
              *     Neighbor Cache entry as follows:
              */
             else
             {
                /* - The link-layer address in the Target Link-Layer Address option
                 *   MUST be inserted in the cache (if one is supplied and differs
                 *   from the already recorded address).
                 */
                if(nd_option_tlla)
                    LL_ADDR_COPY(nd_option_tlla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN);
                
                /* - If the Solicited flag is set, the state of the entry MUST be
                 *   set to REACHABLE. 
                 */
                if(is_solicited) 
                {
                    neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_REACHABLE;
                    /* Reset Reachable Timer.*/
                    neighbor_cache_entry->state_time = RTCS_time_get();
                }
                /* If the Solicited flag is zero and the linklayer
                 *   address was updated with a different address, the state
                 *   MUST be set to STALE. Otherwise, the entrys state remains
                 *   unchanged.
                 */
                else if(is_ll_addr_changed)
                {
                    neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;
                }
                /* - The IsRouter flag in the cache entry MUST be set based on the
                 *   Router flag in the received advertisement. In those cases
                 *   where the IsRouter flag changes from TRUE to FALSE as a result
                 *   of this update, the node MUST remove that router from the
                 *   Default Router List and update the Destination Cache entries
                 *   for all destinations using that neighbor as a router as
                 *   specified in Section 7.3.3.
                 */
                if((neighbor_cache_entry->is_router) && (is_router == 0))
                {
                    nd6_router_list_del(neighbor_cache_entry);
                }                 
            }
        }
    }
    
DROP:    
    return result;    
}

/************************************************************************
* NAME: nd6_neighbor_advertisement_send
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Sends an Neighbor Advertisement message.
*************************************************************************/
uint_32 nd6_neighbor_advertisement_send(IP_IF_PTR if_ptr, in6_addr *ipsrc, in6_addr *ipdest, uint_8 flags)
{
    uint_32                 na_packet_size;
    RTCSPCB_PTR             na_pcb;
    ND6_NA_HEADER_PTR       na_packet;
    ND6_OPTION_LLA_HEADER_PTR nd_option_tlla;
    uint_32                 result = (uint_32)RTCS_ERROR;
	                                   
    na_pcb = RTCSPCB_alloc_send();
                                   
    if (na_pcb)
    {
        na_packet_size = sizeof(ND6_NA_HEADER) + sizeof(ND6_OPTION_HEADER) + if_ptr->DEV_ADDRLEN;

        result = RTCSPCB_insert_header(na_pcb, na_packet_size);
        if (result)
        {
            RTCSLOG_PCB_FREE(na_pcb, result);
            RTCSPCB_free(na_pcb);
        }
        else
        {
            /* Fill ICMP Header */
            na_packet = (ND6_NA_HEADER_PTR)RTCSPCB_DATA(na_pcb);
                                              
            htonc(na_packet->HEAD.TYPE, ICMP6_TYPE_NEIGHBOR_ADVERTISEMENT);
            htonc(na_packet->HEAD.CODE, 0);
                                              
            /* NA header.*/
            na_packet->FLAG[0] = flags;    /* Flag parameter.*/
            _mem_zero(na_packet->RESERVED, sizeof(na_packet->RESERVED));    /* Set to zeros the reserved field.*/
            IN6_ADDR_COPY(ipsrc, (in6_addr *)na_packet->TARGET_ADDR);       /* Set NA target address, the same as for NS.*/ //TBD ???
                                              
            /* Fill Target Link-Layer Address option.*/
            nd_option_tlla = (ND6_OPTION_LLA_HEADER_PTR)((uint_32)na_packet + sizeof(ND6_NA_HEADER));
            nd_option_tlla->HEAD.TYPE[0] = ND6_OPTION_TARGET_LLA; /* Type. */
            nd_option_tlla->HEAD.LENGTH[0] = (if_ptr->DEV_ADDRLEN + sizeof(ND6_OPTION_HEADER))>>3; /* Option size devided by 8,*/
            LL_ADDR_COPY(if_ptr->DEV_ADDR, nd_option_tlla->addr, if_ptr->DEV_ADDRLEN); /* Link-Layer Target address.*/
                                              
            /*
             * Calculate the checksum for the packet now, and let IPv6
             * add the IPv6 pseudo checksum later.
             */
            htons(na_packet->HEAD.CHECKSUM, 0);
            na_pcb->IP_SUM     = IP_Sum_PCB(RTCSPCB_SIZE(na_pcb), na_pcb);
            na_pcb->IP_SUM_PTR = na_packet->HEAD.CHECKSUM;
                                              
                                              
            //TBD Padd if needed.
                                              
            result = IP6_send(na_pcb, IPPROTO_ICMPV6|IPTTL(ND6_HOP_LIMIT), ipsrc, ipdest, if_ptr, 0);
        }
    }
    
    return result;
}

/************************************************************************
* NAME: nd6_neighbor_solicitation_send
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Sends an Neighbor Solicitation Message.
*   Nodes send Neighbor Solicitations to request the link-layer address
*   of a target node while also providing their own link-layer address to
*   the target.
*************************************************************************/
uint_32 nd6_neighbor_solicitation_send(IP_IF_PTR if_ptr /*MUST*/, in6_addr *ipsrc /* NULL for, DAD */, in6_addr *ipdest /*set for NUD,  NULL for DAD & AR */, in6_addr *target_addr)
{
    uint_32                 ns_packet_size;
    RTCSPCB_PTR             ns_pcb;
    ND6_NS_HEADER_PTR     ns_packet;
    ND6_OPTION_LLA_HEADER_PTR nd_option_slla;
    uint_32                 result = (uint_32)RTCS_ERROR;
    in6_addr                _ip_dest;
                                    
    ns_pcb = RTCSPCB_alloc_send();
                                   
    if (ns_pcb)
    {
        /* Check if, this is DAD.
         * Duplicate Address Detection sends Neighbor Solicitation
         * messages with an unspecified source address targeting its own
         * "tentative" address and without SLLAO.*/
        ns_packet_size = sizeof(ND6_NS_HEADER) + ((ipsrc == NULL /* DAD */) ? 0:(sizeof(ND6_OPTION_HEADER) + if_ptr->DEV_ADDRLEN));

        result = RTCSPCB_insert_header(ns_pcb, ns_packet_size);
        if (result != RTCS_OK)
        {
            goto FAILED;
        }
        else
        {   
            /*
             * Neighbor Solicitations are multicast when the node needs
             * to resolve an address and unicast when the node seeks to verify the
             * reachability of a neighbor.
             */
            if(ipdest == NULL)
            { /* AR, DAD */
                /* Generate Solicited Multicast destination address from the target address.*/
                ip6_get_solicited_multicast_addr(target_addr, &_ip_dest);
                ipdest = &_ip_dest;
            }
           /* else  NUD */
            
            /* Fill ICMP Header */
            ns_packet = (ND6_NS_HEADER_PTR)RTCSPCB_DATA(ns_pcb); 
            htonc(ns_packet->HEAD.TYPE, ICMP6_TYPE_NEIGHBOR_SOLICITATION);
            htonc(ns_packet->HEAD.CODE, 0);
             
            /* Fill NS Header.*/
            _mem_zero(ns_packet->RESERVED, sizeof(ns_packet->RESERVED));    /* Set to zeros the reserved field.*/
            IN6_ADDR_COPY(target_addr, (in6_addr *)ns_packet->TARGET_ADDR); /* Set NS target address.*/ 

            /*
             * RFC4861: The link-layer address for the sender. MUST NOT be
             * included when the source IP address is the
             * unspecified address (DAD). Otherwise, on link layers
             * that have addresses this option MUST be included in
             * multicast solicitations and SHOULD be included in
             * unicast solicitations.
             */
            if(ipsrc != NULL /* AR or NUD */)
            {
                /* RFC4861 7.2.2: If the source address of the packet prompting the solicitation is the
                 * same as one of the addresses assigned to the outgoing interface, that
                 * address SHOULD be placed in the IP Source Address of the outgoing
                 * solicitation. Otherwise, any one of the addresses assigned to the
                 * interface should be used.
                 */
                if(ip6_if_is_my_addr(if_ptr, ipsrc) == FALSE)
                {
                    ipsrc = (in6_addr *)ip6_if_select_src_addr(if_ptr, ipdest);
                    
                    if(ipsrc == NULL)
                        goto FAILED; /* Just in case. Should never happen.*/
                }
                                            
                /* Fill Source link-layer address option.*/
                nd_option_slla = (ND6_OPTION_LLA_HEADER_PTR)((uint_32)ns_packet + sizeof(ND6_NS_HEADER));
                nd_option_slla->HEAD.TYPE[0] = ND6_OPTION_SOURCE_LLA; /* Type. */
                nd_option_slla->HEAD.LENGTH[0] = (if_ptr->DEV_ADDRLEN + sizeof(ND6_OPTION_HEADER))>>3; /* Option size devided by 8,*/
                LL_ADDR_COPY(if_ptr->DEV_ADDR, nd_option_slla->addr, if_ptr->DEV_ADDRLEN); /* Source Link-Layer address.*/
                
                //TBD Padd if needed.%8
            }
            else
            {
                /* Source IP address is the
                 * unspecified address for DAD.
                 */
                 ipsrc = (struct in6_addr *)&in6addr_any;
            }                                              
            
            /*
             * Calculate the checksum for the packet now, and let IPv6
             * add the IPv6 pseudo checksum later.
             */
            htons(ns_packet->HEAD.CHECKSUM, 0);
            ns_pcb->IP_SUM     = IP_Sum_PCB(RTCSPCB_SIZE(ns_pcb), ns_pcb);
            ns_pcb->IP_SUM_PTR = ns_packet->HEAD.CHECKSUM;
                                              
            result = IP6_send(ns_pcb, IPPROTO_ICMPV6|IPTTL(ND6_HOP_LIMIT), ipsrc, ipdest, if_ptr, 0);
        }
    }
    return result;
    
FAILED:
    RTCSLOG_PCB_FREE(ns_pcb, result);
    RTCSPCB_free(ns_pcb);
    return result;
}

/************************************************************************
* NAME: nd6_router_solicitation_send
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Sends an Router Solicitation Message.
*   Nodes send Neighbor Solicitations to request the link-layer address
*   of a target node while also providing their own link-layer address to
*   the target.
*************************************************************************/
uint_32 nd6_router_solicitation_send(IP_IF_PTR if_ptr /*MUST*/) 
{
    uint_32             rs_packet_size;
    RTCSPCB_PTR         rs_pcb;
    ND6_RS_HEADER_PTR   rs_packet;
    ND6_OPTION_LLA_HEADER_PTR nd_option_slla;
    uint_32             result = (uint_32)RTCS_ERROR;
    in6_addr *ip_dest;
    in6_addr *ip_src;
                                    
    rs_pcb = RTCSPCB_alloc_send();
    
                                 
    if (rs_pcb)
    {
        
        /* Destination Address is the all-routers multicast address.*/
        ip_dest = (in6_addr *)&in6addr_linklocal_allrouters;
        /* Choose source address.*/
        ip_src = (in6_addr *)ip6_if_select_src_addr(if_ptr, ip_dest);

        rs_packet_size = sizeof(ND6_RS_HEADER) + ((ip_src == NULL /* no address */) ? 0:(sizeof(ND6_OPTION_HEADER) + if_ptr->DEV_ADDRLEN));
 
        result = RTCSPCB_insert_header(rs_pcb, rs_packet_size);
        if (result)
        {
            RTCSLOG_PCB_FREE(rs_pcb, result);
            RTCSPCB_free(rs_pcb);
        }
        else
        {   
        #if RTCSCFG_DEBUG_MSG
            printf("ND6: Send RS\n");
        #endif            
            
            /* Fill ICMP Header */
            rs_packet = (ND6_RS_HEADER_PTR)RTCSPCB_DATA(rs_pcb); 
            htonc(rs_packet->HEAD.TYPE, ICMP6_TYPE_ROUTER_SOLICITATION);
            htonc(rs_packet->HEAD.CODE, 0);
             
            /* Fill RS Header.*/
            _mem_zero(rs_packet->RESERVED, sizeof(rs_packet->RESERVED));    /* Set to zeros the reserved field.*/

            /*
             * RFC4861: The link-layer address of the sender, if
             * known. MUST NOT be included if the Source Address
             * is the unspecified address. Otherwise, it SHOULD
             * be included on link layers that have addresses.
             */
            if(ip_src != NULL )
            {
                                           
                /* Fill Source link-layer address option.*/
                nd_option_slla = (ND6_OPTION_LLA_HEADER_PTR)((uint_32)rs_packet + sizeof(ND6_RS_HEADER));
                nd_option_slla->HEAD.TYPE[0] = ND6_OPTION_SOURCE_LLA;   /* Type. */
                nd_option_slla->HEAD.LENGTH[0] = (if_ptr->DEV_ADDRLEN + sizeof(ND6_OPTION_HEADER))>>3;  /* Option size devided by 8,*/
                LL_ADDR_COPY(if_ptr->DEV_ADDR, nd_option_slla->addr, if_ptr->DEV_ADDRLEN);  /* Source Link-Layer address.*/
                
                //TBD Padd if needed.%8
            }
            else
            {
                /* Source IP address is the
                 * unspecified address.
                 */
                 ip_src = (struct in6_addr *)&in6addr_any;
            }                                              
            
            /*
             * Calculate the checksum for the packet now, and let IPv6
             * add the IPv6 pseudo checksum later.
             */
            htons(rs_packet->HEAD.CHECKSUM, 0);
            rs_pcb->IP_SUM     = IP_Sum_PCB(RTCSPCB_SIZE(rs_pcb), rs_pcb);
            rs_pcb->IP_SUM_PTR = rs_packet->HEAD.CHECKSUM;
                                              
            result = IP6_send(rs_pcb, IPPROTO_ICMPV6|IPTTL(ND6_HOP_LIMIT), ip_src, ip_dest, if_ptr, 0);
        }
    }
    
    return result;
}

/************************************************************************
* NAME: nd6_router_advertisement_receive
* RETURS: RTCS_OK or error code.
* DESCRIPTION: Handles received Router Advertisement message.
*************************************************************************/
uint_32 nd6_router_advertisement_receive(RTCSPCB_PTR rtcs_pcb)
{
    ICMP6_HEADER_PTR        icmp6_packet = (ICMP6_HEADER_PTR)RTCSPCB_DATA(rtcs_pcb);
    uint_32                 icmp6_packet_size = RTCSPCB_SIZE(rtcs_pcb);
    ND6_RA_HEADER_PTR       ra_packet = (ND6_RA_HEADER_PTR)icmp6_packet;
    ND6_OPTION_LLA_HEADER_PTR       nd_option_slla = NULL;
    ND6_OPTION_MTU_HEADER_PTR       nd_option_mtu = NULL;
    ND6_OPTION_PREFIX_HEADER_PTR    nd_option_prefix[RTCSCFG_ND6_PREFIX_LIST_SIZE];
    int                     prefix_index;
    uint_32                 nd_option_offset;
    IP6_HEADER_PTR          ip6_packet = (IP6_HEADER_PTR)RTCSPCB_DATA_NETWORK(rtcs_pcb);

    IP_IF_PTR               if_ptr = rtcs_pcb->IFSRC;
    int                     result = RTCS_OK;
    int                     i;
    ND6_NEIGHBOR_ENTRY_PTR  neighbor_cache_entry;

#if RTCSCFG_DEBUG_MSG    
    printf("ND6: RA Rx.\n");
#endif    
	
    /************************************************************
    * Validation of Router Advertisement Message.
    *************************************************************/	
    if(
        (icmp6_packet_size < sizeof(ND6_RA_HEADER))                 /* Make sure that sizeof(ICMP6_RA_HEADER) <= RTCSPCB_SIZE(pcb).*/
        /* Validation RFC4861 (6.1.2).*/
        ||(ip6_packet->HOP_LIMIT[0] != ND6_HOP_LIMIT)               /* The IP Hop Limit field has a value of 255.*/
        ||(icmp6_packet->CODE[0] != 0)                              /* ICMP Code is 0.*/ 
        )
    { 		
        result = RTCSERR_ICMP_BAD_HEADER;
    }
    else
    {   
        /************************************************************
         * Handle posible options.
         ************************************************************
         * The contents of any defined options that are not specified
         * to be used with Router Advertisement messages MUST be
         * ignored and the packet processed as normal. The only defined 
         * options that may appear are the Source Link-Layer Address, 
         * Prefix Information and MTU options.
         ************************************************************/
        nd_option_offset = sizeof(ND6_RA_HEADER);
        prefix_index = 0;
        while(icmp6_packet_size > nd_option_offset + sizeof(ND6_OPTION_HEADER)) 
        {
            ND6_OPTION_HEADER_PTR nd_option;
    				    
            nd_option =  (ND6_OPTION_HEADER_PTR) ((uint_8 *)icmp6_packet + nd_option_offset) ;
            /* Validation RFC4861 (6.1.2). All included options have a length that is greater than zero.
             */
            if(ntohc(nd_option->LENGTH) == 0)
            { 		
                result = RTCSERR_ICMP_BAD_HEADER;
                goto DROP;
            }
            
            /* The advertisement is scanned for valid options.
             */
            
            /* Source Link-Layer Address option.*/
            if( (ntohc(nd_option->TYPE) == ND6_OPTION_SOURCE_LLA)
    			   && ( ((ntohc(nd_option->LENGTH) << 3) - sizeof(ND6_OPTION_HEADER)) >= if_ptr->DEV_ADDRLEN) )
            {
                nd_option_slla = (ND6_OPTION_LLA_HEADER_PTR)nd_option; /* Target Link-layer Address option is found.*/
            }
            /* MTU option */
            else if( (ntohc(nd_option->TYPE) == ND6_OPTION_MTU)
    			   && ((ntohc(nd_option->LENGTH) << 3) >= sizeof(ND6_OPTION_MTU_HEADER)) )
            {
                 nd_option_mtu = (ND6_OPTION_MTU_HEADER_PTR)nd_option; /* MTU option is found.*/
            }
            /* Prefix option */
            else if( (ntohc(nd_option->TYPE) == ND6_OPTION_PREFIX)
    			   && ((ntohc(nd_option->LENGTH) << 3) >= sizeof(ND6_OPTION_PREFIX_HEADER)) )
            {
                 /* Note that there can be multiple "Prefix information" options included in a router advertisement.*/ 
                 if(prefix_index < RTCSCFG_ND6_PREFIX_LIST_SIZE)
                 {
                    nd_option_prefix[prefix_index] = (ND6_OPTION_PREFIX_HEADER_PTR)nd_option; /* Prefix information.*/
                    prefix_index++;
                 }
            }
            /* else, silently ignore any options they do not recognize
             * and continue processing the message.
             */
    				    
            nd_option_offset += (nd_option->LENGTH[0] << 3);
        }
        
        /************************************************************
         * Set parameters.
         ************************************************************/
        
        /* RFC4861 6.3.4: If the received Cur Hop Limit value is non-zero, the host SHOULD set
         * its CurHopLimit variable to the received value.
         */
        if(ntohc(ra_packet->CUR_HOP_LIMIT) != 0)
        {
            if_ptr->IP6_IF.ND6->cur_hop_limit = ntohc(ra_packet->CUR_HOP_LIMIT);
        }
        
        /* RFC4861 6.3.4: If the received Reachable Time value is non-zero, the host SHOULD set
         * its BaseReachableTime variable to the received value.
         */
        if(ntohl(ra_packet->REACHABLE_TIME) != 0)
        {
            if_ptr->IP6_IF.ND6->reachable_time = ntohl(ra_packet->REACHABLE_TIME);
        }        
        
        /* RFC4861 6.3.4:The RetransTimer variable SHOULD be copied from the Retrans Timer
         * field, if the received value is non-zero.
         */
        if(ntohl(ra_packet->RETRANS_TIMER) != 0)
        {
            if_ptr->IP6_IF.ND6->retrans_timer = ntohl(ra_packet->RETRANS_TIMER);
        } 

        /* RFC4861: Hosts SHOULD copy the options value
         * into LinkMTU so long as the value is greater than or equal to the
         * minimum link MTU [IPv6] and does not exceed the maximum LinkMTU value
         * specified in the link-type-specific document.
         */
        if(nd_option_mtu) 
        {
            if_ptr->IP6_IF.ND6->mtu = ntohl(nd_option_mtu->MTU);
            if(if_ptr->IP6_IF.ND6->mtu < IP6_DEFAULT_MTU)
                if_ptr->IP6_IF.ND6->mtu = IP6_DEFAULT_MTU;
        }
        
        /* RFC4861: If the advertisement contains a Source Link-Layer Address
         * option, the link-layer address SHOULD be recorded in the Neighbor
         * Cache entry for the router (creating an entry if necessary) and the
         * IsRouter flag in the Neighbor Cache entry MUST be set to TRUE. */
        neighbor_cache_entry = nd6_neighbor_cache_get(if_ptr, (in6_addr*)ip6_packet->SOURCE);
        if(nd_option_slla)
        {
            if(neighbor_cache_entry == NULL)
            /* Creating an entry if necessary */
            {
                neighbor_cache_entry = nd6_neighbor_cache_add(if_ptr, (in6_addr*)ip6_packet->SOURCE, nd_option_slla->addr, ND6_NEIGHBOR_STATE_STALE);
            }
            else
            {
                /* If a cache entry already exists and is
                 * updated with a different link-layer address, the reachability state
                 * MUST also be set to STALE.*/
                if( !LL_ADDR_ARE_EQUAL(nd_option_slla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN) )
                {
                    LL_ADDR_COPY(nd_option_slla->addr, neighbor_cache_entry->ll_addr, if_ptr->DEV_ADDRLEN);
                    
                    neighbor_cache_entry->state = ND6_NEIGHBOR_STATE_STALE;
                }
            }
            
            /* Sends any packets queued for the neighbor awaiting address resolution.
             */ 
            nd6_neighbor_send_waiting_pcb(if_ptr, neighbor_cache_entry);
        }
        else
        {
            if(neighbor_cache_entry == NULL)
            {
                neighbor_cache_entry = nd6_neighbor_cache_add(if_ptr, (in6_addr*)ip6_packet->SOURCE, NULL, ND6_NEIGHBOR_STATE_STALE);
            }
        }
        
        /*
         * RFC4861: If the address is already present in the hosts Default Router
         * List as a result of a previously received advertisement, reset
         * its invalidation timer to the Router Lifetime value in the newly
         * received advertisement.
         * If the address is already present in the hosts Default Router
         * List and the received Router Lifetime value is zero, immediately
         * time-out the entry.
         */
        nd6_router_list_add( neighbor_cache_entry, ntohs(ra_packet->ROUTER_LIFETIME));
        
        
        /*************************************************************
         * Prcess Prefix options.
         *************************************************************/
        if(prefix_index > 0)
        {
            for(i=0; i<prefix_index; i++)
            {
                ND6_PREFIX_ENTRY_PTR prefix_entry;
            #if RTCSCFG_DEBUG_MSG 
                {
                    char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
                    inet_ntop(AF_INET6, (char*)nd_option_prefix[i]->PREFIX, numaddr, sizeof(numaddr));  
                    
                    printf("Prefix[%d]= %s \n", i, numaddr);
                }
            #endif /* RTCSCFG_DEBUG_MSG */
                 
                /* RFC4861: A router SHOULD NOT send a prefix
                 * option for the link-local prefix and a host SHOULD
                 * ignore such a prefix option.*/
                if( !IN6_IS_ADDR_LINKLOCAL((in6_addr*)nd_option_prefix[i]->PREFIX)
                /* RFC4861: The value of Prefered Lifetime field MUST NOT exceed
                 * the Valid Lifetime field to avoid preferring
                 * addresses that are no longer valid.*/
                 && (ntohl(nd_option_prefix[i]->PREFERED_LIFETIME) <= ntohl(nd_option_prefix[i]->VALID_LIFETIME)) )
                {
                    /*************************************************************
                     * Prefix Information option with the on-link flag set. 
                     *************************************************************/
                    if( (ntohc(nd_option_prefix[i]->FLAG) & ND6_OPTION_FLAG_L) == ND6_OPTION_FLAG_L )
                    {
                        prefix_entry = nd6_prefix_list_get(if_ptr, (in6_addr*)nd_option_prefix[i]->PREFIX);
                        
                        /* RFC4861: If the prefix is not already present in the Prefix List, and the
                         * Prefix Information options Valid Lifetime field is non-zero,
                         * create a new entry for the prefix and initialize its
                         * invalidation timer to the Valid Lifetime value in the Prefix
                         * Information option.*/
                        if(prefix_entry == NULL)
                        {
                            if(ntohl(nd_option_prefix[i]->VALID_LIFETIME) != 0)
                            {
                                /* Create a new entry for the prefix.*/
                                nd6_prefix_list_add(if_ptr, (in6_addr*)nd_option_prefix[i]->PREFIX, 
                                                             ntohc(nd_option_prefix[i]->PREFIX_LENGTH), 
                                                             ntohl(nd_option_prefix[i]->VALID_LIFETIME));
                            }
                        }
                        else
                        {
                            /* RFC4861: If the prefix is already present in the hosts Prefix List as
                             * the result of a previously received advertisement, reset its
                             * invalidation timer to the Valid Lifetime value in the Prefix
                             * Information option. If the new Lifetime value is zero, time-out
                             * the prefix immediately.*/
                            if(ntohl(nd_option_prefix[i]->VALID_LIFETIME) != 0)
                            {
                                /* Reset Timer. */
                                prefix_entry->lifetime = ntohl(nd_option_prefix[i]->VALID_LIFETIME);
                                prefix_entry->creation_time = RTCS_time_get()/1000;
                            }
                            else
                            {   
                                /* Time-out the prefix immediately. */
                                nd6_prefix_list_del(prefix_entry);
                            }
                        }
                    }
                    
                    /*************************************************************
                     * Stateless Address Autconfiguration.
                     *************************************************************/
                     
                    /* For each Prefix-Information option in the Router Advertisement:*/
                    if( ((ntohc(nd_option_prefix[i]->FLAG) & ND6_OPTION_FLAG_A) == ND6_OPTION_FLAG_A )
                        && (ntohl(nd_option_prefix[i]->VALID_LIFETIME) != 0) )
                    {
                        int j;
                        IP6_IF_ADDR_INFO_PTR addr_info = NULL;
                        
                        /* RFC4862 5.5.3:If the advertised prefix is equal to the prefix of an address
                         * configured by stateless autoconfiguration in the list, the
                         * preferred lifetime of the address is reset to the Preferred
                         * Lifetime in the received advertisement. */
                         
                        /* Lookup the address */
                        for(j = 0; j < IP6_IF_ADDRESSES_MAX; j++)
                        {
                            if((if_ptr->IP6_IF.address[j].ip_addr_state != IP6_ADDR_STATE_NOT_USED)
                                && (if_ptr->IP6_IF.address[j].ip_addr_type == IP6_ADDR_TYPE_AUTOCONFIGURABLE)
                                && (ip6_addr_pefix_cmp((in6_addr*)nd_option_prefix[i]->PREFIX, &if_ptr->IP6_IF.address[j].ip_addr, ntohc(nd_option_prefix[i]->PREFIX_LENGTH)) == TRUE) )
                            {
                                addr_info = &if_ptr->IP6_IF.address[j];
                                break;
                            }    
                        }
                        
                        if( addr_info != NULL)
                        {
                            /* RFC4862 5.5.3: The specific action to
                             * perform for the valid lifetime of the address depends on the Valid
                             * Lifetime in the received advertisement and the remaining time to
                             * the valid lifetime expiration of the previously autoconfigured
                             * address. We call the remaining time "RemainingLifetime" in the
                             * following discussion:*/
                            if( (ntohl(nd_option_prefix[i]->VALID_LIFETIME) > (60*60*2) /* 2 hours */)
                               ||( ntohl(nd_option_prefix[i]->VALID_LIFETIME) >  ((addr_info->creation_time + addr_info->lifetime) - (RTCS_time_get()/1000)) ) 
                              )
                            {
                               /* 1. If the received Valid Lifetime is greater than 2 hours or
                                *    greater than RemainingLifetime, set the valid lifetime of the
                                *    corresponding address to the advertised Valid Lifetime. */
                                addr_info->lifetime = ntohl(nd_option_prefix[i]->VALID_LIFETIME);
                            
                            }
                            else
                            {
                               /* 2. If RemainingLifetime is less than or equal to 2 hours, ignore
                                *    the Prefix Information option with regards to the valid
                                *    lifetime, unless the Router Advertisement from which this
                                *    option was obtained has been authenticated (e.g., via Secure
                                *    Neighbor Discovery [RFC3971]). If the Router Advertisement
                                *    was authenticated, the valid lifetime of the corresponding
                                *    address should be set to the Valid Lifetime in the received
                                *    option.
                                * 3. Otherwise, reset the valid lifetime of the corresponding
                                *    address to 2 hours. */
                                addr_info->lifetime = (60*60*2) /* 2 hours */;
                            }
                            addr_info->creation_time = RTCS_time_get()/1000;
                        }
                        else
                        {
                        	/* RFC4862 5.5.3: If the prefix advertised is not equal to the prefix of an
                             * address configured by stateless autoconfiguration already in the
                             * list of addresses associated with the interface (where "equal"
                             * means the two prefix lengths are the same and the first prefixlength
                             * bits of the prefixes are identical), and if the Valid
                             * Lifetime is not 0, form an address (and add it to the list) by
                             * combining the advertised prefix with an interface identifier. */
                        	ip6_if_bind_addr (if_ptr, (in6_addr *)(&nd_option_prefix[i]->PREFIX[0]), IP6_ADDR_TYPE_AUTOCONFIGURABLE, 
                        	                 ntohl(nd_option_prefix[i]->VALID_LIFETIME), ntohc(nd_option_prefix[i]->PREFIX_LENGTH));
                        }
                    }
                    /* else. RFC4862: If the Autonomous flag is not set, silently ignore the Prefix Information option.*/
                }
            } /* for() end. */
        } /* End processing prefix options.*/
    }
DROP:    
    return result;    
}

/************************************************************************
* NAME: nd6_rd_start
* RETURS: None.
* DESCRIPTION: Start the Router Discovery for the interface.
*************************************************************************/
void nd6_rd_start(IP_IF_PTR if_ptr)
{
    if_ptr->IP6_IF.ND6->rd_transmit_counter = ND6_MAX_RTR_SOLICITATIONS-1;
    if_ptr->IP6_IF.ND6->rd_time = RTCS_time_get();  /* Save send time.*/
    nd6_router_solicitation_send(if_ptr);
    
    //TBD Randomise first send.
}

/************************************************************************
* NAME: nd6_rd_timer
* RETURS: None.
* DESCRIPTION: Timer routine used by Router Discovery.
*************************************************************************/
static void nd6_rd_timer(IP_IF_PTR if_ptr )
{
    if(if_ptr->IP6_IF.ND6->rd_transmit_counter > 0)
    {
        if(nd6_default_router_get(if_ptr) == NULL)
        /* Router is not found yet.*/
        {
            if((uint_32)(RTCS_time_get() - if_ptr->IP6_IF.ND6->rd_time ) > ND6_MAX_RTR_SOLICITATION_DELAY)
            {
                if_ptr->IP6_IF.ND6->rd_transmit_counter--;
                if_ptr->IP6_IF.ND6->rd_time = RTCS_time_get();  /* Save send time.*/
                nd6_router_solicitation_send(if_ptr);
            }
        }
        else
        /* Router is found.*/
        {
        #if RTCSCFG_DEBUG_MSG 
            printf("RD: ROUTER is FOUND!!\n");    
        #endif            
            if_ptr->IP6_IF.ND6->rd_transmit_counter = 0; /* Stop timer.*/
        }
    }
}

/************************************************************************
* NAME: nd6_dad_start
* RETURS: None.
* DESCRIPTION: Start the Duplicate Address Detection for the address.
*************************************************************************/
void nd6_dad_start(IP_IF_PTR if_ptr, IP6_IF_ADDR_INFO_PTR addr_info)
{
#if RTCSCFG_ND6_DAD_TRANSMITS > 0 
    if(addr_info && (addr_info->ip_addr_state == IP6_ADDR_STATE_TENTATIVE))
    {
        /* To check an address, a node sends RTCSCFG_ND6_DAD_TRANSMITS Neighbor
         * Solicitations, each separated by 1 second(TBD)..
         */
        addr_info->dad_transmit_counter = RTCSCFG_ND6_DAD_TRANSMITS; 
        addr_info->state_time = RTCS_time_get();  /* Save state time.*/
        nd6_neighbor_solicitation_send(if_ptr, NULL, NULL, &addr_info->ip_addr);
    }
#endif /* RTCSCFG_ND6_DAD_TRANSMITS */
    
    //TBD may be add more flags - for example per interface or gloval counter.
}

/************************************************************************
* NAME: nd6_dad_timer
* RETURS: None.
* DESCRIPTION: Timer routine used by Duplicate Address Detection.
*************************************************************************/
static void nd6_dad_timer(IP_IF_PTR if_ptr )
{

#if RTCSCFG_ND6_DAD_TRANSMITS > 0 
    int i;
    IP6_IF_ADDR_INFO_PTR addr_info;
    
    for(i=0; i<IP6_IF_ADDRESSES_MAX; i++)
    {
        addr_info = &if_ptr->IP6_IF.address[i];
        if(addr_info->ip_addr_state == IP6_ADDR_STATE_TENTATIVE)
        {
            if((uint_32)(RTCS_time_get() - addr_info->state_time ) > if_ptr->IP6_IF.ND6->retrans_timer )
            {
                addr_info->dad_transmit_counter--;
                if(addr_info->dad_transmit_counter == 0)
                {
                    /* DAD succeeded, for this address.
                    * Once an address is determined to be unique,
                    * it may be assigned to an interface.*/
                    addr_info->ip_addr_state = IP6_ADDR_STATE_PREFERRED;
                    
                    #if RTCSCFG_DEBUG_MSG 
                    {
                        char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
                        inet_ntop(AF_INET6, (char*)addr_info->ip_addr.s6_addr, numaddr, sizeof(numaddr));
                        printf("ND6: %s is PREFERED NOW\n", numaddr);
                    }
                    #endif /* RTCSCFG_DEBUG_MSG */
                }
                else
                {
                    addr_info->state_time = RTCS_time_get();
                    nd6_neighbor_solicitation_send(if_ptr, NULL, NULL, &addr_info->ip_addr);
                }
            }
        }
    }
#endif /* RTCSCFG_ND6_DAD_TRANSMITS */    
}

/************************************************************************
* NAME: nd6_dad_failed
* RETURS: None.
* DESCRIPTION: Called when DAD is failed.
*************************************************************************/
static void nd6_dad_failed(IP_IF_PTR if_ptr, IP6_IF_ADDR_INFO_PTR addr_info)
{
#if RTCSCFG_DEBUG_MSG 
    char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
    inet_ntop(AF_INET6, (char*)addr_info->ip_addr.s6_addr, numaddr, sizeof(numaddr));  
    printf("%s DAD is FAILED!!!!\n", numaddr);
#endif /* RTCSCFG_DEBUG_MSG */

    /* Just remove address, or TBD mark it as dupicate.*/
    ip6_if_unbind_addr ( if_ptr, addr_info );
}


//////////////// For DEBUG needs only. //////////////////////////////////

/************************************************************************
* NAME: nd6_debug_print_prefix_list
* RETURS: None.
* DESCRIPTION: Prints prefix list. For DEBUG needs only.
*************************************************************************/
void nd6_debug_print_prefix_list(IP_IF_PTR if_ptr )
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_PREFIX_ENTRY_PTR entry ;
    
    if(if_ptr == NULL)
        if_ptr = ip6_if_get_by_n(0); /* Get the first one.*/
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    
    if (nd6_cfg)
    {
        printf("Prefix List:\n");
        for(i=0; i<ND6_PREFIX_LIST_SIZE; i++)
        {
            if(nd6_cfg->prefix_list[i].state != ND6_PREFIX_STATE_NOTUSED)
            {
                char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
                entry = &nd6_cfg->prefix_list[i];
                
                /* Print entry.*/
                inet_ntop(AF_INET6, (char*)entry->prefix.s6_addr, numaddr, sizeof(numaddr));  
              
                printf("[%d] %s \\ %d State(%d) LifeTime(%d)[%d]\n", i, numaddr, entry->prefix_length, entry->state,  entry->lifetime, entry->creation_time); 
            }
        }

    }
}

/************************************************************************
* NAME: nd6_debug_print_neighbor_cache
* RETURS: None.
* DESCRIPTION: Prints Neighbor Cache. For DEBUG needs only.
*************************************************************************/
void nd6_debug_print_neighbor_cache(IP_IF_PTR if_ptr )
{
    ND6_CFG_PTR    nd6_cfg;
    int i;
    ND6_NEIGHBOR_ENTRY_PTR entry ;
    
    if(if_ptr == NULL)
        if_ptr = ip6_if_get_by_n(0); /* Get the first one.*/
    
    nd6_cfg = if_ptr->IP6_IF.ND6;
    
    if (nd6_cfg)
    {
        printf("Neighbor Cache:\n");
        for(i=0; i<ND6_NEIGHBOR_CACHE_SIZE; i++)
        {
            if(nd6_cfg->neighbor_cache[i].state != ND6_NEIGHBOR_STATE_NOTUSED)
            {
                char numaddr[sizeof("abcd:abcd:abcd:abcd:abcd:abcd:abcd:abcd")]={0};
                entry = &nd6_cfg->neighbor_cache[i];
                
                /* Print entry.*/
                inet_ntop(AF_INET6, (char*)entry->ip_addr.s6_addr, numaddr, sizeof(numaddr));  
                printf("[%d] %s State(%d) IsRouter(%d) RouterLifeTime(%d)[%d]\n", i, numaddr, entry->state, entry->is_router, entry->router_lifetime, entry->creation_time); 
            }
        }

    }
}

#endif /* RTCSCFG_ENABLE_IP6 */
